(******************************************************************************

 ProtoVeriPhy

 Module: distance/DelayedKeyDisclosure.thy

 Copyright (C) 2008-12 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
               2008 Patrick Schaller <patrick.schaller@inf.ethz.ch>
 License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Security Analysis of Protocols based on Delayed Key Disclosure *}

theory DelayedKeyDisclosure
imports MessageDerivationHashChains SystemSimps SystemInvariants
begin

subsection {* Parameters of the TESLA protocol *}

consts
 cn        :: nat      -- {* length of hash chain *}
 valwin    :: time     -- {* validity window *}
 Br        :: nat      -- {* honest broadcaster *}
 NBr       :: nat      -- {* used for initial Nonce *}

specification (cn)
 cn_lb: "cn \<ge> 2"
by (rule_tac x=2 in exI, auto)

specification (valwin)
  valwin_pos: "valwin > 0"
by (rule_tac x=1 in exI, auto)

consts
 t0 :: time

specification(t0)
  t0_pos [simp,intro]: "t0 \<ge> 0"
by auto

text {* corner cases: $k = 0 \to cn$, $k > cn \to hashNonce$ *}
 
definition
 ht :: "nat \<Rightarrow> real" where
 "ht k = t0 + (real k)*valwin"

text {* $inI 0 x \to x \in [t0,t0[$ *}

definition
  inI :: "nat \<Rightarrow> real \<Rightarrow> bool" where
  "inI k x = ((x \<ge>  (ht (k - 1))) \<and> (x < ht k))"

(*
release schedule:
  H^n      H^n     H^n-1        H^2      H^1     H^0     H^0   H^0 ...
   |
  K_0      K0      K1           Kcn-3    Kcn-2   Kcn-1   Kcn   Kcn+1
[ I1    [ I2   [ I3    [  ... [ Icn    [ Icn+1   [ Icn+2 [      [
t0     t1     t2       t3     tcn-1    tcn     tcn+1  tcn+2
 

use schedule:
  H^(cn-1) H^(cn-2)        H^0    H^0  ...      
   |       |               |      |
  K1       K2              Kcn    Kcn+1 ...
[ I1   [   I2      [ ... [ Icn  [
t0    t1                 tcn-2
*)

subsection {* assumptions about the environment *}

specification (clocktime)
  clocktime_Br[simp] : "clocktime Br t = t"
  apply best
  done

(* Br can create hashSecret anyways because it's "from his own nonceset" *)

text {* the function hashn returns for a
given n and a message m the nth element of the hash-chain, where m is
the starting element (i.e., it applies n times the hash function to
m*}

definition
  publicHash :: msg where
  "publicHash \<equiv> hashn cn (Nonce (Honest Br) NBr)"

definition
  initStateDKD :: "agent => msg set" where
  "initStateDKD A = {publicHash}"

lemma hashn_parts_IK:
  assumes a: "k < cn"
  shows "hashn k (Nonce (Honest Br) NBr) \<notin> parts (initStateDKD A)" using a
  apply (auto simp add: initStateDKD_def publicHash_def parts_hashn)
  apply (rotate_tac 1)
  apply (erule make_pos_goal)
  apply (rule hashn_inj2)
  apply auto
done

interpretation INITSTATE_NOHASHCHAIN_FREE initStateDKD NBr "Honest Br" cn
proof -
  have A: "\<And>B NA A. Nonce B NA \<notin> parts (initStateDKD A)"
    apply (auto simp add: initStateDKD_def publicHash_def)
    apply (insert cn_lb, case_tac cn)
    apply force
    apply (subgoal_tac "Nonce B NA = hashn (Suc nat) (Nonce (Honest Br) NBr)") defer
    apply (force)
    apply (thin_tac "Nonce B NA = hashn cn (Nonce (Honest Br) NBr)")
    apply (thin_tac "cn=Suc nat")
    apply (simp only: hashn.simps)
    apply auto
    done
  have B:
    "\<And>k A. k < cn \<Longrightarrow> hashn k (Nonce (Honest Br) NBr) \<notin> parts (initStateDKD A)"
    apply (erule hashn_parts_IK)
    done
  have C:
    "\<And>k A X. Hash \<lbrace>hashn k (Nonce (Honest Br) NBr), X\<rbrace> \<notin> parts (initStateDKD A)"
    apply (auto simp add: initStateDKD_def publicHash_def)
    done
  thus "INITSTATE_NOHASHCHAIN_FREE initStateDKD NBr (Honest Br) cn"
    apply (unfold_locales)
    apply (erule B)
    apply (rule C)
    apply (rule A)
    done
qed

definition
  dkd1 :: "msg step"
 where
  "dkd1 tr A t =
     (UN i n. {ev. ev = (\<lbrace>HMAC i (Number n), HKey (i - 2)\<rbrace>,SendEv 0 []) \<and> inI i t \<and>
                   A = Br})"

definition
  dkd2 :: "msg step"
 where
  "dkd2 tr Reciv t =
     (UN n i t_reca t_recb nb. 
           {ev. ev = (\<lbrace>(Number n), Number (int i)\<rbrace>,ClaimEv) \<and>
                (t_reca, Recv (Rec (Honest Reciv)) \<lbrace>HMAC i (Number n),HKey (i - 2)\<rbrace>)
                \<in> set tr \<and>
                inI i t_reca \<and>
                (t_recb, Recv (Rec (Honest Reciv)) \<lbrace>HMAC (i+2) (Number nb), HKey i\<rbrace>)
                \<in> set tr \<and>
                i < cn})"

definition
  dkd_proto :: "msg proto"
  where "dkd_proto = {dkd1,dkd2}"

interpretation PROTOCOL_EXECUTABLE Crypt Nonce MPair Hash Number parts subterms DM initStateDKD dkd_proto sys Key
  apply (unfold_locales)
  apply (auto simp add: dkd_proto_def initStateDKD_def occursAt.simps
            split: event.split split_if)
  apply (auto simp add: dkd1_def dkd2_def HMAC_def)
  apply (rule mac_derivable)
  apply (auto simp add: recv_a_view_a_r send_a_view_a_r)
done


inductive_set
 delayed :: "(msg trace) set"
 where
  Nil [intro] : "[] \<in> delayed"

 | Fake:
  "\<lbrakk> tr \<in> delayed;  t >=  maxtime tr;
     X \<in> DM (Intruder I) (knowsI (Intruder I) tr)\<rbrakk>
  \<Longrightarrow> (t, Send (Tx (Intruder I) j) X [])  # tr \<in> delayed"

 | Con :
  "\<lbrakk> tr \<in> delayed; (tsend, Send (Tx A i) m L) \<in> set tr;
      cdistM (Tx A i) (Rx B j) = Some tab;
      trecv >= maxtime tr; trecv \<ge> tsend + tab  \<rbrakk>
   \<Longrightarrow> (trecv, Recv (Rx B j) m) # tr \<in> delayed"

 (* Protocol rules for authenticated delayed *)
 | Broadcast: (* for i=0, the interval is empty *)
   "\<lbrakk> tr \<in> delayed; t \<ge> maxtime tr;
      inI i (clocktime Br t) \<rbrakk>
    \<Longrightarrow> (t, Send (Tr (Honest Br)) \<lbrace>HMAC i (Number n), HKey (i - 2)\<rbrace> [])#tr \<in> delayed"

 | Conclusion: (* we do not accept messages with HKey cn = hashSecret, since this
                  key is used twice *)
   "\<lbrakk> tr \<in> delayed; t \<ge> maxtime tr;
      i < cn;
      (t_reca, Recv (Rec (Honest Reciv)) \<lbrace>HMAC i (Number n),HKey (i - 2)\<rbrace>)
      \<in> set (view Reciv tr);
      inI i t_reca;
      (t_recb, Recv (Rec (Honest Reciv)) \<lbrace>HMAC (i+2) (Number nb), HKey i\<rbrace>)
      \<in> set (view Reciv tr)
    \<rbrakk>
    \<Longrightarrow> (t, Claim (Honest Reciv) \<lbrace>(Number n), Number (int i)\<rbrace>) # tr \<in> delayed"

lemma abstr_equal: "delayed = sys"
proof auto
  fix tr
  assume r: "tr \<in> sys"
  show "tr \<in> delayed" using r
  proof (induct tr)
    case Nil
    show ?case by auto
  next
    case (Fake tr t X I j)
    thus ?case apply -
      apply (rule delayed.Fake)
      apply auto
      done
  next
    case (Con tr tsend A i m B j tab trecv)
    thus ?case apply -
      apply (rule delayed.Con)
      apply auto
      done
  next
    case (Proto step m pEv A tr t)
    thus ?case apply -
      apply (unfold dkd_proto_def)
      apply auto
      apply (subgoal_tac "dkd1 \<in> dkd_proto") prefer 2
      apply (force simp add: dkd_proto_def)
      apply (simp add: events_occur_at) prefer 2
      apply (subgoal_tac "dkd2 \<in> dkd_proto") prefer 2
      apply (force simp add: dkd_proto_def)
      apply (simp add: events_occur_at) defer
      apply (auto simp add: dkd1_def)
      apply (rule delayed.Broadcast) prefer 4
      apply  (auto simp add: recv_a_view_a_r send_a_view_a_r dkd2_def)
      apply (rule_tac t_reca="t_reca" and t_recb="t_recb" in delayed.Conclusion)
      apply  (auto simp add: recv_a_view_a_r send_a_view_a_r dkd2_def)
    done
  qed
next
  fix tr
  assume r: "tr \<in> delayed"
  show "tr \<in> sys" using r
  proof (induct tr)
    case Nil
    show ?case by auto
  next
    case (Fake tr t X I j)
    thus ?case apply -
      apply (rule sys.Fake)
      apply auto
      done
  next
    case (Con tr tsend A i m B j tab trecv)
    thus ?case apply -
      apply (rule sys.Con)
      apply auto
      done
  next    
    case (Broadcast tr t i n)
    with prems have "(t,createEv Br (SendEv 0 []) \<lbrace>HMAC i (Number n), HKey (i - 2)\<rbrace>) # tr \<in> sys"
      apply -
      apply (rule_tac step=dkd1 in sys.Proto)
      apply (auto simp add: dkd_proto_def dkd1_def)
      apply (unfold DM_analz_synth knowsI_def)
      apply (rule_tac step=dkd1 in messages_derivable [simplified DM_analz_synth])
      apply (auto simp add: dkd_proto_def dkd1_def)
      done
    thus ?case using prems by auto
  next
    case (Conclusion tr t i t_reca Reciv n t_recb nb)
    with prems have "(t,createEv Reciv (ClaimEv) \<lbrace>Number n, Number (int i)\<rbrace>) # tr \<in> sys"
      apply (rule_tac step=dkd2 in sys.Proto)
      apply (auto simp add: dkd_proto_def dkd2_def)
      done
    thus ?case by auto
  qed
qed

lemmas [simp,intro] = abstr_equal [THEN sym]

subsection {* There exists a possible trace *}

text {* We will prove the existence of a "normal" protocol execution under the assumption that the clock-offset is smaller than the length of the validity window. Note that if the this not the case, then the receiver will never accept a message, and therefore will never introduce a claim-event.*}

lemma dkd_possibility:
 assumes distab: "cdistM (Tr (Honest Br)) (Rec (Honest Reciv)) \<noteq> None"
 assumes distableval: "(cdist (Tr (Honest Br)) (Rec (Honest Reciv))) < valwin"
 assumes clockrest: "inI 1 (clocktime Reciv (t0 + cdist (Tr (Honest Br)) (Rec (Honest Reciv))))"
 shows "\<exists> te i. \<exists> tr \<in> delayed. (te, Claim (Honest Reciv)\<lbrace>(Number n), Number (i)\<rbrace>) \<in> set tr"
proof -
 let ?dBR = "cdist (Tr (Honest Br)) (Rec (Honest Reciv))"
 from valwin_pos
 have "(t0,Send (Tr (Honest Br))
              \<lbrace>HMAC 1 (Number n),
                HKey (1 - 2)\<rbrace> []) # [] \<in> delayed" (is "?tr1 \<in> delayed")
    apply -
    apply (rule delayed.Broadcast)  apply (auto simp add: inI_def ht_def)
    done
 with distab t0_pos have "(t0 + ?dBR, Recv (Rec (Honest Reciv)) \<lbrace>HMAC 1 (Number n),
                                                          HKey (1 - 2)\<rbrace>)#?tr1 \<in> delayed"
                   (is "?tr2 \<in> delayed")
  apply -
  apply (rule_tac tab="cdist (Tr (Honest Br)) (Rec (Honest Reciv))" in delayed.Con)
  apply (auto)
  apply (auto simp add: cdist_def intro: cdistnoneg_some add_nonneg_nonneg)
 done
 with valwin_pos and distableval
 have "(t0+ 2 * valwin, Send (Tr (Honest Br))
               \<lbrace>HMAC 3 (Number n1), HKey (3 - 2)\<rbrace> []) # ?tr2 \<in> delayed" (is "?tr3 \<in> delayed")
   apply -
   apply (rule delayed.Broadcast)
   apply (assumption)
   apply (simp)
   apply (auto simp add: inI_def ht_def intro: add_nonneg_nonneg)
   done
 with distab valwin_pos distableval cdistnoneg_some t0_pos
   have a: "(t0+2*valwin + ?dBR, Recv (Rec (Honest Reciv))
                                     \<lbrace>HMAC 3 (Number n1), HKey 1\<rbrace>)#?tr3 \<in> delayed"
   (is "?tr4 \<in> delayed")
   apply -
   apply (rule_tac tab="cdist (Tr (Honest Br)) (Rec (Honest Reciv))" in delayed.Con)
   apply (assumption)
   apply (auto)
   apply (subgoal_tac "y \<ge> 0") defer
   apply (auto simp add: cdist_def intro: cdistnoneg_some)
   apply (rule add_nonneg_nonneg)
   apply (insert valwin_pos, auto)
   apply (rule add_nonneg_nonneg)+
   apply auto
   done
  with a distableval distab distableval cdistnoneg_some valwin_pos clockrest
    have "(t0 + 2*valwin + ?dBR,
           Claim (Honest Reciv) \<lbrace>(Number n), Number (int 1)\<rbrace>) # ?tr4 \<in> delayed"
   apply -
   apply (rule delayed.Conclusion [where t_reca="clocktime Reciv (t0 +?dBR)" and i=1
                                   and   nb=n1
                                   and   t_recb="clocktime Reciv (t0+2*valwin + ?dBR)"])
   apply assumption+
   apply (auto simp add: maxtime.psimps(2) cdist_def)
   apply (rule add_nonneg_nonneg, auto)+
   apply (insert cn_lb,simp)
   apply (rule_tac
          A="set (view Reciv [(t0 + y,
                              Recv (Rec (Honest Reciv)) \<lbrace>HMAC (Suc 0) (Number n), HKey 0\<rbrace>)])"
          in subsetD)
   apply (simp add: view_def)
   apply (simp add: view_def) defer
   apply (rule_tac
          A="set (view Reciv [(t0 + 2 * valwin + y,
                     Recv (Rec (Honest Reciv)) \<lbrace>HMAC 3 (Number n1), HKey (Suc 0)\<rbrace>)])"
          in subsetD)
   apply (simp add: view_def)
   apply (auto simp add: view_def)
   apply (subgoal_tac "Suc (Suc (Suc 0)) = 3") prefer 2
   apply force
   apply (simp (no_asm_simp))
   done
 thus ?thesis
   apply auto
   apply force
 done
qed

subsection {* Hashchain elements are secret until published *}

text {* note that only Br can derive hashn k hashSecret early, but nobody can
        send something containing hashn k hashSecret earlier *}

lemma hk_secret:
  assumes deriv: "HKey k \<in> parts (DM A (knowsI A tr))
                  \<or> (HKey k \<in> parts {X} \<and> (t, Send (Tx B txi) X L) \<in> set tr)" and
          kleqcn: "k <= cn" and notBr: "A \<noteq>  Honest Br" and  kpos: "k > 0" and
          delayed:"tr \<in> delayed"
  shows "maxtime tr  \<ge>  ht (k+1)" using delayed deriv notBr kleqcn kpos
proof (induct tr arbitrary: k A B X t L txi rule: delayed.induct)
  case Nil thus ?case using prems
    apply -
    apply (simp add: HKey_def)
    apply (drule hashn_derivI_hashn_received)
    apply auto
  done
next
  case (Fake tr ti Y I j)
  let ?ev = "(ti, Send (Tx (Intruder I) j) Y [])"
  (* adversary sends a message containing H^n(hashSecret), so the message must
     be derivable from the trace without the Send. We can just use the IH here *)
  show ?case proof cases
    assume "HKey k \<in> parts {Y}"
    with `Y \<in> DM (Intruder I) (knowsI (Intruder I) tr)`
    have "HKey k \<in> parts (DM (Intruder I) (knowsI (Intruder I) tr))" apply -
      apply (rule parts.elem_trans)
      by auto
    with prems show ?case apply -
      apply (rule maxtime_cons)
      apply (rule prems(7))
      apply force+
      done
  next
    (* now the hashSecret is no part of the new message Y, also use IH in
       the different cases *)
    assume "HKey k \<notin> parts {Y}"
    show ?case proof cases
      assume "HKey k \<in> parts (DM A (knowsI A (?ev#tr)))"
      hence "HKey k \<in> parts (DM A (knowsI A tr))"
        apply (auto simp add: DM_analz_synth knowsI_def)
	done
      with prems show ?thesis apply -
        apply (rule maxtime_cons)
        apply (rule prems(7))
        apply force+
      done
    next
      assume na: "HKey k \<notin> parts (DM A (knowsI A (?ev#tr)))"
      show ?case proof cases
        assume a: "HKey k \<in> parts {X} \<and> (t, Send (Tx B txi) X L) \<in> set (?ev#tr)"
        with `HKey k \<notin> parts {Y}` have "?ev \<noteq> (t, Send (Tx B txi) X L)"
          by auto
        with a have "(t, Send (Tx B txi) X L) \<in> set tr" by auto
        with prems show ?case apply -
          apply (rule maxtime_cons)
          apply (rule prems(7))
          apply force+
        done
      next
        assume nb: "\<not> (HKey k \<in> parts {X} \<and> (t, Send (Tx B txi) X L) \<in> set (?ev#tr))"
        with prems(10) na show ?case apply auto done
      qed
    qed
  qed
next
  case (Conclusion tr tc ic t_reca Reciv nc t_recb n2)
  let ?ev = "(tc, Claim (Honest Reciv) \<lbrace>Number nc, Number (int ic)\<rbrace>)"
  from prems have dm: "(HKey k \<in> parts (DM A (knowsI A (?ev#tr)))) =
                       (HKey k \<in> parts (DM A (knowsI A tr)))"
    apply (auto simp add: DM_analz_synth knowsI_def) done
  with prems have "HKey k \<in> parts (DM A (knowsI A tr))
                   \<or> (HKey k \<in> parts {X} \<and> (t, Send (Tx B txi) X L) \<in> set tr)"
    by auto
  with prems show ?case apply -
    apply (rule maxtime_cons)
    apply (rule prems(7)) by assumption+
next
  case (Broadcast tr tb ib n)
  have "\<exists> ibb. (ibb - 2) \<le>  cn \<and> HKey (ib - 2) = HKey (ibb - 2) \<and> (ibb \<le> ib)"
    by (rule HKey_overflow)
  then obtain ibb where ibbeq: "HKey (ib - 2) = HKey (ibb - 2)" and ibble: "ibb - 2 \<le> cn" and
                        ibbleib: "ibb \<le> ib" by auto
  let ?ev="(tb, Send (Tr (Honest Br))
                     \<lbrace>HMAC ib (Number n), HKey (ib - 2)\<rbrace> [])"
  show ?case proof cases
    assume "HKey k \<in> parts (DM A (knowsI A (?ev#tr)))"
    hence "HKey k \<in> parts (DM A (knowsI A tr))"
      by (auto simp add: DM_analz_synth knowsI_def)
    with prems show ?case apply -
      apply (rule maxtime_cons)
      apply (rule prems(7))
      by force+
  next
    assume  "HKey k \<notin> parts (DM A (knowsI A (?ev#tr)))"
    hence hn: "(HKey k \<in> parts {X} \<and> (t, Send (Tx B txi) X L) \<in> set (?ev#tr))"
      using prems(10) ibbeq by auto
    show ?case proof cases
      assume eveq: "?ev = (t, Send (Tx B txi) X L)"
      with eveq hn prems(12) ibbeq ibble have keq: "k=ibb - 2"
        apply auto
        apply (case_tac "cn - k",
	       auto simp add: hashSecret_def hashn.simps MACM_def HKey_def HMAC_def)
        apply (case_tac "cn - k", auto)
        apply (case_tac "cn - k", auto)
        apply (case_tac "nat", auto)
        apply (case_tac "cn - k", auto)
        apply (case_tac "cn - k", auto)
        apply (case_tac "cn - (ibb - 2)", auto)
        apply (case_tac "cn - (ibb - 2)", auto)
        apply (drule_tac X="Nonce (Honest Br) NBr" in hashn_inj1)
        apply auto
        done
      with prems(8) have tbeq: "maxtime (?ev#tr) = tb" by auto
      with prems(13) keq tbeq valwin_pos ibbeq have "ht (k+1) <= ht (ibb - 1)"
        apply (auto simp add: ht_def)
        done
      also have "... <= ht (ib - 1)" using prems(13) keq tbeq valwin_pos ibbleib
        apply (auto simp add: ht_def)
        done
      also have "... <= maxtime (?ev#tr)"using prems tbeq
        apply (simp add: inI_def) done
      finally show ?case by auto
    next
      assume "?ev \<noteq> (t, Send (Tx B txi) X L)"
      with hn have "(HKey k \<in> parts {X} \<and> (t, Send (Tx B txi) X L) \<in> set tr)"
        by auto
      with prems show ?case apply -
        apply (rule maxtime_cons)
        apply (rule prems(7))
        apply force+
      done
    qed
  qed
next
  case (Con tr tsend C ic Y Lc D j tab trecv)
  let ?ev = "(trecv, Recv (Rx D j) Y)"
  show ?case proof cases
    assume "HKey k \<in> parts (DM A (knowsI A (?ev#tr)))"
    with prems
    have "\<exists> X t i j. (t, Recv (Rx A i) X) \<in> set (?ev#tr) \<and> HKey j \<in> parts {X} \<and> j >= k
                     \<and> j \<le> cn"
      apply -
      apply (simp only: HKey_def)
      apply (drule hashn_derivI_hashn_received)
      apply force
      apply force
      apply (elim exE)
      apply (rule_tac x=Xb in exI)
      apply (rule_tac x=tb in exI)
      apply (rule_tac x=i in exI)
      apply (rule_tac x="cn-ja" in exI)
      apply auto
      done
    then obtain Z t i jh where rtr: "(t, Recv (Rx A i) Z) \<in> set (?ev#tr)" and
                               hnparts: "HKey jh  \<in> parts {Z}" and jhlecn: "jh \<le> cn" and
                               jleqk: "jh >= k" by auto
    from prems have "?ev#tr \<in> delayed" apply - apply (rule delayed.Con) by auto
    with prems rtr have
          ex: "\<exists>E j tsend Le. (tsend, Send (Tx E j) Z Le) \<in> set (?ev#tr)"
      apply -
      apply (rule_tac tb=t and RB="Rx A i" in send_before_recv_notime)
      by auto
    then obtain tsend E i Le where "(tsend, Send (Tx E i) Z Le) \<in> set (?ev#tr)" by auto
    with hnparts have hn: "HKey jh \<in> parts {Z} \<and> (tsend, Send (Tx E i) Z Le) \<in> set (tr)"
      by auto
    have "ht (k + 1) \<le> ht (jh + 1)" using jleqk valwin_pos
      apply (simp only: ht_def)
      apply auto
      done
    also have "ht (jh+1) \<le> maxtime ((trecv, Recv (Rx D j) Y) # tr)" using prems hn
      apply -
      apply (rule maxtime_cons)
      apply (rule prems(7))
      apply (rule disjI2)
      apply assumption
      apply force defer
      apply (insert jleqk, force)
      apply (auto simp add: inI_def)
      done
    finally show ?case by auto
  next
    assume "HKey k \<notin> parts (DM A (knowsI A (?ev#tr)))"
    with prems have "(HKey k \<in> parts {X} \<and> (t, Send (Tx B txi) X L) \<in> set tr)"
      by auto
    with prems show ?case apply -
      apply (rule maxtime_cons)
      apply (rule prems(7)) by auto
  qed
qed

subsection {* MAC Origination *}

text {* if there is a MACed number/message on the network before the corresponding
  key is disclosed, then this MAC must originate from the broadcaster *}

lemma hmac_originates_before_disclosure:
  assumes dkd: "tr \<in> delayed" and
          intr: "(t, Send (Tx A txi) Y L) \<in> set tr" and
          parts: "Hash \<lbrace>HKey k, Number n\<rbrace> \<in> parts {Y}" and
          before: "maxtime tr < ht (k+1)" and
          kle: "k < cn" and
          kgez: "k > 0"
  shows "\<exists> tb. (tb, Send (Tr (Honest Br))
                         \<lbrace>HMAC k (Number n), HKey (k - 2)\<rbrace> []) \<in> set tr"
   using dkd intr before kle kgez parts
proof (induct tr arbitrary: A k n Y L t txi rule: delayed.induct)
  case Nil thus ?case by auto
next
  case (Con tr tsend C ic Y Lc D j tab trecv)
  thus ?case by auto
next
  case (Conclusion tr tc t_reca Reciv ic n t_rec2 n2)
  thus ?case by auto
next
  case (Broadcast tr tb ib nb)
  let ?ev="(tb, Send (Tr (Honest Br))
                     \<lbrace>Hash[HKey ib] (Number nb) , HKey (ib - 2)\<rbrace> [])"
  (* the corner cases are interesting:
          HKey cn,   HKey cn - 2 = hashSecret, H^2 hashSecret
          HKey cn+1, HKey cn - 1 = hashSecret, H hashSecret 
          HKey cn+2, HKey cn     = hashSecret, hashSecret 
         
          But they can not happen because k < cn
  *)
  show ?case proof cases
    assume eveq: "?ev = (t, Send (Tx A txi) Y L)"
    hence meq: "Y = \<lbrace>Hash[HKey ib] (Number nb) , HKey (ib - 2)\<rbrace>" by auto
    thm prems(15)
    from meq prems(15) have hkeq: "HKey k = HKey ib"
      apply (simp add: MACM_def)
      apply (auto simp add: HKey_def hashSecret_def)
      done
    from meq prems(15) have neq: "n = nb"
      by (auto simp add: HKey_def hashSecret_def)
    from hkeq prems(13) have iblcn: "ib < cn"
      apply (auto simp add: HKey_def)
      apply (drule_tac k="cn-k" and X="hashSecret" in hashn_inj1)
      apply auto
      done
    with prems(13) hkeq iblcn have keq: "k=ib"
      apply (auto simp add: HKey_def)
      apply (drule hashn_inj1)
      apply auto
      done
    with eveq iblcn prems(13) have keq1: "HKey (ib - 2) = HKey (k - 2)"
      by (auto simp add: HKey_def)

    with keq iblcn neq show ?case apply -
      apply (rule_tac x=tb in exI) apply auto
      done
  next
    assume "?ev \<noteq> (t, Send (Tx A txi) Y L)"
    with prems have "(t, Send (Tx A txi) Y L) \<in> set tr" by (auto simp add: HMAC_def)
    with prems show ?case apply -
      apply (rule trace_weaken)
      apply (rule prems(8))
      apply auto
      done
  qed
next
  case (Fake tr ti Z I j)
  let ?ev = "(ti, Send (Tx (Intruder I) j) Z [])"
  show ?case proof cases
    assume "?ev = (t, Send (Tx A txi) Y L)"
    hence "Z=Y" by auto
    with prems have dm: "Hash \<lbrace>hashn (cn -k) hashSecret,Number n\<rbrace>
                         \<in> parts (DM (Intruder I) (knowsI (Intruder I) tr))"
      apply -
      apply (rule_tac G="{Y}" in parts.trans) defer
      apply (auto simp add: HKey_def)
      done
    (* proof sketch: mac derived ==> either Key received or MAC received
                     implies either Key send (use h_secret) or MAC send: use IH *)
    
    with prems
    have a: "(\<exists>Y t i j. (t, Recv (Rx (Intruder I) i) Y) \<in> set tr \<and>
                        hashn j hashSecret \<in> parts {Y} \<and> j \<le> (cn - k)) \<or>
             (\<exists>Y t i. (t, Recv (Rx (Intruder I) i) Y) \<in> set tr \<and>
                      Hash \<lbrace>hashn (cn - k) hashSecret, Number n\<rbrace> \<in> parts {Y})"
      apply -
      apply (unfold HKey_def)
      apply (erule mac_derivI_mac_or_hashn_received)
      apply auto
      done
    show ?thesis proof cases
      assume "(\<exists>Y t i j. (t, Recv (Rx (Intruder I) i) Y) \<in> set tr \<and>
                     hashn j hashSecret \<in> parts {Y} \<and> j \<le> (cn - k))"
      then obtain Z th ih jh where itr: "(th, Recv (Rx (Intruder I) ih) Z) \<in> set tr" and
                                   parts: "hashn jh hashSecret \<in> parts {Z}" and
	                           jhleq: "jh \<le> (cn - k)" by auto
      with prems itr have
        ex: "\<exists>D txis t_send Ld. (t_send, Send (Tx D txis) Z Ld) \<in> set tr"
      apply -
      apply (rule_tac tb=th and RB="Rx (Intruder I) ih" in send_before_recv_notime)
      by (auto)
      then obtain t_send D txis Ld where sitr: "(t_send, Send (Tx D txis) Z Ld) \<in> set tr"
        by auto
      from prems(12) have maxt: "maxtime tr < ht (k + 1)" by auto
      from jhleq prems have "k \<le> cn - jh" apply -
        apply (rule_tac c=jh in add_le_imp_le_right)
        apply simp
        done
      with itr parts valwin_pos jhleq have "ht (k + 1) \<le> ht ((cn - jh) + 1)"
        apply (auto simp add: ht_def)
        done
      also have "... \<le> maxtime tr" using prems sitr parts jhleq apply -
        apply (rule_tac X=Z in hk_secret) defer
        apply force
        apply force
        apply force
        apply force
        apply (rule disjI2)
        apply (auto simp add: HKey_def)
	done
      finally show ?thesis using maxt by auto
    next
      assume "\<not> (\<exists>Y t i j. (t, Recv (Rx (Intruder I) i) Y) \<in> set tr \<and>
                           hashn j hashSecret \<in> parts {Y} \<and> j \<le> (cn - k))"
      with a have "(\<exists>Y t i. (t, Recv (Rx (Intruder I) i) Y) \<in> set tr \<and>
                      Hash \<lbrace>hashn (cn - k) hashSecret, Number n\<rbrace> \<in> parts {Y})" by auto 
      then obtain M th ih where ritr: "(th, Recv (Rx (Intruder I) ih) M) \<in> set tr" and
                                parts: "Hash \<lbrace>hashn (cn - k) hashSecret, Number n\<rbrace> \<in> parts {M}"
        by auto
      with ritr have "\<exists> A i t_send La. (t_send, Send (Tx A i) M La) \<in> set tr" using prems
	apply -
	apply (rule_tac tb=th and RB="Rx (Intruder I) ih" in send_before_recv_notime)
	by (auto)
      then obtain t_send D txis Ld where sitr: "(t_send, Send (Tx D txis) M Ld) \<in> set tr"
        by auto
      thus ?case using prems sitr parts apply -
        apply (rule trace_weaken)
        apply (rule_tac Y=M in prems(8))
        apply (auto simp add: HKey_def)
        done
    qed
  next
    assume "?ev \<noteq> (t, Send (Tx A txi) Y L)"
    with prems have "(t, Send (Tx A txi) Y L) \<in> set tr" by auto
    with prems show ?case apply -
      apply (rule trace_weaken)
      apply (rule prems(8))
      apply auto
      done
  qed
qed

lemma BR_send_interval:
  assumes dkd: "tr \<in> delayed" and
          klcn: "k < cn" and
          intr: "(t_send, Send (Tr (Honest Br))
                               \<lbrace>Hash[HKey k] (Number n), X\<rbrace> []) \<in> set tr"
  shows "inI k t_send" using dkd intr klcn
proof (induct tr arbitrary: t_send k n X)
  case Nil thus ?case by auto
next
  case (Con tr tsend C ic Y D j tab trecv)
  thus ?case by auto
next
  case (Conclusion tr tc t_reca Reciv ic n t_rec2 n2)
  thus ?case by auto
next
  case (Fake tr ti Y I j)
  thus ?case by auto
next
  case (Broadcast tr tb ib nb)
  let ?ev="(tb, Send (Tr (Honest Br))
                     \<lbrace>Hash[HKey ib] (Number nb), HKey (ib - 2)\<rbrace> [])"
  show ?case proof cases
    assume eq: "?ev = (t_send, Send (Tr (Honest Br)) \<lbrace> Hash[HKey k] (Number n), X\<rbrace> [])"
    hence b: "t_send=tb" by auto
    have "0 \<le> maxtime tr" by auto
    also have "... \<le> t_send" using prems by auto
    finally have "0 \<le> t_send" by auto
    with prems have "ib > 0" apply (auto simp add: inI_def ht_def HMAC_def)
      apply (case_tac "ib=0", auto)
      done
    from eq prems(9) have "ib < cn"
      apply (auto simp add: HKey_def)
      apply (drule_tac k="cn-ib" and X="hashSecret" in hashn_inj1)
      apply auto
      done
    with prems(10) have a: "k=ib" using eq apply (auto simp add: HKey_def)
      apply (drule hashn_inj1) by auto
    with prems show ?case apply (auto simp add: inI_def) done
  next
    assume "?ev \<noteq> (t_send, Send (Tr (Honest Br)) \<lbrace> Hash[HKey k] (Number n), X\<rbrace> [])"
    with prems have "(t_send, Send (Tr (Honest Br)) \<lbrace> Hash[HKey k] (Number n), X\<rbrace> []) \<in> set tr"
      by (auto simp add: HMAC_def)
    with prems show ?case apply -
      apply (rule prems(5))
      apply auto
      done
  qed
qed

subsection {* T-Authentication for TESLA *}

text {*  The main security guarantee we can get is what we call in our
BAP-paper T-authentication. This implies that if there is a Claim-Event
in a given trace, where the Claim-Event refers to a certain message
(Number n) and a certain interval (Number (int i)), then indeed there
exists a send-event of the honest broadcast node in the interval i,
where the honest broadcast node has sent Number n.*}

lemma view_elem_ex: (* move to SystemSimps *)
  "(t,ev) \<in> (set (view A tr)) \<Longrightarrow> \<exists> t'. (t',ev) \<in> set tr \<and> clocktime A t' = t"
by (auto simp add: view_def split: split_if_asm)

lemma t_authentication:
 assumes intr: "tr \<in> delayed" and
         loosely_synced: "!!t. t - clocktime Reciv t  < valwin" and
         igz: "i > 0" and
         dev: "(t, Claim (Honest Reciv) \<lbrace>(Number n), Number (int i)\<rbrace>) \<in> set tr"
 shows "\<exists> t_send X. (t_send, Send (Tr (Honest Br))
                                  \<lbrace>HMAC i (Number n), X\<rbrace> []) \<in> set tr
                    \<and> (inI i t_send)"
 using intr dev igz
proof (induct tr arbitrary: t n i X)
  case Nil thus ?case by simp
next
  case Fake thus ?case by simp
next
  case Con thus ?case by simp
next
  case  Broadcast thus ?case by force
next
  case (Conclusion tr tc ic ct_reca Recivc nc ct_recb n2)
  let ?ev = "(tc, Claim (Honest Reciv) \<lbrace>Number nc, Number (int ic)\<rbrace>)"
  let ?msg = "\<lbrace>Hash[HKey ic] (Number nc), HKey (ic - 2)\<rbrace>"
  show ?case proof cases
    assume meq: "(t, Claim (Honest Reciv) \<lbrace>Number n, Number (int i)\<rbrace>) =
                 (tc, Claim (Honest Recivc) \<lbrace>Number nc, Number (int ic)\<rbrace>)"
    from meq have ieq: "i=ic" and teq: "t = tc" and neq: "n=nc" by auto
    with meq ieq have
      "\<exists> t_reca. (t_reca, Recv (Rec (Honest Reciv)) \<lbrace>HMAC ic (Number n),HKey (i - 2)\<rbrace>)
                  \<in> set tr \<and> clocktime Reciv t_reca = ct_reca" using prems apply -
      apply (rule view_elem_ex)
      by auto
    then obtain t_reca where atr: "(t_reca, Recv (Rec (Honest Reciv)) \<lbrace>HMAC ic (Number n),HKey (i - 2)\<rbrace>)
                                   \<in> set tr"
                             and cta: "clocktime Reciv t_reca = ct_reca"
      by auto
    with meq ieq prems have
      "\<exists>B ti tsi Lb. (tsi, Send (Tx B ti) ?msg Lb) \<in> set tr
               \<and> t_reca >= cdist (Tx B ti) (Rec (Honest Reciv))
               \<and> cdistM (Tx B ti) (Rec (Honest Reciv)) \<noteq> None
               \<and> tsi \<le> t_reca - cdist (Tx B ti) (Rec (Honest Reciv))"
      apply - apply (rule send_before_recv)
      by (auto simp add: HMAC_def)
    then obtain tsi B ti Lb where sitr: "(tsi, Send (Tx B ti) ?msg Lb) \<in> set tr" and
                                     "tsi \<le> t_reca - cdist (Tx B ti) (Rec (Honest Reciv))" and
                                     "cdistM (Tx B ti) (Rec (Honest Reciv)) \<noteq> None"
      
      by auto
    hence tsile: "tsi \<le> t_reca" apply - apply auto
      apply (unfold cdist_def, simp)
      apply (drule cdistnoneg_some)
      apply arith
      done
    let ?sev = "(tsi, Send (Tx B ti) ?msg Lb)"
    let ?trb = "?sev#(beforeEvent ?sev tr)"
    from sitr prems have bitr: "?trb \<in> delayed" apply -
      apply (rule beforeEvent_cons_event_delayed [simplified])
      apply auto
      done
    have z1: "real (Suc ic) = real ic + (1::real)" apply auto done
    hence d: "real ic * valwin + (1::real) * valwin = real (Suc ic) * valwin"
      apply (simp add: z1)
      by (rule Groebner_Basis.class_semiring.semiring_rules(2))
    from prems have c: "ct_reca < ht (ic)" using prems apply (auto simp add: inI_def) done
    from tsile have "tsi \<le> t_reca" by auto
    also have "... < ht (ic + 1)" using cta valwin_pos c loosely_synced [of t_reca]
      apply (auto simp add: ht_def)
      apply (subgoal_tac "t_reca < t0 + (real ic * valwin + (1::real) * valwin)") prefer 2
      apply force
      apply (simp only: d)
      done
    finally have tsiless: "tsi < ht (ic + 1)" .
    hence "maxtime ?trb = tsi" using prems(5) sitr apply (auto)
      apply (insert beforeEvent_maxtime [of tr tsi "snd ?sev"])
      apply auto
      done
    with bitr sitr tsile bitr prems tsiless have "\<exists> tb. (tb, Send (Tr (Honest Br))
                           \<lbrace>HMAC ic (Number nc), HKey (ic - 2)\<rbrace> []) \<in> set ?trb"  apply -
      apply (rule hmac_originates_before_disclosure)
      apply (auto simp add: inI_def)
      done
    then obtain tb where "(tb, Send (Tr (Honest Br))
      \<lbrace>HMAC ic (Number nc), HKey (ic - 2)\<rbrace> []) \<in> set ?trb" by auto
    with prems(4) sitr have a:"(tb, Send (Tr (Honest Br))
      \<lbrace>HMAC ic (Number nc), HKey (ic - 2)\<rbrace> []) \<in> set tr" apply -
      apply (rule beforeEvent_prefix)
      apply auto
      done
    with prems have b: "inI ic tb " apply -
      apply (rule_tac tr=tr in BR_send_interval)
      apply (auto simp add: HMAC_def)
      done
    with a ieq neq show ?case apply - apply (rule_tac x=tb in exI)
      apply (rule_tac x="HKey (ic - 2)" in exI)
      apply (auto simp add: HMAC_def)
      done
  next
    assume "(t, Claim (Honest Reciv) \<lbrace>Number n, Number (int i)\<rbrace>) \<noteq>
            (tc, Claim (Honest Recivc) \<lbrace>Number nc, Number (int ic)\<rbrace>)"
    with prems have int: "(t, Claim (Honest Reciv) \<lbrace>Number n, Number (int i)\<rbrace>) \<in> set tr"
      by auto
    from int prems have "\<exists>t_send X.
          (t_send, Send (Tr (Honest Br)) \<lbrace>HMAC i (Number n), X\<rbrace> [])
          \<in> set tr \<and> inI i t_send" apply -
      apply (rule prems(6))
      apply auto
      done
    thus ?case apply -
      apply auto
      done
  qed
qed

end