(******************************************************************************

  ProtoVeriPhy

  Module: distance/MessageTheory.thy

  Copyright (C) 2007-2012 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
  License: BSD (see COPYRIGHT)

******************************************************************************)

header{*Message Theory Locale*}

theory MessageTheory imports Misc begin

subsection{* The Notion of Subterms *}

locale MESSAGE_THEORY_SUBTERM_NOTION =
  fixes   f   :: "'msg set \<Rightarrow> 'msg set"
  assumes inj[intro]: "X \<in> H \<Longrightarrow> X \<in> f H"
  and     singleton: "X \<in> f H ==> \<exists>Y\<in>H. X \<in> f {Y}"
  and     mono: "G \<subseteq> H ==> f G \<subseteq> f H"
  and     idem [simp]: "f (f H) = f H"

begin

subsubsection{*Idempotence and Transitivity *}

lemma empty [simp]: "f {} = {}"
by (auto dest: singleton)

lemma emptyE [elim!]: "X\<in> f{} ==> P"
by simp

lemma increasing: "H \<subseteq> f H"
by auto

lemma subset_iff [simp]: "(f G \<subseteq> f H) = (G \<subseteq> f H)"
  apply (rule iffI)
  apply (iprover intro: subset_trans increasing)  
  apply (frule mono, simp)
done

lemma trans: "[| X\<in> f G;  G \<subseteq> f H |] ==> X \<in> f H"
  apply (drule mono)
  apply (subgoal_tac "f G \<subseteq> f H")
  apply (erule rev_subsetD)
by auto

subsubsection{*Unions *}

lemma Un_subset1: "f(G) \<union> f(H) \<subseteq> f(G \<union> H)"
by auto

lemma Un_subset2: "f (G \<union> H) \<subseteq> f (G) \<union> f (H)"
  apply auto
  apply (drule singleton)
  apply auto
  apply (erule trans)
  apply force
  apply (erule contrapos_np)
  apply (erule trans)
  apply force
done

lemma Un [simp]: "f(G \<union> H) = f(G) \<union> f(H)"
by (intro equalityI Un_subset1 Un_subset2)

lemma insert: "f (insert X H) = f {X} \<union> f H"
  apply (subst insert_is_Un [of _ H])
  apply (simp only: Un)
done

lemma insert2:
     "f (insert X (insert Y H)) = f {X} \<union> f {Y} \<union> f H"
  apply (simp add: Un_assoc)
  apply (simp add: insert [symmetric])
done

lemma UN_subset1: "(\<Union>x\<in>A. f(H x)) \<subseteq> f(\<Union>x\<in>A. H x)"
by (intro UN_least mono UN_upper)

lemma UN_subset2: "f(\<Union>x\<in>A. H x) \<subseteq> (\<Union>x\<in>A. f(H x))"
  apply auto
  apply (drule singleton)
  apply auto
  apply (rule_tac x=a in bexI) prefer 2
  apply force
  apply (erule trans)
  apply force
done

lemma UN [simp]: "f (\<Union>x\<in>A. H x) = (\<Union>x\<in>A. f (H x))"
by (intro equalityI UN_subset1 UN_subset2)

text{*This allows @{text blast} to simplify occurrences of 
  @{term "parts(G\<union>H)"} in the assumption.*}
 
lemmas in_parts_UnE = Un [THEN equalityD1, THEN subsetD, THEN UnE] 
declare in_parts_UnE [elim!]

lemma insert_subset: "insert X (f H) \<subseteq> f(insert X H)"
by auto

text{*Cut*}
lemma cut:
     "[| Y\<in> f (insert X G);  X\<in> f H |] ==> Y\<in> f (G \<union> H)"
  apply (erule trans)
by auto

(* standard  attribute breaks locales *)
lemmas insertI = subset_insertI [THEN mono, THEN subsetD]

lemma cut_eq [simp]: "X\<in> f H ==> f (insert X H) = f H"
by (force dest!: cut intro: insertI)

lemmas insert_eq_I = equalityI [OF subsetI insert_subset]

lemma bUnion [simp]:
  "f (\<Union>x\<in>H. G x) = (\<Union>x\<in>H. f (G x))"
by auto

lemma set: "X \<in> f {m. m = a \<and> P} \<Longrightarrow> X \<in> f {m. m = a}"
  apply auto
  apply (erule trans)
  apply auto
done

lemma elem_trans:
  assumes a: "X \<in> f {Y}" and b: "Y \<in> f H"
  shows "X \<in> f H" using a b
proof -
  from a have c: "{X} \<subseteq> f {Y}" by auto
  with b have "{Y} \<subseteq> f H" by auto
  with c show ?thesis apply - apply (rule trans) by auto
qed

lemma fst_set: "X \<in> f (fst ` {ev. ev = (a,b) \<and> C}) \<Longrightarrow> X \<in> f {a}"
  apply (erule rev_subsetD)
  apply auto
done

lemma mono_elem: "\<lbrakk> x \<in> f H; H \<subseteq> G \<rbrakk> \<Longrightarrow> x \<in> f G"
  apply (drule mono)
by (erule rev_subsetD)

end

subsection{* Required Constructors for Message Theories*}

locale MESSAGE_THEORY_DATA =
  fixes Key   :: "key \<Rightarrow> 'msg"
  and   Crypt :: "key \<Rightarrow> 'msg \<Rightarrow> 'msg"
  and   Nonce :: "agent \<Rightarrow> nat \<Rightarrow> 'msg"
  and   MPair :: "'msg \<Rightarrow> 'msg \<Rightarrow> 'msg"
  and   Hash  :: "'msg \<Rightarrow> 'msg"
  and Number  :: "int \<Rightarrow> 'msg"
begin

definition
  MACM :: "['msg,'msg] => 'msg"                       ("(4Hash[_] /_)" [0, 1000])
    --{*Message Y paired with a MAC computed with the help of X*}
 where
    "Hash[X] Y == MPair (Hash (MPair X Y)) Y"

end

subsection {* Message Derivation: Constructors, parts, subterms, and DM *}

locale MESSAGE_THEORY_PARTS = MESSAGE_THEORY_DATA Key  +
  parts: MESSAGE_THEORY_SUBTERM_NOTION parts
    for Key :: "key \<Rightarrow> 'msg" and parts :: "'msg set \<Rightarrow> 'msg set"

locale MESSAGE_THEORY_SUBTERM = MESSAGE_THEORY_PARTS _ _ _ _ _ Key  +
  subterms: MESSAGE_THEORY_SUBTERM_NOTION subterms
    for Key :: "key \<Rightarrow> 'msg" and subterms :: "'msg set \<Rightarrow> 'msg set" +
  assumes parts_subset_subterms: "!!H. parts H \<subseteq> subterms H"
begin

lemmas parts_in_subterms = parts_subset_subterms[THEN subsetD]

end

locale MESSAGE_THEORY_DM = MESSAGE_THEORY_SUBTERM _ _ _ _ _ _  Key for Key :: "key \<Rightarrow> 'msg" +
  fixes DM   :: "agent \<Rightarrow> 'msg set \<Rightarrow> 'msg set"
  fixes LowHam :: "'msg set"
  fixes distort :: "'msg \<Rightarrow> 'msg \<Rightarrow> 'msg"
  fixes components :: "'msg set \<Rightarrow> 'msg set"

locale MESSAGE_DERIVATION = MESSAGE_THEORY_DM _ _ _ _ _ _ _ Key for Key :: "nat \<Rightarrow> 'msg" + 
  assumes nonce_subterms_DM_nonce:
          "!! A.
            Nonce B NB \<in> subterms (DM A H) \<Longrightarrow>
            A \<noteq> B
           \<Longrightarrow> Nonce B NB \<in> subterms H"
  assumes nonce_parts_DM_nonce:
          "!! A.
            Nonce B NB \<in> parts (DM A H) \<Longrightarrow>
            A \<noteq> B
           \<Longrightarrow> Nonce B NB \<in> parts H"
  and      key_parts_DM_key:
           "!!A.
             Key k \<in> parts (DM A H)
            \<Longrightarrow> Key k \<in> parts H"
  and     sig_subterms_DM_sig_or_key:
          "!!H A.
            Crypt k msig \<in> subterms (DM A H)
           \<Longrightarrow> Crypt k msig \<in> subterms H
              \<or> Key k \<in> parts H"
  and     mac_subterms_DM_mac_or_key:
          "Hash (MPair (Key k) m) \<in> subterms (DM A H)
           \<Longrightarrow> Hash (MPair (Key k) m) \<in> subterms H
              \<or> Key k \<in> parts H"

(*
  and     distort_LowHam:
          "distort X Y \<in> LowHam \<Longrightarrow> \<exists> d \<in> LowHam. X = distort Y d"

  and     distort_comm:
          "distort X Y = distort Y X"

  and     key_parts_distortion:
          "\<lbrakk> d \<in> LowHam; Key k \<in> parts {distort m d} \<rbrakk>
           \<Longrightarrow> Key k \<in> parts {m}"


  and     key_not_LowHam:
          "\<lbrakk> d \<in> LowHam; Key k \<in> subterms {distort m d} \<rbrakk>
           \<Longrightarrow> Key k \<in> subterms {m}"

  and     nonce_not_LowHam:
          "\<lbrakk> d \<in> LowHam; Nonce A N \<in> subterms {distort m d} \<rbrakk>
           \<Longrightarrow> Nonce A N \<in> subterms {m}"

  and     crypt_not_LowHam:
          "\<lbrakk> d \<in> LowHam; Crypt E F \<in> subterms {distort m d} \<rbrakk>
           \<Longrightarrow> Crypt E F \<in> subterms {m}"

  and     hash_not_LowHam:
          "\<lbrakk> d \<in> LowHam; Hash c \<in> subterms {distort m d} \<rbrakk>
           \<Longrightarrow> Hash c \<in> subterms {m}"
*)
  and     components_subset_parts: (* implies subset subterms *)
          "x \<in> components S \<Longrightarrow> x \<in> parts S"

  and     key_components_parts:
          "Key k \<in> parts S \<Longrightarrow> \<exists> m \<in> components S. Key k \<in> parts {m}"

  and     nonce_components_subterm:
          "Nonce A N \<in> subterms S \<Longrightarrow> \<exists> m \<in> components S. Nonce A N \<in> subterms {m}"

  and     hash_components_subterm:
          "Hash c \<in> subterms S \<Longrightarrow> \<exists> m \<in> components S. Hash c \<in> subterms {m}"

  and     crypt_components_subterm:
          "Crypt k m \<in> subterms S \<Longrightarrow> \<exists> M \<in> components S. Crypt k m \<in> subterms {M}"



end