(******************************************************************************

  ProtoVeriPhy

  Module: distance/TimeSyncESPS.thy

  Copyright (C) 2010-12 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
  License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Security Analysis of the ESPS time synchronization protocol *}

theory TimeSyncESPS
imports Event SystemCoffset SystemOrigination MessageTheoryFree
begin

locale INITSTATE_SYMKEYS_NN = INITSTATE_SYMKEYS + INITSTATE_NONONCE

text {* Here we define the locale describing the necessary initial
state for the protocol, i.e., agents have pairwise shared keys and
their knowledge initially does not contain nonces of other agents. *}

definition
  initStateESPS :: "agent \<Rightarrow> msg set" where
  "initStateESPS A == Key`((symKey A)`UNIV)"

interpretation INITSTATE_SYMKEYS_NN Crypt Nonce MPair Hash Number parts subterms DM
               initStateESPS Key
by (unfold_locales, auto simp add: initStateESPS_def dest: injective_symKey)


text {* min\_delay denotes the minimal delay it takes for a protocol
        message to be transmitted to one of the honest nodes. More
        precisely, the minimal time it takes between sending of the
        first bit until the complete message is received.  bound\_delay
        is the maximal delay that is expected and a messages that
        takes longer is assumed to be dealyed by the adversary and
        will not be accepted.  *}

consts
  min_delay :: real
  bound_delay :: real

specification (min_delay)
  min_delay_pos: "min_delay > (0::real)"
by (rule_tac x=1 in exI, arith)

specification (bound_delay)
  bound_delay_pos: "bound_delay > (0::real)"
  bound_delay_min: "bound_delay \<ge> min_delay"
by (insert min_delay_pos,rule_tac x="min_delay + 1" in exI,simp)

text {* The following definitions have to be understood in the context
of the challenge-response structure of the time-synchronization
protocol. All time stamps denote local clock-readings of the following
events:
t1: the initiator starts the protocol by sending a fresh nonce
(sending starts) 
t2: the responder has received the challenge (reception finished)
t3: the responder starts the transmission of the response
t4: the initiator has received the response (reception finished)

Note that we assume a constant clock-offset for the duration of the
protocol execution, since synchronization would not make much sense,
if arbitrary clock-offset were allowed (e.g., consider the case where
a clock stands still.)  *}

definition
  offset :: "real \<Rightarrow> real \<Rightarrow> real \<Rightarrow> real \<Rightarrow> real"
 where
  "offset t1 t2 t3 t4 \<equiv> ((t2 - t1) - (t4 - t3))/2"

definition
  delay :: "real \<Rightarrow> real \<Rightarrow> real \<Rightarrow> real \<Rightarrow> real"
 where
  "delay t1 t2 t3 t4 \<equiv> ((t2 - t1) + (t4 - t3))/2"


subsection {* Direct and Parameterized Definition and their Equality *}

subsubsection {* Step Functions for Authenticated Ranging *}

definition
  sps1 :: "msg step"
 where
 "sps1 tr A t =
    (UN NA B. {ev. ev = (\<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace>,SendEv 0 []) \<and>
                   Nonce (Honest A) NA \<notin> usedI tr})"

definition
  sps2 :: "msg step"
 where
 "sps2 tr B tsend =
    (UN NA A A' trec_start trec_complete NB.
       {ev. ev = (Nonce (Honest B) NB,
                  SendEv 0 [Agent A, Nonce A' NA, Real trec_complete]) \<and>
            (trec_start, Recv (Rx (Honest B) 0) \<lbrace>Agent A, Agent (Honest B), Nonce A' NA\<rbrace>) \<in> set tr \<and>
            trec_complete \<le> tsend \<and>
            Nonce (Honest B) NB \<notin> usedI tr \<and>
            trec_complete \<ge> trec_start + min_delay})"

definition
  sps3 :: "msg step"
 where
 "sps3 tr B ts =
    (UN NA A A' trec_complete trec_start tsend NB.
       {ev. ev = (Hash[Key (symKey (Honest B) A)]
                    \<lbrace>Agent (Honest B), Agent A, Nonce A' NA, Nonce (Honest B) NB,
                     Real trec_complete, Real tsend\<rbrace>,
                  SendEv 0 []) \<and>
                  (trec_start, Recv (Rx (Honest B) 0) \<lbrace>Agent A, Agent (Honest B), Nonce A' NA\<rbrace>) \<in> set tr \<and>
                  (tsend, Send (Tr (Honest B)) (Nonce (Honest B) NB)
                       [Agent A, Nonce A' NA, Real trec_complete]) \<in> set tr})"

definition
  sps4 :: "msg step"
 where
 "sps4 tr A tclaim =
    (UN NA B B' t1 t2 t3 t4_start t4_complete t5 NB.
       {ev. ev = (\<lbrace>Agent B, Real (offset t1 t2 t3 t4_complete)\<rbrace>,ClaimEv) \<and>
            (t5, Recv (Rx (Honest A) 0)
                   Hash[Key (symKey B (Honest A))]
                     \<lbrace>Agent B, Agent (Honest A), Nonce (Honest A) NA, Nonce B' NB, Real t2, Real t3\<rbrace>)
            \<in> set tr \<and>

            (t4_start, Recv (Rx (Honest A) 0) (Nonce B' NB)) \<in> set tr \<and>
            
            (t1, Send (Tr (Honest A)) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> []) \<in> set tr \<and>
            t4_complete \<le> tclaim \<and>
            t4_complete \<ge> t4_start + min_delay \<and>
            (delay t1 t2 t3 t4_complete) \<le> bound_delay })"

definition
  sps_proto :: "msg proto" where
  "sps_proto = {sps1,sps2,sps3,sps4}"

lemmas sps_defs = sps_proto_def sps1_def sps2_def sps3_def sps4_def

text {* The following locale defines the necessary conditions on the
  initial knowledge distribution (distribution of public and private
  keys) and requires that, according to the protocol specification, no
  private keys are sent over the network. The locale is then
  instantiated using the cryptographic functions, the intruders'
  abilities to derive messages, and the step functions of the
  synchronization protocol. *}

locale PROTOCOL_ESPS =
  PROTOCOL_SYMKEYS_NOKEYS+PROTOCOL_NONONCE+INITSTATE_SYMKEYS_NN

interpretation PROTOCOL_ESPS Crypt Nonce MPair Hash Number parts subterms DM initStateESPS Key sps_proto
 apply (unfold_locales)
 apply (auto simp add: sps_defs messagesProtoTr_def messagesProtoTrHonest_def
              split: event.split split_if dest: parts.fst_set)
done

lemma (in PROTOCOL_EXECUTABLE) sys_Proto_exec:
  "\<lbrakk> sys = sys_param; tr \<in> sys_param; maxtime tr \<le> t;
     step \<in> proto; (m, pEv) \<in> step (timetrans A tr) A (clocktime A t)\<rbrakk>
   \<Longrightarrow> (t, createEv A pEv m) # tr \<in> sys"
  apply (rule sys.Proto)
  apply simp+
  apply (subst events_occur_at)
  apply simp+
  apply (rule messages_derivableI)
  apply simp
  apply (subst events_occur_at)
  apply simp+
done

text {* The following interpretation of the locale PROTOCOL\_EXECUTABLE
  ensures that agents only look at their own views and all messages are
  derivable. *}

lemma sps_proto_induct[consumes 1]:
  "\<lbrakk> step \<in> sps_proto; P sps1; P sps2; P sps3; P sps4 \<rbrakk> \<Longrightarrow> P step"
by (unfold sps_proto_def, blast)

interpretation PROTOCOL_EXECUTABLE Crypt Nonce MPair Hash Number parts subterms DM initStateESPS  sps_proto sys Key
proof (unfold_locales)
  fix step :: "msg step" and
      m :: msg and
      pEv :: "msg protoEv" and
      A :: nat and
      tr :: "(real \<times> msg event) list" and
      t :: real
  assume "step \<in> sps_proto"
  assume "(m, pEv) \<in> step (view A tr) A (clocktime A t)"
  show "m \<in> DM (Honest A) (knows (Honest A) tr \<union> initStateESPS (Honest A))" using prems
  proof (induct step rule: sps_proto_induct)
    case 1 show ?case using prems 
      by (auto simp  add: sps_defs)
  next
    case 2 show ?case using prems 
      by (auto simp  add: sps_defs)
  next
    case 3 show ?case using prems apply -
      apply (auto simp  add: sps_defs)
      apply (unfold MACM_def)
      apply (auto simp add: DM_analz_synth initStateESPS_def)
      apply (rule synth.MPair)
      apply (rule synth.Hash)
      apply (rule synth.MPair)
      apply force
      apply (rule synth.MPair)
      apply force
      apply (rule synth.MPair)
      apply force
      apply (rule synth.MPair) defer
      apply (force intro!: synth.MPair)
      apply (rule synth.MPair)
      apply force
      apply (rule synth.MPair)
      apply force
      apply (rule synth.MPair) defer
      apply (force intro!: synth.MPair)
      apply (drule view_elem_ex)
      apply (erule exE)
      apply (drule Recv_imp_knows_A) back
      apply (rule synth.Inj)
      apply (rule UnI2)
      apply (drule analz.Inj)
      apply (drule analz.Snd)
      apply (drule analz.Snd)
      apply (rule subsetD [OF analz_mono]) prefer 2
      apply force+
      apply (drule view_elem_ex)
      apply (erule exE)
      apply (drule Recv_imp_knows_A) back
      apply (rule synth.Inj)
      apply (rule UnI2)
      apply (drule analz.Inj)
      apply (drule analz.Snd)
      apply (drule analz.Snd)
      apply (rule subsetD [OF analz_mono]) prefer 2
      apply force+
      done
  next
    case 4 show ?case using prems
      by (auto simp  add: sps_defs)
  qed
next
  fix tr ::"(real \<times> msg event) list" and
      step :: "msg step" and
      A :: nat and
      t :: real
  assume "tr \<in> PROTOCOL.sys DM initStateESPS sps_proto"
  assume "step \<in> sps_proto"
  show   "step (view A tr) A t = step (timetrans A tr) A t" using prems
    apply -
    apply (auto simp add: sps_defs)
    apply (auto simp add: nonce_view_fresh [simplified sps_proto_def]
                          nonce_view_used [simplified sps_proto_def])
    apply (auto simp add: recv_a_view_a_r send_a_view_a_r)
    apply (rule_tac x=NA in exI)
    apply (rule_tac x=B' in exI)
    apply (rule_tac x=t1 in exI)
    apply (rule_tac x=t2 in exI)
    apply (rule_tac x=t3 in exI)
    apply (rule_tac x=t4_start in exI)
    apply (rule_tac x=t4_complete in exI)
    apply force
    apply (rule_tac x=NA in exI)
    apply (rule_tac x=B' in exI)
    apply (rule_tac x=t1 in exI)
    apply (rule_tac x=t2 in exI)
    apply (rule_tac x=t3 in exI)
    apply (rule_tac x=t4_start in exI)
    apply (rule_tac x=t4_complete in exI)
    apply force
    done
qed

subsubsection {* In the following we define the a set of rules that
  directly defines the set of trace for the ESPS time-synchronization
  protocol. *}

inductive_set
  ESPS :: "(msg trace) set"

 where
   Nil [intro] : "[] \<in> ESPS"

 | Fake: 
   "\<lbrakk> tr \<in> ESPS; t \<ge> maxtime tr;
      X \<in> DM (Intruder I) (knowsI (Intruder I) tr)\<rbrakk>
   \<Longrightarrow> (t, Send (Tx (Intruder I) j) X [])  # tr \<in> ESPS"

 | Con : 
   "\<lbrakk> tr \<in> ESPS; trecv \<ge> maxtime tr;
      (tsend, Send (Tx A i) m L) \<in> set tr;
      cdistM (Tx A i) (Rx B j) = Some tab;
      trecv \<ge> tsend + tab \<rbrakk>
    \<Longrightarrow> (trecv, Recv (Rx B j) m) # tr \<in> ESPS"

 -- {* Protocol rules for ESPS, correspond to step functions *}
 | ESPS1:  
   "\<lbrakk> tr \<in> ESPS; t >=  maxtime tr ;
      \<not> (usedI tr (Nonce (Honest A) NA)) \<rbrakk> 
        (* cannot use set notation here: infer_arities: bad term (Isabelle Bug?) *)
   \<Longrightarrow> (t, Send (Tx (Honest A) 0) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> []) # tr \<in> ESPS"

 | ESPS2:
   "\<lbrakk> tr \<in> ESPS;  tsend \<ge>  maxtime tr;
      \<not> (usedI tr (Nonce (Honest B) NB)); 
      trec_complete \<ge> trec_start + min_delay;
      tsend \<ge> trec_complete;
      (trec_start, Recv (Rx (Honest B) 0) \<lbrace>Agent A, Agent (Honest B), Nonce A' NA\<rbrace>) \<in> set tr \<rbrakk>
   \<Longrightarrow> (tsend, Send (Tx (Honest B) 0) (Nonce (Honest B) NB)
                   [Agent A, Nonce A' NA, Real (trec_complete + coffset B)])
      # tr \<in> ESPS"

 | ESPS3:
   "\<lbrakk> tr \<in> ESPS;  ts \<ge>  maxtime tr;
      (trec_start, Recv (Rx (Honest B) 0) \<lbrace>Agent A, Agent (Honest B), Nonce A' NA\<rbrace>) \<in> set tr \<and>
      (tsend, Send (Tx (Honest B) 0) (Nonce (Honest B) NB)
        [Agent A, Nonce A' NA, Real trec_complete]) \<in> set tr\<rbrakk>
   \<Longrightarrow> (ts,
        Send (Tx (Honest B) 0)
             Hash[Key (symKey (Honest B) A)]
                 \<lbrace>Agent (Honest B), Agent A, Nonce A' NA, Nonce (Honest B) NB,
                  Real (trec_complete),
                  Real (tsend + coffset B)\<rbrace>
             [])
      # tr \<in> ESPS"

 | ESPS4:
   "\<lbrakk> tr \<in> ESPS; t \<ge> maxtime tr;
      delay (t1 + coffset A) t2 t3 (t4_complete + coffset A) \<le> bound_delay;
      t4_complete \<ge> t4_start + min_delay;
      t \<ge> t4_complete;
      (t5, Recv (Rx (Honest A) 0)
                Hash[Key (symKey B (Honest A))]
                    \<lbrace>Agent B, Agent (Honest A), Nonce (Honest A) NA, Nonce B' NB, Real t2, Real t3\<rbrace>)
      \<in> set tr;
      (t4_start, Recv (Rx (Honest A) 0) (Nonce B' NB)) \<in> set tr;
      (t1, Send (Tr (Honest A)) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> []) \<in> set tr\<rbrakk>
  \<Longrightarrow> (t, Claim (Honest A) \<lbrace>Agent B,
                           Real (offset (t1 + coffset A) t2 t3 (t4_complete + coffset A))\<rbrace>) # tr
      \<in> ESPS"

text {* We obtain a simpler induction principle for the protocol since
  it is executable and deltaonly. *}

lemmas proto_induct =
  sys.induct [simplified derivable_removable remove_occursAt]

subsubsection {* Equality for the direct and the parameterized definitions *}

text {* We show that both inductive definitions define the
        same set of traces. *}

lemma abstr_equal: "ESPS = sys"
proof auto
  fix tr
  assume r: "tr \<in> sys"
  show "tr \<in> ESPS" using r
  proof (induct tr rule: proto_induct)
    case 1 with prems show ?case by simp
  next
    case 2 with prems show ?case by (blast intro: ESPS.Nil)
  next
    case 4 with prems show ?case by (blast intro: ESPS.Con)
  next
    case 3 with prems show ?case by (blast intro: ESPS.Fake)
  next
    case (5 tr t step m pEv A)
    thus ?case
      apply (auto simp add: sps_defs)
      apply (rule ESPS.ESPS1)
      apply force+
      apply (simp add: mem_def)
      apply (unfold timetrans_def)
      apply auto
      apply (drule_tac trec_start="a" and tsend=t and B=A and NB=NB and
	               trec_complete="trec_complete - coffset A" in ESPS.ESPS2)
      apply assumption
      apply (force simp add: mem_def)
      apply force+
      apply (drule_tac B=A and trec_complete="trec_complete" in ESPS.ESPS3)
      apply assumption prefer 2
      apply force
      apply (rule conjI)
      apply assumption+
      apply (drule_tac t=t and A=A and B'=B' and
	               t4_complete="t4_complete - coffset A" in ESPS.ESPS4)
      apply auto
      done
      qed
    next
      fix tr
      assume r: "tr \<in> ESPS"
      show "tr \<in> sys" using r
      proof(induct tr rule: ESPS.induct)
        case Nil
        with prems show ?case by force
      next
        case (Fake tr ts X I j)
        with prems show ?case by (auto intro: sys.Fake)
      next
        case (Con tr tsend C i X D j dCD trecv)
        with prems show ?case by (auto intro: sys.Con)
      next
        case (ESPS1 tr ts C NC D)
        with prems have "(ts,createEv C (SendEv 0 []) \<lbrace>Agent (Honest C), Agent D, Nonce (Honest C) NC\<rbrace>) # tr \<in> sys"
          apply -
          apply (rule_tac step=sps1 in sys_Proto_exec)
          apply (force simp add: sps_defs)+
          apply (auto simp add: sps_defs)
          apply (simp add: mem_def)
          done
        thus ?case by (auto simp add: createEv.psimps)
      next
    case (ESPS2 tr tsend B NB trec_complete trec_start A A' NA)
    with prems have
      "(tsend, 
        createEv B
                 (SendEv 0 [Agent A, Nonce A' NA, Real (trec_complete + coffset B)])
                 (Nonce (Honest B) NB)) # tr \<in> sys"
      apply -
      apply (rule_tac step=sps2 in sys_Proto_exec)
      apply force+
      apply (force simp add: sps_defs)
      apply (auto simp add: sps_defs)
      apply (rule_tac x="trec_start + coffset B" in exI)
      apply auto
      apply (auto simp add: in_timetrans)
      apply (auto simp add: mem_def)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (ESPS3 tr ts trec_start B A A' NA tsend NB trec_complete)
    with prems have
      "(ts, 
        createEv B
                 (SendEv 0 [])
                 (Hash[Key (symKey (Honest B) A)]
                    \<lbrace>Agent (Honest B), Agent A, Nonce A' NA, Nonce (Honest B) NB,
                     Real (trec_complete), Real (tsend  + coffset B)\<rbrace>)) # tr \<in> sys"
      apply -
      apply (rule_tac step=sps3 in sys_Proto_exec)
      apply force+
      apply (force simp add: sps_defs)
      apply (auto simp add: sps_defs)
      apply (auto simp add: in_timetrans)
      apply (rule_tac x="trec_start + coffset B" in exI)
      apply auto
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (ESPS4 tr t t1 A t2 t3 t4_complete t4_start t5 B NA B' NB)
    with prems have
      "(t,
        createEv A ClaimEv
                 (\<lbrace>Agent B, Real (offset (t1 + coffset A) t2 t3 (t4_complete + coffset A))\<rbrace>)) # tr \<in> sys"
      apply - 
      apply (rule_tac step=sps4 in sys_Proto_exec)
      apply (force simp add: sps_defs)+
      apply (auto simp add: sps_defs)
      apply (rule_tac x=NA in exI)
      apply (rule_tac x=B' in exI)
      apply (auto simp add: in_timetrans)
      apply (rule_tac x="t1 + coffset A" in exI)
      apply (rule_tac x=t2 in exI)
      apply (rule_tac x=t3 in exI)
      apply (rule_tac x="t4_start + coffset A" in exI)
      apply (rule_tac x="t4_complete + coffset A" in exI)
      apply (intro conjI)
      apply force defer defer
      apply (auto simp add: offset_def)
      apply (rule_tac x="t5 + coffset A" in exI)
      apply auto
      done
    thus ?case by (auto simp add: createEv.psimps)
  qed
qed

lemmas [simp,intro] = abstr_equal [THEN sym]

subsection {* A "possibility property"*}

text {* If certain conditions on the
communication distance between prover and verifier are met, the
protocol may complete succesfully, i.e., under certain conditions a
verifier introduces a claim-event into a communication trace.*}

lemma delay_offset_remove[simp]:
  "delay (ta + ca) (tb + cb) (tc + cb) (td + ca) = delay ta tb tc td"
  by (unfold delay_def, auto)

lemma ESPS_possibility:
 assumes bound_ok: "delay 0
                          (cdist (Tr (Honest A)) (Rec (Honest (A+1))) + min_delay)
                          (cdist (Tr (Honest A)) (Rec (Honest (A+1))) + min_delay)
                          (cdist (Tr (Honest A)) (Rec (Honest (A+1))) + min_delay + cdist (Tr (Honest (A+1))) (Rec (Honest A)) + min_delay)  \<le> bound_delay"
 assumes distab: "cdistM (Tr (Honest A)) (Rec (Honest (A+1))) \<noteq> None"
 assumes distba: "cdistM (Tr (Honest (A+1))) (Rec (Honest A)) \<noteq> None"
 shows "\<exists>ta dt. \<exists>tr \<in> ESPS. (ta,Claim (Honest A) \<lbrace>Agent (Honest (A+1)), Real dt\<rbrace>)
        \<in> set tr"
proof -
  let ?dab = "cdist (Tr (Honest A)) (Rec (Honest (A+1)))" and
      ?dba = "cdist (Tr (Honest (A+1))) (Rec (Honest A))" and
      ?fa = "Honest A" and
      ?fb = "Honest (A+1)" and
      ?na = "Nonce (Honest A) NA" and
      ?nb = "Nonce (Honest (A+1)) NB"
  let ?msg1 = "\<lbrace>Agent ?fa, Agent ?fb, ?na\<rbrace>"
  let ?msg2 = "Hash[Key (symKey ?fb ?fa)]
                    \<lbrace>Agent ?fb, Agent ?fa, ?na, ?nb,
                     Real ((?dab + min_delay) + coffset (A+1)), Real ((?dab + min_delay)  + coffset (A+1))\<rbrace>"
  let ?t1 = 0 and
      ?t2 = "?dab+min_delay + coffset (A+1)"
  let ?t3 = "?t2" and
      ?t4 = "?dab+min_delay+(?dba+min_delay)"
  let ?done = "Claim ?fa \<lbrace>Agent ?fb, Real (offset (?t1 + coffset A) ?t2 ?t3 (?t4 + coffset A))\<rbrace>"
  from distab have dabnoneg: "?dab \<ge> 0"
    by (auto simp add: cdist_def intro: cdistnoneg_some)
  from distba have dbanoneg: "?dba \<ge> 0"
    by (auto intro: cdistnoneg_some simp add: cdist_def)
  from dabnoneg dbanoneg have snoneg: "?dab+?dba \<ge> 0" by auto
  have "(0,Send (Tr ?fa) ?msg1 [])#[] \<in> ESPS" (is "?tr1 \<in> ESPS")
    apply - apply (rule ESPS.ESPS1)
    apply (auto simp add: usedI_def)
    by (simp add: mem_def [THEN Pure.symmetric, where x1="Nonce (Honest A) NA"])
  with distab dabnoneg have "(0+?dab,Recv (Rec ?fb) ?msg1)#?tr1 \<in> ESPS" (is "?tr2 \<in> ESPS")
     apply -
     apply (rule_tac tab="cdist (Tr (Honest A)) (Rec (Honest (A+1)))" in ESPS.Con)
     by (auto simp add: cdist_def)
  with dabnoneg have "(?dab + min_delay,Send (Tr ?fb) ?nb [Agent ?fa, ?na, Real (?dab + min_delay + coffset (A+1))])#?tr2 \<in> ESPS"
    (is "?tr3 \<in> ESPS") apply -
    apply (rule_tac tr="?tr2" in ESPS.ESPS2)
    prefer 6
    apply force+
    apply (insert min_delay_pos)
    apply force defer
    apply force+
    apply (auto simp add: usedI_def initStateESPS_def)
    apply (simp add: mem_def [THEN Pure.symmetric, where x1="Nonce (Honest (Suc A)) NB"])
    done
  with dabnoneg have "(?dab + min_delay,Send (Tr ?fb) ?msg2 [])#?tr3 \<in> ESPS"
    (is "?tr4 \<in> ESPS") apply -
    apply (rule_tac tr="?tr3" in ESPS.ESPS3)
    apply auto
    apply (insert min_delay_pos)
    apply auto
    done
  hence "((?dab+min_delay)+?dba,Recv (Rec ?fa) ?msg2)#?tr4 \<in> ESPS"
    (is "?tr5 \<in> ESPS")
    using snoneg distba dbanoneg apply -
    apply (insert min_delay_pos)
    apply (rule ESPS.Con)
    apply auto
    by (auto simp add: cdist_def)
  hence "((?dab+min_delay)+?dba,Recv (Rec ?fa) ?nb)#?tr5 \<in> ESPS"
    (is "?tr6 \<in> ESPS")
    using snoneg distba dbanoneg apply -
    apply (insert min_delay_pos)
    apply (rule ESPS.Con)
    apply auto
    by (auto simp add: cdist_def)
  hence "(?dab + min_delay + ?dba + min_delay, ?done)#?tr6 \<in> ESPS" (is "?tr7 \<in> ESPS")
    apply -
    apply (insert min_delay_pos)
    apply (rule_tac tr="?tr6" in ESPS.ESPS4)
    apply (insert dbanoneg snoneg)
    apply (insert bound_ok)
    apply (auto simp add: delay_offset_remove [where ta=0 and ca="coffset A" and cb="coffset (A+1)", simplified])
    done
  -- {* This is the witness trace: @{term [display] ?tr5} *}
  thus ?thesis
    apply -
    apply (rule_tac x="?dab + min_delay+(?dba+min_delay)" in exI)
    apply (rule_tac x="offset (?t1 + coffset A) ?t2 ?t3 (?t4 + coffset A)" in exI)
    apply (intro bexI) by auto
qed

subsection {* Security Proof: If the protocol has been successfully
completed by a verifier, then his conclusions about the offset between
his own and the prover's clock are correct.  *}

definition
  abs :: "real \<Rightarrow> real"
 where
  "abs x \<equiv> if 0 \<le> x then x else - x"

lemma minus_eq: "a = b ==> - (a ::real) = - b" by simp

lemma claim_implies:
  assumes sps: "tr \<in> ESPS"
  assumes claim: "(t,Claim (Honest A) \<lbrace>Agent B, Real off\<rbrace>) \<in> set tr"
  shows "\<exists> t1 t2B t3B t4_start t4 NA B' NB t t5.
            (delay (t1 + coffset A) t2B t3B (t4 + coffset A)) \<le> bound_delay \<and>
            t4 \<ge> t4_start + min_delay \<and>
            t \<ge> t4 \<and>
            off = offset (t1 + coffset A) t2B t3B (t4 + coffset A) \<and>
            (t4_start, Recv (Rx (Honest A) 0) (Nonce B' NB)) \<in> set tr \<and>
            (t5,
             Recv (Rx (Honest A) 0)
                      Hash[Key (symKey B (Honest A))]
                          \<lbrace>Agent B, Agent (Honest A), Nonce (Honest A) NA,
                           Nonce B' NB, Real t2B, Real t3B\<rbrace>)
            \<in> set tr \<and>
            (t1, Send (Tr (Honest A)) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> [])
            \<in> set tr"
  using assms
by (induct rule: ESPS.induct, force+)

lemma nonce_send_honest:
  assumes sps: "tr \<in> ESPS" and
          send: "(ta, Send (Tx (Honest A) i) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> L) \<in> set tr" 
  shows   "Nonce (Honest A) NA
           \<notin> usedI (beforeEvent (ta, Send (Tx (Honest A) i) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> L) tr)"
  using assms
proof (induct rule: ESPS.induct)
  case Nil thus ?case by simp
next
  case (Con tr tc C mc D tab)
  hence "(ta, Send (Tx (Honest A) i) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> L) \<in> set tr" by simp
  with Con.hyps prems show ?case by simp
next
  case (Fake tr mintr I tsend)
  hence "(ta, Send (Tx (Honest A) i)  \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> L) \<in> set tr" by simp
  with Fake.hyps prems show ?case by simp
next
  case (ESPS2 tr tsend trec_complete trec_start C D D' ND)
  hence "(ta, Send (Tx (Honest A) i)  \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> L) \<in> set tr" by simp
  with ESPS2.hyps prems show ?case by simp
next
  case (ESPS3 tr t t1 D t2 t3 t4_complete t4_start C ND)
  hence "(ta, Send (Tx (Honest A) i) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace>   L) \<in> set tr" by simp
  with ESPS3.hyps prems show ?case by simp
next
  case (ESPS1 tr t D ND C)
  with ESPS1.hyps prems show ?case by (auto, simp add: mem_def)  
next
  case (ESPS4 tr t t1 C t2 t3 t4_complete t4_start t5 D NC D' ND)
  hence "(ta, Send (Tx (Honest A) i) \<lbrace>Agent (Honest A),Agent B,Nonce (Honest A) NA\<rbrace> L) \<in> set tr"
    by simp
  with ESPS4.hyps prems show ?case by simp
qed

lemma used_beforeEvent:
  "X \<in> used (beforeEvent ev evs) \<Longrightarrow> X \<in> used evs"
by (erule contrapos_pp, induct evs rule: trace_induct, auto split: event.split_asm)


lemma nonce_send_honest2:
  assumes sps:"tr \<in> ESPS" and
          send: "(t3, Send (Tx (Honest B) i) (Nonce (Honest B) NB)
                           [Agent A, Nonce A' NA, Real t2]) \<in> set tr" 
  shows   "\<exists> t2_start.
             t2 - coffset B \<le> t3 \<and>
             t2_start + min_delay \<le> t2 - coffset B \<and>
             (t2_start, Recv (Rx (Honest B) 0) \<lbrace>Agent A, Agent (Honest B), Nonce A' NA\<rbrace>)
             \<in> set tr \<and>
             Nonce (Honest B) NB
             \<notin> usedI (beforeEvent (t3, Send (Tx (Honest B) i) (Nonce (Honest B) NB) L) tr)"
  using assms
proof (induct rule: ESPS.induct)
  case Nil thus ?case by simp
next
  case (Con tr tc C mc D tab)
  hence "(t3, Send (Tx (Honest B) i) (Nonce (Honest B) NB) [Agent A, Nonce A' NA, Real t2]) \<in> set tr" by simp
  with Con.hyps prems show ?case by force
next
  case (Fake tr mintr I tsend)
  hence "(t3, Send (Tx (Honest B) i) (Nonce (Honest B) NB) [Agent A, Nonce A' NA, Real t2]) \<in> set tr" by simp
  with Fake.hyps prems show ?case by simp
next
  case (ESPS4 tr t t1 C t2' t3' t4_complete t4_start t5 D NC D' ND)
  hence "(t3, Send (Tx (Honest B) i) (Nonce (Honest B) NB) [Agent A, Nonce A' NA, Real t2]) \<in> set tr" by simp
  with ESPS4.hyps prems show ?case by simp
next
  case (ESPS3 tr t t1 D t2' t3' t4_complete t4_start C ND)
  hence "(t3, Send (Tx (Honest B) i) (Nonce (Honest B) NB) [Agent A, Nonce A' NA, Real t2]) \<in> set tr" by simp
  with ESPS3.hyps prems show ?case by simp
next
  case (ESPS1 tr t D ND C)
   hence "(t3, Send (Tx (Honest B) i) (Nonce (Honest B) NB) [Agent A, Nonce A' NA, Real t2]) \<in> set tr" by simp
  with ESPS1.hyps prems show ?case by simp
next
  case (ESPS2 tr tsend B NB trec_complete trec_start A A' NA)
  thus ?case 
    apply -
    apply auto
    apply (force simp add: mem_def) defer defer
    apply (force simp add: mem_def)
    apply (auto simp add: usedI_def mem_def)
    apply (simp add: mem_def [THEN Pure.symmetric, where x1="Nonce (Honest B) NB"])
    prefer 2
    apply (simp add: mem_def [THEN Pure.symmetric, where x1="Nonce (Honest B) NB"])
    apply (rule used_beforeEvent)
    apply force
    apply (rule used_beforeEvent)
    apply force
    done
qed

lemma prefix_ESPS:
  assumes "ys \<in> ESPS" and "ys = x # xs" shows "xs \<in> ESPS" using assms
by (induct "ys" rule: ESPS.induct,auto)

lemma monotone_ESPS:
   assumes "ys \<in> ESPS" and "ys = (tb, evb) # tr" and "(ta, eva) \<in> set tr" shows "ta \<le> tb" using assms
by (induct "ys" rule: ESPS.induct, auto intro: maxtime_geq_elem)


lemma earlier_implies_beforeEvent:
  assumes "tr \<in> ESPS"
  assumes "(ta, eva) \<in> set tr"
  assumes "(tb, evb) \<in> set tr"
  assumes "tb < ta"
  shows   "(tb, evb) \<in> set (beforeEvent (ta, eva) tr)" using assms
  apply -
  apply (induct tr)
  apply force
  apply (frule_tac x=a and xs=tr in prefix_ESPS)
  apply simp
  apply (case_tac "a =  (ta, eva)")
  apply simp
  apply (auto dest: monotone_ESPS)
done

lemma used_beforeEvent_later:
  assumes sps: "tr \<in> ESPS"
  assumes "(ta, Send (Tx A i) ma La) \<in> set tr"
  assumes "x \<notin> used (beforeEvent (ta, Send (Tx A i) ma La) tr)"
  assumes "(tb,  Recv (Rx B j) mb) \<in> set tr"
  assumes "x \<in> subterms {mb}"
  shows   "ta \<le> tb" using assms
  apply -
  apply (frule send_before_recv [simplified abstr_equal [THEN sym]])
  apply simp
  apply (elim exE)
  apply auto
  apply (subgoal_tac "tsend \<le> tb") prefer 2
  apply (erule order_trans)
  apply (subgoal_tac "cdist (Tx Aa ia) (Rx B j) \<ge> 0")
  apply force
  apply (subgoal_tac "y = cdist (Tx Aa ia) (Rx B j)")
  apply (rule cdistnoneg_some)
  apply force
  apply (force simp add: cdist_def)
  apply (rule_tac y=tsend in order_trans)
  prefer 2
  apply assumption
  apply (rotate_tac 2)
  apply (erule contrapos_pp)
  apply (subgoal_tac "tsend < ta") defer
  apply force
  apply (frule_tac ta=ta and tb=tsend in earlier_implies_beforeEvent)
  apply force+
  apply simp
  apply (rule Send_imp_used_parts)
  apply force
done
 

lemma mac_implies:
  assumes sps: "tr \<in> ESPS"
  assumes      "(tsend, Send (Tx (Honest E) i) me Le) \<in> set tr"
  assumes      "(E = A \<or> E = B)"
  assumes      "Hash
                    \<lbrace>Key (symKey (Honest B) (Honest A)),
                     Agent (Honest B), Agent (Honest A),
                     Nonce A' NA, NBB, msg.Real t2B, msg.Real t3B\<rbrace>
                \<in> subterms {me}"
  shows (* step ESPS2 must have happened *)
        "\<exists> trec_start trec_complete NB ts.
           E = B \<and>
           (NBB = Nonce (Honest B) NB) \<and>
           (ts, Send (Tr (Honest B)) (Nonce (Honest B) NB)
                        [Agent (Honest A), Nonce A' NA, Real trec_complete]) \<in> set tr \<and>
           me = Hash[Key (symKey (Honest B) (Honest A))]
                 \<lbrace>Agent (Honest B), Agent (Honest A), Nonce A' NA, Nonce (Honest B) NB,
                  Real (trec_complete),
                  Real (ts + coffset B)\<rbrace>"
  using assms
proof (induct rule: ESPS.induct)
  case Nil thus ?case by simp
next
  case (Con tr tc C mc D tab) thus ?case by simp
next
  case (Fake tr mintr I tsend) thus ?case by simp
next
  case (ESPS1 tr t D ND C)
  have "(tsend, Send (Tx (Honest E) i) me Le) \<in> set tr" using prems by force
  thus ?case using prems by simp
next
  case (ESPS2 tr tsend trec_complete trec_start C D D' ND) thus ?case by force
next
  case (ESPS4 tr t t1 C t2 t3 t4_complete t4_start t5 D NC ND) thus ?case by simp
next
  case (ESPS3 tr ts trec_start D C C' NC tsend' ND trec_complete)
  show ?case proof cases
    assume "(tsend, Send (Tx (Honest E) i) me Le) \<in> set tr"
    thus ?thesis using prems by simp
  next
    assume "(tsend, Send (Tx (Honest E) i) me Le) \<notin> set tr"
    thus ?thesis using prems(3-) by simp
    qed
qed

lemma ESPS_inequalities:
  assumes sps: "tr \<in> ESPS"
  assumes AnotB: "A \<noteq> B"
  assumes claim: "(ta,Claim (Honest A) \<lbrace>Agent (Honest B), Real off\<rbrace>) \<in> set tr"
  shows "\<exists> t1 t2 t3 t4. 
             min_delay \<le> t2 - t1 \<and>
             t2 \<le> t3 \<and>
             min_delay \<le> t4 - t3 \<and>
             off = offset (t1 + coffset A) (t2 + coffset B) (t3 + coffset B) (t4 + coffset A) \<and>
             delay t1 t2 t3 t4 \<le> bound_delay"
proof -
  (* step ESPS3 must have been executed *)
  obtain t1 t2B t3B t4_start t4 NA B' NB t t5 where
    A: "(delay (t1 + coffset A) t2B t3B (t4 + coffset A)) \<le> bound_delay" and
    B: "t4 \<ge> t4_start + min_delay" and
    C: "t \<ge> t4" and
    C2: "off =  offset (t1 + coffset A) t2B t3B (t4 + coffset A)" and
    C3: "(t4_start, Recv (Rx (Honest A) 0) (Nonce B' NB)) \<in> set tr" and
    D: "(t5,
         Recv (Rx (Honest A) 0)
              Hash[Key (symKey (Honest B) (Honest A))]
                  \<lbrace>Agent (Honest B), Agent (Honest A), Nonce (Honest A) NA,
                   Nonce B' NB,
                   Real t2B, Real t3B\<rbrace>)
        \<in> set tr" and
    E: "(t1, Send (Tr (Honest A)) \<lbrace>Agent (Honest A),Agent (Honest B),Nonce (Honest A) NA\<rbrace> [])
        \<in> set tr" using sps claim
    apply -
    apply (drule claim_implies)
    by auto
  let ?mmac = "\<lbrace>Agent (Honest B), Agent (Honest A), Nonce (Honest A) NA, Nonce B' NB, Real t2B, Real t3B\<rbrace>"
  let ?mac  = "Hash[Key (symKey (Honest B) (Honest A))] ?mmac"
  (* Either A or B must have sent something containing the MAC *)
  obtain t5 me i E Le where
    F: "(t5, Send (Tx (Honest E) i) me Le) \<in> set tr" and
    G: "(E = A \<or> E = B)" and
    H: "Hash \<lbrace>Key (symKey (Honest B) (Honest A)), ?mmac\<rbrace> \<in> subterms {me}" and
    I: "Hash \<lbrace>Key (symKey (Honest B) (Honest A)), ?mmac\<rbrace>
        \<notin> used (beforeEvent (t5, Send (Tx (Honest E) i) me Le) tr)"
    using sps D
    apply -
    apply (drule send_before_recv [simplified abstr_equal [THEN sym]])
    apply assumption
    apply (elim exE)
    apply (insert sps)
    apply (drule_tac mc="?mac" in mac_originates [simplified abstr_equal [THEN sym]])
    apply force+
    done
  (* The first send of the MAC must come from step ESPS3 *)
  obtain t2 t3 where
    J: "E = B" and
    K: "(Nonce B' NB = Nonce (Honest B) NB)" and
    L: "(t3, Send (Tr (Honest B)) (Nonce (Honest B) NB)
                  [Agent (Honest A), Nonce (Honest A) NA, Real t2]) \<in> set tr" and
(*    M: "(t2_start, Recv (Rx (Honest B) 0) \<lbrace>Agent (Honest A), Agent (Honest B), Nonce (Honest A) NA\<rbrace>) \<in> set tr" and *)
    N: "me = Hash[Key (symKey (Honest B) (Honest A))]
                 \<lbrace>Agent (Honest B), Agent (Honest A), Nonce (Honest A) NA, Nonce (Honest B) NB,
                  Real (t2),
                  Real (t3 + coffset B)\<rbrace>" using sps F G H
    apply -
    apply (drule mac_implies)
    apply assumption+
    apply force
    done
  obtain t2_start where
      O: "t2 - coffset B \<le> t3" and
      P: "t2_start + min_delay \<le> t2 - coffset B" and
      Q: "(t2_start, Recv (Rx (Honest B) 0) \<lbrace>Agent (Honest A), Agent (Honest B), Nonce (Honest A) NA\<rbrace>)
          \<in> set tr" and
      R: "Nonce (Honest B) NB
          \<notin> usedI (beforeEvent (t3, Send (Tx (Honest B) 0) (Nonce (Honest B) NB) [Agent (Honest A), Nonce (Honest A) NA, Real t2]) tr)"
    using sps prems
    apply -
    apply (drule nonce_send_honest2)
    by auto
  have S: "t1 \<le> t2_start" using prems
    apply -
    apply (frule_tac A="Honest A" and B="Honest B" and NA=NA and
                     mb="\<lbrace>Agent (Honest A),Agent (Honest B),Nonce (Honest A) NA\<rbrace>" and
                     ma = "\<lbrace>Agent (Honest A),Agent (Honest B),Nonce (Honest A) NA\<rbrace>"
           in fresh_nonce_earliest_recv [simplified abstr_equal [THEN sym]]) defer
    apply force+
    apply (insert cdistl_noneg [where A="Honest A" and B="Honest B"], force)
    apply (drule nonce_send_honest)
    apply (unfold usedI_def)
    apply force+
    done    
    (* This follows from the freshness of the Nonce NB *)
  
  hence T: "t3 \<le> t4_start" using L C3 K J sps AnotB R
    apply -
    apply simp
    apply (frule_tac A="Honest B" and B="Honest A" and NA=NB and
                     mb="Nonce (Honest B) NB" and
                     ma = "Nonce (Honest B) NB"
           in fresh_nonce_earliest_recv [simplified abstr_equal [THEN sym]]) defer
    apply force+
    apply (insert cdistl_noneg [where A="Honest B" and B="Honest A"], force)
    apply (unfold usedI_def)
    apply simp
    done
  have U: "t3B = t3 + coffset B" using N H  by simp
  have V: "t2B = t2" using N H by simp
  (* The recv in ESPS2 must have happened after the challenge has been sent in step ESPS1 *)
  thus ?thesis using Q R O P C2 A N S T B R U V
    apply -
    apply (rule_tac x=t1 in exI)
    apply (rule_tac x="t2 - coffset B" in exI)
    apply (rule_tac x=t3 in exI)
    apply (rule_tac x=t4 in exI)
    apply (auto simp add: delay_def)
    done
qed

lemma ESPS_secure:
  assumes sps: "tr \<in> ESPS"
  assumes AnotB: "A \<noteq> B"
  assumes claim: "(ta,Claim (Honest A) \<lbrace>Agent (Honest B), Real off\<rbrace>) \<in> set tr"
  shows "abs (off - coffset B + coffset A) \<le> 2* bound_delay - 2* min_delay"
proof -
  obtain t1 t2 t3 t4 where
    A: "min_delay \<le> t2 - t1" and
    B: "t2 \<le> t3" and
    C: "min_delay \<le> t4 - t3" and
    D: "off = offset (t1 + coffset A) (t2 + coffset B) (t3 + coffset B) (t4 + coffset A)" and
    E: "delay t1 t2 t3 t4 \<le> bound_delay" using sps claim AnotB
    apply -
    apply (drule ESPS_inequalities)
    apply force+
    done
  then have F: "off - coffset B + coffset A = offset t1 t2 t3 t4"
    by (auto simp add: offset_def)
  have G: "t2 - t1 \<le> 2 * bound_delay - min_delay" using A B C E
    by (auto simp add: delay_def)
  have H: "t4 - t3 \<le> 2 * bound_delay - min_delay" using A B C E
    by (auto simp add: delay_def)
  have I: "abs (offset t1 t2 t3 t4) \<le> 2 * bound_delay - 2 * min_delay" using A B G H
  proof cases
    assume "t2 - t1  \<ge> t4 - t3"
    hence "offset t1 t2 t3 t4 \<ge> 0" by (auto simp add: offset_def)
    thus ?thesis using A B C G H apply -
      by (auto simp add: abs_def offset_def)
  next
    assume "\<not>(t2 - t1 \<ge> t4 - t3)"
    hence  "t2 - t1 < t4 - t3" by simp
    hence "offset t1 t2 t3 t4 < 0" by (simp add: offset_def)
    hence "abs (offset t1 t2 t3 t4) = ((t4 - t3) - (t2 - t1)) / 2" apply (auto simp add: abs_def offset_def)
      apply (subgoal_tac "- ((t2 * 2 - t1 * 2 - (t4 * 2 - t3 * 2)) / 2) = - (t2 - t1 - (t4 - t3))")
      prefer 2
      apply (rule minus_eq)
      by auto
    thus ?thesis using A B C G H by simp 
  qed
  thus ?thesis using F G by simp  
qed

end
