(******************************************************************************

  ProtoVeriPhy

  Module: distance/UsoundDistanceBounding.thy

  Copyright (C) 2007-12 Benedikt Schmidt
  License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Security Analysis of an Ultrasound Distance Bounding Protocol *}

theory UsoundDistanceBounding imports SystemCoffset SystemOrigination MessageTheoryFree begin

subsection {* Direct and Parameterized Definition and their Equality *}

locale INITSTATE_SIG_NN = INITSTATE_PKSIG + INITSTATE_NONONCE

text {* here we define the initstate for the protocol, then interpret it as \_SIG\_NN *}

definition
  initStateUdb :: "agent \<Rightarrow> msg set" where
  "initStateUdb A == Key`(insert (priSK A) (pubSK`UNIV))"

interpretation INITSTATE_SIG_NN Crypt Nonce MPair Hash Number parts subterms DM initStateUdb Key
apply (unfold_locales)
apply (auto simp add: initStateUdb_def)
done

definition
  udb1 :: "msg step"
 where
  "udb1 tr A t =
     (UN NA.
         {ev. ev = (Nonce (Honest A) NA,SendEv 0 []) \<and>

              Nonce (Honest A) NA \<notin> usedI tr})"

definition
  udb2 :: "msg step"
 where
  "udb2 tr B t =
      (UN trec NA C.
         {ev. ev = (Crypt (priSK (Honest B))
                          (Nonce C NA),SendEv 1 []) \<and>

               (trec, Recv (Rec (Honest B)) (Nonce C NA)) \<in> set tr})"

definition
  udb3 :: "msg step"
 where
  "udb3 tr A tdone =
      (UN tsend trec NA B'.
        {ev. ev = (\<lbrace>Agent B', Real ((trec - tsend) * vu)\<rbrace>,ClaimEv) \<and>

           (trec,
            Recv (Ru (Honest A)) (Crypt (priSK B') (Nonce (Honest A) NA)))
            \<in> set tr \<and>
           (tsend, Send (Tr (Honest A)) (Nonce (Honest A) NA) []) \<in> set tr})"

definition
  udb_proto :: "msg proto" where
  "udb_proto \<equiv> {udb1, udb2, udb3}"

text {* now we interpret the protocol as PKSIGNOKEYS and later as EXECUTABLEPLUSDELTAONLY*}

locale PROTOCOL_UDB1 =
  PROTOCOL_PKSIG_NOKEYS+PROTOCOL_NONONCE+INITSTATE_SIG_NN

lemmas udb_defs = udb_proto_def udb1_def udb2_def udb3_def

interpretation PROTOCOL_UDB1 Crypt Nonce MPair Hash Number parts subterms DM initStateUdb Key udb_proto
  apply (unfold_locales)
  apply (auto simp add: udb_defs messagesProto_def messagesProtoTrHonest_def
              split: event.split split_if dest: parts.fst_set)
done

interpretation PROTOCOL_EXECUTABLE  Crypt Nonce MPair Hash Number parts subterms DM initStateUdb udb_proto sys Key
  apply (unfold_locales)
  apply (auto simp add: udb_defs initStateUdb_def DM_analz_synth
                        messagesProto_def messagesProtoTrHonest_def
              split: event.split split_if)
  apply (drule view_elem_ex, erule exE, drule Recv_imp_knows_A)
  apply (auto intro!: synth.Crypt)
  apply (auto simp add: nonce_view_fresh [simplified udb_proto_def]
                        nonce_view_used [simplified udb_proto_def])
  apply (auto simp add: recv_a_view_a_r send_a_view_a_r
                        recv_a_view_a_u [simplified] send_a_view_a_u [simplified])
done

interpretation PROTOCOL_DELTAONLY Crypt Nonce MPair Hash Number parts subterms DM initStateUdb Key udb_proto
  apply unfold_locales
  apply (auto simp add: udb_defs in_timetrans)
  apply force prefer 2
  apply (rule_tac x="trec + coffset A" in exI, force)
  apply (intro exI conjI) prefer 2
  apply force prefer 2
  apply force apply force
  apply (rule_tac x="tsend + coffset A" in exI,
         rule_tac x="trec + coffset A" in exI)
  apply auto
done

interpretation PROTOCOL_DELTA_EXEC Crypt Nonce MPair Hash Number parts subterms DM initStateUdb Key udb_proto sys
by unfold_locales

inductive_set
  udb :: "(msg trace) set"
 where
   Nil [intro] : "[] \<in> udb"
 | Fake: 
   "\<lbrakk> tr \<in> udb;  t >=  maxtime tr;
      X \<in> DM (Intruder I) (knowsI (Intruder I) tr)\<rbrakk>
   \<Longrightarrow> (t, Send (Tx (Intruder I) j) X [])  # tr \<in> udb"

 | Con : 
   "\<lbrakk> tr \<in> udb; (tsend, Send (Tx A i) m L) \<in> set tr;
       cdistM (Tx A i) (Rx B j) = Some tab;
       trecv >= maxtime tr; trecv \<ge> tsend + tab  \<rbrakk>
    \<Longrightarrow> (trecv, Recv (Rx B j) m) # tr \<in> udb"

 (* Protocol rules for integrity regions *)
 | UDB1:  
   "\<lbrakk> tr \<in> udb; t >=  maxtime tr;
      \<not> (usedI tr (Nonce (Honest A) NA))  \<rbrakk> (* infer_arities: bad term when using set notation *)
   \<Longrightarrow> (t, Send (Tx (Honest A) 0) (Nonce (Honest A) NA) []) # tr \<in> udb"

 | UDB2:  
   "\<lbrakk> tr \<in> udb; tsend >=  maxtime tr;
      (trec, Recv (Rec (Honest B)) (Nonce A NA)) \<in> set tr \<rbrakk>
   \<Longrightarrow> (tsend,
        Send (Tu (Honest B)) (Crypt (priSK (Honest B)) (Nonce A NA)) [])
       # tr \<in> udb"

 | UDB3:  
   "\<lbrakk> tr \<in> udb; tdone \<ge> maxtime tr;
      (trec, Recv (Ru (Honest A)) (Crypt (priSK B) (Nonce (Honest A) NA))) \<in> set tr;
      (tsend, Send (Tr (Honest A)) (Nonce (Honest A) NA) []) \<in> set tr \<rbrakk>
  \<Longrightarrow> (tdone, Claim (Honest A) \<lbrace>Agent B, Real ((trec - tsend) * vu)\<rbrace>) # tr
      \<in> udb"

text {* obtain a simpler induction rule for protocol since it is executable and deltaonly *}

lemmas proto_induct =
  sys.induct [simplified derivable_removable remove_occursAt timetrans_removable]

text {* now we show that both define the same set of traces *}

lemma abstr_equal: "udb = sys"
proof auto
  fix tr
  assume r: "tr \<in> sys"
  show "tr \<in> udb" using r
  proof(induct tr rule: proto_induct)
    case 1 show ?case using r by auto
  next
    case 2
    with prems show ?case by auto
  next
    case (3 tr t X I j)
    with prems show ?case by (auto intro: udb.Fake)
  next
    case (4 tr ts X I j)
    with prems show ?case
      apply - apply (rule udb.Con) by auto
  next
    case (5 tr step m t pEv A)
    thus ?case
    apply (auto simp add: udb_defs)
    apply (auto intro: udb.UDB1 udb.UDB3 udb.UDB2 [simplified])
    apply (rule udb.UDB1)
    apply auto
    apply (auto simp add: mem_def)
    done
  qed
next
  fix tr
  assume r: "tr \<in> udb"
  show "tr \<in> sys" using r
  proof(induct tr rule: udb.induct)
    case Nil
    with prems show ?case by auto
  next
    case (Fake tr ts X I j)
    with prems show ?case by (auto intro: sys.Fake)
  next
    case (Con tr tsend C i X D j dCD trecv)
    with prems show ?case by (auto intro: sys.Con)
  next
    case (UDB1 tr ts C NA)
    with prems have "(ts,
                      createEv C (SendEv 0 []) (Nonce (Honest C) NA)) # tr \<in> sys"
      apply -
      apply (rule_tac step=udb1 in sys_Proto) apply (auto simp add: udb_defs)
      apply (auto simp add: mem_def)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (UDB2 tr tsend trecv B A NA)
    with prems have
      "(tsend, createEv B
                 (SendEv 1 [])
                 (Crypt (priSK (Honest B)) (Nonce A NA))) # tr \<in> sys"
      apply - apply (rule_tac step=udb2 in sys_Proto)
      apply (auto simp add: udb_defs)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (UDB3 tr tdone trecv A B NA tsend)
    with prems have
      "(tdone,createEv A ClaimEv \<lbrace>Agent B, Real ((trecv - tsend)*vu)\<rbrace>) # tr \<in> sys"
      apply - apply (rule_tac step=udb3 in sys_Proto) apply (auto simp add: udb_defs)
      done
   thus ?case by (auto simp add: createEv.psimps)
  qed
qed

lemmas [simp,intro] = abstr_equal [THEN sym]

subsection {*A "possibility property": there are traces that reach the end*}

lemma udb_possibility:
 assumes distab: "cdistM (Tr (Honest A)) (Rec (Honest B)) \<noteq> None"
 assumes distba: "cdistM (Tu (Honest B)) (Ru (Honest A)) \<noteq> None"
 shows "\<exists>ta dt. \<exists>tr \<in> udb. (ta,Claim (Honest A) \<lbrace>Agent (Honest B), Real dt\<rbrace>)
        \<in> set tr"
proof -
  let ?dab = "cdist (Tr (Honest A)) (Rec (Honest B))" and
      ?dba = "cdist (Tu (Honest B)) (Ru (Honest A))" and
      ?fa = "Honest A" and
      ?fb = "Honest B" and
      ?na = "Nonce (Honest A) NA"
  let ?msg2 = "Crypt (priSK ?fb) ?na" and
      ?done = "Claim ?fa \<lbrace>Agent ?fb, Real (((?dab+?dba) - 0)*vu)\<rbrace>"
  from distab have dabnoneg: "?dab \<ge> 0"
    by (auto simp add: cdist_def intro: cdistnoneg_some)
  from distba have dbanoneg: "?dba \<ge> 0"
    by (auto simp add: cdist_def intro: cdistnoneg_some)
  from dabnoneg dbanoneg have snoneg: "?dab+?dba \<ge> 0" by auto
  have "(0,Send (Tr ?fa) ?na [])#[] \<in> udb" (is "?tr1 \<in> udb")
    apply - apply (rule udb.UDB1) apply auto
    apply (auto simp add: usedI_def initStateUdb_def)
    apply (simp only: mem_def [of "Nonce (Honest A) NA", THEN Pure.symmetric])
    apply auto
    done

  with distab dabnoneg have "(0+?dab,Recv (Rec ?fb) ?na)#?tr1 \<in> udb"
     (is "?tr2 \<in> udb")
     apply -
     apply (rule_tac tab="cdist (Tr (Honest A)) (Rec (Honest B))" in udb.Con)
     by (auto simp add: cdist_def)
  with dabnoneg have "(?dab,Send (Tu ?fb) ?msg2 [])#?tr2 \<in> udb"
    (is "?tr3 \<in> udb") apply -
    by (rule_tac tr="?tr2" in udb.UDB2, auto)
  hence "(?dab+?dba,Recv (Ru ?fa) ?msg2)#?tr3 \<in> udb"
    (is "?tr4 \<in> udb")
    using snoneg distba dbanoneg apply - apply (simp) apply (rule udb.Con)
    by (auto simp add: cdist_def)
  hence "(?dab+?dba, ?done)#?tr4 \<in> udb" (is "?tr5 \<in> udb")
    apply (rule_tac tr="?tr4" in udb.UDB3) apply auto
    apply (rule dbanoneg [simplified]) apply (rule snoneg [simplified]) done
  -- {* This is the witness trace: @{term [display] ?tr5} *}
  thus ?thesis by (intro exI bexI, auto)
qed

lemma honest_send_no_other_sig:
  assumes rang: "tr \<in> udb" and
          A: "(tb, Send (Tx (Honest B) i) (Crypt (priSK B') (Nonce A NA)) L) \<in> set tr"
  shows   "B'=Honest B"
  using rang A
by (induct tr, auto)

lemma recv_before_send_sig:
  assumes rang: "tr \<in> udb" and
          A: "(tb, Send (Tx (Honest B) i) mb L) \<in> set tr" and
          B: "Crypt (priSK (Honest Ba)) (Nonce A NA)
              \<in> subterms {mb}"
  shows   "\<exists> trec. (trec, Recv (Rx (Honest B) 0) (Nonce A NA)) \<in> set tr \<and> trec \<le> tb"
  using rang A B
apply (induct tr, auto)
apply (rule_tac x=trec in exI, auto)
apply (rule maxtime_geq_elem)
by auto

lemma nonce_send_honest:
  assumes send: "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) L) \<in> set tr" and
          rang: "tr \<in> udb"
  shows   "Nonce (Honest A) NA
           \<notin> usedI (beforeEvent (ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) L) tr)"
  using rang send
proof (induct rule: udb.induct)
next
  case Nil thus ?case by auto
next
  case (Con tr tsend)
  hence "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) L) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
next
  case (Fake tr mintr I tsend)
  hence "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) L) \<in> set tr" by auto
  with Fake.hyps prems show ?case by (auto)
next
  case (UDB2 tr trecv D NB tsend)
  hence "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) L) \<in> set tr" by auto
  with UDB2.hyps prems show ?case by (auto)
next
  case (UDB3 tr trecv D E NB tsend)
  hence "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) L) \<in> set tr" by auto
  with UDB3.hyps prems show ?case by (auto)
next
  case (UDB1 tr NB tsend D)
  with UDB1.hyps prems show ?case apply auto
    apply (auto simp add: mem_def)
    done
qed

lemma sequence_of_steps:
  assumes trace:"tr \<in> udb" and
          send:"(tsend, Send (Tx (Honest A) 0) (Nonce (Honest A) NA) []) \<in> set tr" and
          recv:"(trec, Recv (Ru (Honest A)) (Crypt (priSK B) (Nonce (Honest A) NA))) \<in> set tr"
  shows "tsend \<le> trec" using prems
proof (induct rule:udb.induct)
 case Nil thus ?case by simp
next
  case (Con tr2 tsend2 A2 i2 m2 L2 B2 j2 tabd trecv2) with Con.hyps show ?case 
    apply -
    apply clarsimp
    apply (case_tac "(trecv2,
      Recv (Rx (Honest A) (Suc 0))
       (Crypt (invKey (pubSK B)) (Nonce (Honest A) NA)))
     \<in> set tr2")
    prefer 2
    apply (erule disjE)
    apply simp
    apply auto
    apply (rule maxtime_geq_elem)
    apply auto
 done
next
  case Fake thus ?case by simp
next
 case UDB2 thus ?case by simp
next 
 case UDB3 thus ?case by simp
next
 case (UDB1 tr2 ts A2 NA2) with UDB1.hyps  show ?thesis 
   apply -
   apply auto
   apply (subgoal_tac "\<exists>C i tsend3 L.
      (tsend3, Send (Tx C i)  ((Crypt (priSK B) (Nonce (Honest A2) NA2))) L) \<in> set tr2 \<and>
      cdist (Tx C i) (Ru (Honest A2))\<le> trec \<and>
      cdistM (Tx C i) (Ru (Honest A2))\<noteq> None \<and> tsend3 \<le> trec - cdist (Tx C i) (Ru (Honest A2))")
   prefer 2
   apply (rule send_before_recv)
   apply auto
   apply (drule_tac t="tsend3" and Y="Nonce (Honest A2) NA2" in Send_imp_parts_used)
   apply auto
   apply (simp add: usedI_def)
   apply (simp add: mem_def Un_def Collect_def)
done
qed



lemma claim_trace_measurement_posh:
  assumes rang: "tr \<in> udb" and
          claim:"(tdone, Claim (Honest A) \<lbrace>Agent B, Real ((trec - tsend) * vu)\<rbrace>) \<in> set tr"
  shows "tsend \<le> trec"
  using rang claim
proof (induct rule: udb.induct)
  case Nil thus ?case by simp
next
  case Con thus ?case by simp
next 
  case Fake thus ?case by simp
next
  case UDB1 thus ?case by simp
next
  case UDB2 thus ?case by simp
next 
  case (UDB3 trn tdonen trecn An Bn NAn tsendn)  with UDB3.hyps show ?case
    apply -
    apply (case_tac "(tdone, Claim (Honest A) \<lbrace>Agent B, msg.Real ((trec - tsend) * vu)\<rbrace>) \<in> set trn")
    apply assumption
    apply clarsimp
    apply (erule disjE)
    apply (insert vu_pos)
    apply simp
    apply (drule_tac  sequence_of_steps)
    apply simp+
done
qed

lemma tracetime_non_negative_udb: 
    assumes "tr \<in> udb" and "(t,ev) \<in> set tr" shows "0 \<le> t"
using prems
apply -
apply (rule tracetime_non_negative)
apply (insert abstr_equal)
apply auto
done


lemma claim_trace_measurement_pos:
  assumes udb: "tr \<in> udb" and 
  believe: "(tdone, Claim (Honest A) \<lbrace>Agent (B), Real d\<rbrace>) \<in> set tr"
  shows   "d \<ge> 0" using udb believe
proof (induct rule: udb.induct)
  case Nil thus ?case by simp
next
  case Con thus ?case by simp
next 
  case Fake thus ?case by simp
next
  case UDB1 thus ?case by simp
next
  case UDB2 thus ?case by simp
next
  case (UDB3 trn tdonen trecn AN Bn NAn tsendn) with UDB3.hyps show ?case
    apply -
    apply (frule udb.UDB3)
    apply simp+
    apply (erule disjE)
    apply simp
    prefer 2
    apply assumption
    apply (frule_tac tr="(tdonen, Claim (Honest AN) \<lbrace>Agent Bn, msg.Real ((trecn - tsendn) * vu)\<rbrace>) # trn" and tdone="tdonen" and tsend="tsendn" and trec="trecn" and A="AN" and B="Bn" in claim_trace_measurement_posh)
    apply simp
    apply (insert vu_pos)
    apply (subgoal_tac "trecn - tsendn >= 0")
    prefer 2 
    apply arith
    apply (unfold le_less)
    apply (rotate_tac -1)
    apply (erule disjE)
    apply (rule disjI1)
    apply (rule mult_pos_pos)
    apply assumption
    apply assumption
    apply (rule disjI2)
    apply simp
  done
qed

fun
  usound_receiver :: "receiver \<Rightarrow> bool" where
  "usound_receiver (Rx B j) = 
     (\<forall>A i. cdistM (Tx A i) (Rx B j) \<noteq> None \<longrightarrow>
          vu * (the (cdistM (Tx A i) (Rx B j))) \<ge> pdist A B)"

lemma us_speed_some:
  assumes us:   "usound_receiver (Rx B j)" and
          some: "cdistM (Tx A i) (Rx B j) = Some y"
  shows "vu * y \<ge>  pdist A B" using some
proof -
  have "cdistM (Tx A i) (Rx B j) \<noteq> None \<longrightarrow>
         vu * (the (cdistM (Tx A i) (Rx B j))) \<ge> pdist A B" using us by auto
  with some have "vu * the (cdistM (Tx A i) (Rx B j)) \<ge>  pdist A B" by auto
  with some show ?thesis by auto
qed

subsection {* A Relay Attack is Captured by the Model *}

text {* The relay attack is captured in our model if two intruders
        with the right communication abilities are positioned right. *}
lemma udb_attack_relay:
 assumes notregion: "cdist (Tu (Honest B)) (Ru (Intruder C)) +
                     cdist (Tu (Intruder D)) (Ru (Honest A)) +
                     cdist (Tr (Intruder C)) (Rec (Intruder D)) +
                     cdist (Tr (Honest A)) (Rec (Honest B))
                     <  (pdist (Honest A) (Honest B)) / vu"
 assumes distAB:   "cdistM (Tr (Honest A)) (Rec (Honest B)) \<noteq> None"
 assumes distBIc:  "cdistM (Tu (Honest B)) (Ru (Intruder C)) \<noteq> None"
 assumes distIcId: "cdistM (Tr (Intruder C)) (Rec (Intruder D)) \<noteq> None"
 assumes distIdA:  "cdistM (Tu (Intruder D)) (Ru (Honest A)) \<noteq> None"
 shows "\<exists>ta dt. \<exists>tr \<in> udb. (ta,Claim (Honest A) \<lbrace>Agent (Honest B), Real dt\<rbrace>)
        \<in> set tr \<and> dt < pdist (Honest A) (Honest B)"
proof -
  let ?dAB   = "cdist (Tr (Honest A)) (Rec (Honest B))" and
      ?dIcId = "cdist (Tr (Intruder C)) (Rec (Intruder D))" and
      ?dBIc  = "cdist (Tu (Honest B)) (Ru (Intruder C))" and
      ?dIdA  = "cdist (Tu (Intruder D)) (Ru (Honest A))" and
      ?fa    = "Honest A" and
      ?fb    = "Honest B" and
      ?ic    = "Intruder C" and
      ?id    = "Intruder D" and
      ?na    = "Nonce (Honest A) NA"
  let ?msg2  = "Crypt (priSK ?fb) ?na" and
      ?done  = "Claim ?fa \<lbrace>Agent ?fb, Real ((?dAB+?dBIc+?dIcId+?dIdA - 0)
                                           *vu)\<rbrace>"
  from distAB   have dABnoneg: "?dAB \<ge> 0"
    by (auto simp add: cdist_def intro: cdistnoneg_some)
  from distIcId have dIcIdnoneg: "?dIcId \<ge> 0"
    by (auto simp add: cdist_def intro: cdistnoneg_some)
  from distBIc  have dBIcnoneg: "?dBIc \<ge> 0"
    by (auto simp add: cdist_def intro: cdistnoneg_some)
  from distIdA  have dIdAnoneg: "?dIdA \<ge> 0"
    by (auto simp add: cdist_def intro: cdistnoneg_some)

  with distBIc distAB dABnoneg dBIcnoneg have dAIcnoneg: "?dAB+?dBIc \<ge> 0" by auto
  with distBIc distAB dABnoneg dBIcnoneg dIcIdnoneg dAIcnoneg
    have dAIdnoneg: "?dAB+?dBIc+?dIcId \<ge> 0" by auto
  with distBIc distAB dABnoneg dBIcnoneg dIcIdnoneg dAIcnoneg dAIdnoneg distIdA dIdAnoneg
    have dAAnoneg: "?dAB+?dBIc+?dIcId+?dIdA \<ge> 0" by auto
  with distBIc distAB dABnoneg dBIcnoneg dIcIdnoneg dAIcnoneg dAIdnoneg distIdA dIdAnoneg
    have dIcAnoneg: "?dIcId+?dIdA \<ge> 0" by auto
  with distBIc distAB dABnoneg dBIcnoneg dIcIdnoneg dAIcnoneg dAIdnoneg distIdA dIdAnoneg
    have dBAnoneg: "?dBIc+(?dIcId+?dIdA) \<ge> 0" by auto
  with distBIc distAB dABnoneg dBIcnoneg dIcIdnoneg dAIcnoneg dAIdnoneg distIdA dIdAnoneg
    have dAAnoneg2: "?dAB+(?dBIc+(?dIcId+?dIdA)) \<ge> 0" by auto

  have "(0,Send (Tr ?fa) ?na [])#[] \<in> udb" (is "?tr1 \<in> udb")
    apply - apply (rule udb.UDB1) apply auto apply (simp add: usedI_def)
      apply (simp add: mem_def [of "Nonce (Honest A) NA", THEN Pure.symmetric])
      done
  with distAB dABnoneg have "(0+?dAB,Recv (Rec ?fb) ?na)#?tr1 \<in> udb" (is "?tr2 \<in> udb")
     apply -
     apply (rule_tac tab="cdist (Tr (Honest A)) (Rec (Honest B))" in udb.Con)
     by (auto simp add: cdist_def)
  with dABnoneg have "(?dAB,Send (Tu ?fb) ?msg2 [])#?tr2 \<in> udb"
    (is "?tr3 \<in> udb") apply -
    by (rule_tac tr="?tr2" in udb.UDB2, auto simp add: cdist_def)
  hence "(?dAB+?dBIc,Recv (Ru ?ic) ?msg2)#?tr3 \<in> udb"
    (is "?tr4 \<in> udb")
    using distBIc dAIcnoneg dBIcnoneg
     apply - apply (simp) apply (rule udb.Con [simplified]) by (auto simp add: cdist_def)
  hence "(?dAB+?dBIc,Send (Tr ?ic) ?msg2 [])#?tr4 \<in> udb"
    (is "?tr5 \<in> udb")
    using distBIc dAIcnoneg dBIcnoneg
     apply - apply (simp) apply (rule udb.Fake [simplified]) apply auto
     apply (auto simp add: DM_analz_synth knowsI_def)
     done
  hence "(?dAB+?dBIc+?dIcId,Recv (Rec ?id) ?msg2)#?tr5 \<in> udb"
    (is "?tr6 \<in> udb")
    using distBIc dAIcnoneg dBIcnoneg dAIdnoneg distIcId dIcIdnoneg
     apply - apply (simp) apply (rule udb.Con [simplified])
     by (auto simp add: cdist_def)
  hence "(?dAB+?dBIc+?dIcId,Send (Tu ?id) ?msg2 [])#?tr6 \<in> udb"
    (is "?tr7 \<in> udb")
    using distBIc dAIcnoneg dBIcnoneg dAIdnoneg distIcId dIcIdnoneg
     apply - apply (simp) apply (rule udb.Fake [simplified])
     by (auto simp add: DM_analz_synth knowsI_def)
  hence "(?dAB+?dBIc+?dIcId+?dIdA,Recv (Ru ?fa) ?msg2)#?tr7 \<in> udb"
    (is "?tr8 \<in> udb")
    using distBIc dAIcnoneg dBIcnoneg dAAnoneg distIcId dIcIdnoneg dIdAnoneg distIdA
     apply - apply (simp) apply (rule_tac A="?id" in udb.Con)
     by (auto simp add: cdist_def)
  hence "(?dAB+?dBIc+?dIcId+?dIdA, ?done)#?tr8 \<in> udb" (is "?tr9 \<in> udb")
    apply (rule_tac udb.UDB3) apply auto
    apply (rule dIdAnoneg [simplified]) apply (rule dIcAnoneg [simplified])
    apply (rule dBAnoneg [simplified]) apply (rule dAAnoneg [simplified]) done
  -- {* This is the witness trace: @{term [display] ?tr5} *}
  thus ?thesis using notregion vu_pos vc_pos apply (intro exI bexI) apply auto
  apply (rule_tac c="1/vu" in mult_less_imp_less_right)
  by auto
qed

subsection {* Security Proof with Private Space Assumption *}

lemma udb_secure_private_space:
  assumes
  udb: "tr \<in> udb" and
  us_recv: "usound_receiver (Ru (Honest A))" and
  region: "!!I.  pdist (Intruder I) (Honest A) \<ge> pdist B (Honest A)" and
  believe: "(tdone, Claim (Honest A) \<lbrace>Agent B, Real d\<rbrace>) \<in> set tr"
  shows   "d \<ge> pdist (Honest A) B"
  using prems
proof (cases "Honest A=B")
  case True
  thus ?thesis using prems
    apply (simp add: pdist_equal_zero)
    apply (erule claim_trace_measurement_pos)
    apply auto
    done
next
case False
show ?thesis using prems
proof (induct tr arbitrary: A B trec t rule: udb.induct)
  case (UDB3 tr t trecv D E NA tsend)
  let ?x = "(t, Claim (Honest D) \<lbrace>Agent E, Real ((trecv - tsend)*vu)\<rbrace>)" and
      ?ev = "((tdone, Claim (Honest A) \<lbrace>Agent B, Real d\<rbrace>))"
  show ?case
  proof cases
    assume "?x = ?ev"
    hence "A=D" and "E=B" and "t=tdone" and tdelta:"d=(trecv - tsend)*vu" by auto
    let ?sigmsg =  "Crypt (priSK E) (Nonce (Honest A) NA)"
    from `A=D` `E=B` prems have
      ononce: "(tsend, Send (Tr (Honest A)) (Nonce (Honest A) NA) []) \<in> set tr" and
      sig: "(trecv, Recv (Ru (Honest A)) ?sigmsg) \<in> set tr"
    by auto
    with prems have
          ex: "\<exists>C i tcsend Lc. (tcsend, Send (Tx C i) ?sigmsg Lc) \<in> set tr
               \<and> trecv >= cdist (Tx C i) (Ru (Honest A))
               \<and> cdistM (Tx C i) (Ru (Honest A)) \<noteq> None
               \<and> tcsend \<le> trecv - cdist (Tx C i) (Ru (Honest A))"
      apply - apply (rule send_before_recv)
      by (auto)
    from ex obtain C tcsend i Lc
      where sigintrC: "(tcsend, Send (Tx C i) ?sigmsg Lc) \<in> set tr" and
                      "trecv >= cdist (Tx C i) (Ru (Honest A))" and
                      "cdistM (Tx C i) (Ru (Honest A)) \<noteq> None" and
            q: "tcsend \<le> trecv - cdist (Tx C i) (Ru (Honest A))"
    by auto
    (* show that the send must have occured after the sending of the challenge nonce if A \<noteq> C *)
    have tcs: "Honest A \<noteq> C \<Longrightarrow> tsend \<le> tcsend" proof - 
      assume "Honest A \<noteq> C"
      hence "cdistl (Honest A) C \<le> tcsend - tsend" using sigintrC ononce apply -
	apply (rule fresh_nonce_earliest_send)
	apply (auto intro: prems)
	apply (drule nonce_send_honest)
	by (auto simp add: usedI_def intro: prems)
      thus ?thesis using cdistl_noneg[where A="Honest A" and B=C] by auto
    qed
    show ?thesis
    proof cases
      (* an honest agent must have sent the reponse for himself *)
      assume "\<exists> D. C = Honest D"
      then obtain D where D: "C = Honest D" by auto
      hence BD: "B = Honest D" using `(tcsend, Send (Tx C i) ?sigmsg Lc) \<in> set tr` prems apply -
	apply (rule honest_send_no_other_sig) by auto
      have "Honest A \<noteq> C" using prems BD D by auto
      hence "tsend \<le>  trecv - cdist (Tx B i) (Ru (Honest A))" using tcs q BD D by auto
      hence p4: "cdist (Tx B i) (Ru (Honest A)) \<le> trecv - tsend" by auto
      have "pdist B (Honest A) \<le> vu * cdist (Tx C i) (Ru (Honest A))"
        using `cdistM (Tx C i) (Ru (Honest A)) \<noteq> None`
              `usound_receiver (Ru (Honest A))`
	      BD D
        apply -
        apply (rule_tac j=1 in us_speed_some)
	by (auto simp add: cdist_def)
      also have "... \<le> vu * (trecv - tsend)" using p4 apply -
        apply (rule mult_mono1)
        by (insert vu_pos, auto simp add: BD D)
      finally show ?thesis using tdelta by (auto simp add: real_mult_commute pdist_symm)
    next
      assume "\<not> (\<exists> D. C = Honest D)"
      hence "\<exists> D. C = Intruder D" apply (case_tac C) by auto
      then obtain D where D: "C = Intruder D" by auto
      have "Honest A \<noteq> C" using prems D by auto
      hence "tsend \<le>  trecv - cdist (Tx (Intruder D) i) (Ru (Honest A))" using tcs q D by auto
      hence p4: "cdist (Tx (Intruder D) i) (Ru (Honest A)) \<le> trecv - tsend" by auto
      have "pdist B (Honest A) \<le> pdist (Intruder D) (Honest A)" using prems by auto
      also have "... \<le> vu * cdist (Tx C i) (Ru (Honest A))"
        using `cdistM (Tx C i) (Ru (Honest A)) \<noteq> None`
              `usound_receiver (Ru (Honest A))` D
        apply -
        apply (rule_tac j=1 in us_speed_some)
	by (auto simp add: cdist_def)
      also have "... \<le> vu * (trecv - tsend)" using p4 apply -
        apply (rule mult_mono1)
        by (insert vu_pos, auto simp add: D)
      finally show ?thesis using tdelta by (auto simp add: real_mult_commute pdist_symm)
    qed
  next
    assume "?x \<noteq> ?ev"
    with `?ev \<in> set (?x#tr)` have "?ev \<in> set tr" by auto
    with prems show ?thesis by auto
  qed
next
  case Nil thus ?case by auto
next
  case (Fake tr mintr I tsend)
  hence "((tdone, Claim (Honest A) \<lbrace>Agent B, Real d\<rbrace>)) \<in> set tr" by auto
  with Fake.hyps prems show ?case by (auto)
next
  case (UDB2 tr trecv D NB tsend)
  hence "((tdone, Claim (Honest A) \<lbrace>Agent B, Real d\<rbrace>)) \<in> set tr" by auto
  with UDB2.hyps prems show ?case by (auto)
next
  case (UDB1 tr NB tsend D)
  hence "((tdone, Claim (Honest A) \<lbrace>Agent B, Real d\<rbrace>)) \<in> set tr" by auto
  with UDB1.hyps prems show ?case by (auto)
next
  case (Con tr tc C mc D tab)
  hence "((tdone, Claim (Honest A) \<lbrace>Agent B, Real d\<rbrace>)) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
qed
qed

end
