(******************************************************************************

 ProtoVeriPhy

 Module: distance/BrandsChaum_attack.thy

 Copyright (C) 2011-2012 Benedikt Schmidt
 License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Security analysis of the signature based Brands-Chaum protocol
          which results in a proof that there is a trace that violates
          distance-bounding security. *}

theory BrandsChaum_attack imports SystemCoffset SystemOrigination MessageTheoryXor3 begin

locale INITSTATE_SIG_NN = INITSTATE_PKSIG + INITSTATE_NONONCE

definition
  initStateMd :: "agent \<Rightarrow> msg set" where
  "initStateMd A == Key`({priSK A} \<union> (pubSK`UNIV))"

interpretation INITSTATE_SIG_NN Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components
               initStateMd Key
  apply (unfold_locales, auto simp add: initStateMd_def dest: injective_symKey)
  apply (drule subterms.singleton)
  apply (auto dest: injective_symKey)
  apply (drule subterms.singleton)
  apply (auto dest: injective_symKey)
  apply (drule subterms.singleton)
  apply (auto dest: injective_symKey simp add: MACM_def)
done

(* Informal multiset rewriting description of the protocol:

  md1(P): Fr(NP_P) \<longrightarrow> S_1(NP_P), Out(Hash(NP_P))

  md2(V): In(Com), Fr(NV_V) \<longrightarrow> S_2(Com,NV_V), Out (NV_V)

  md3(P): In(NV), S_1(NP_P) \<longrightarrow> Out( NV \<oplus> NP_P), S_3(NP_P,NV)

  // we do not send the open for the commit since it can be already
  // computed by V as NV \<oplus> response
  md4(P): S_3(NP_P,NV) \<longrightarrow> Out(SIG_{SK(P)}( NV_V, NP \<oplus> NV_V) ))

  md5(V): S_2(Hash(NP,P) ,NV_V),
          In(NV_V \<oplus> NP)
          In(SIG_{SK(P)}( NV_V, NP) )
          \<longrightarrow>
          Claim(..)

*)
definition
  md1 :: "msg step"
 where
  "md1 tr P t =
     (UN NP. {ev. ev = ( Hash  (Nonce (Honest P) NP)
                       , SendEv 0 [Number 1, Nonce (Honest P) NP]) \<and>
                  Nonce (Honest P) NP \<notin> usedI tr})"


definition
  md2 :: "msg step"
 where
  "md2 tr V t =
     (UN NV COM trec.
        {ev. ev = (Nonce (Honest V) NV,  SendEv 0 [Number 2, COM, Nonce (Honest V) NV]) \<and>
             Nonce (Honest V) NV \<notin> usedI tr \<and>
             (trec, Recv (Rec (Honest V)) COM) \<in> set tr})"

definition
  md3 :: "msg step"
 where
  "md3 tr P t =
     (UN NP NV trec tsend1 COM.
        {ev. ev = ( Xor NV (Nonce (Honest P) NP) 
                  , SendEv 0 [Number 3, Nonce (Honest P) NP, NV]) \<and>
             (tsend1, Send (Tr (Honest P)) COM [Number 1, Nonce (Honest P) NP]) \<in> set tr \<and>
             (trec,   Recv (Rec (Honest P)) NV) \<in> set tr})"

definition
  md4 :: "msg step"
 where
  "md4 tr P t =
     (UN NP NV tsend trecv.
        {ev. ev = ( Crypt (priSK (Honest P))
                   (\<lbrace> NV, (Xor NV (Nonce (Honest P) NP)) \<rbrace>)
                  , SendEv 0 []) \<and>
             (trecv, Recv (Rec (Honest P)) NV) \<in> set tr \<and> (* not strictly neccessary *)
             (tsend, Send (Tr (Honest P))
                          (Xor NV (Nonce (Honest P) NP))
                          [Number 3, Nonce (Honest P) NP, NV])
             \<in> set tr})"

definition
  md5 :: "msg step"
 where
  "md5 tr V t =
     (UN NP NV P trec1 trec2 tsend CHAL.
       {ev. ev = (\<lbrace>Agent P, Real ((trec1 - tsend) * vc/2)\<rbrace>, ClaimEv) \<and>
            (trec2, Recv (Rec (Honest V))
                  ( Crypt (priSK P)
                   \<lbrace> Nonce (Honest V) NV, Xor (Nonce (Honest V) NV) NP\<rbrace>)) \<in> set tr \<and>
            (trec1, Recv (Rec (Honest V)) (Xor (Nonce (Honest V) NV) NP)) \<in> set tr \<and>
            (tsend, Send (Tr (Honest V))  CHAL [Number 2, Hash NP, Nonce (Honest V) NV ]) \<in> set tr})"

definition
  md_proto :: "msg proto" where
  "md_proto = {md1,md2,md3,md4,md5}"

lemmas md_defs = md_proto_def md1_def md2_def md3_def md4_def md5_def


locale PROTOCOL_MD = PROTOCOL_PKSIG_NOKEYS+PROTOCOL_NONONCE+INITSTATE_SIG_NN

interpretation PROTOCOL_MD Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components initStateMd Key md_proto
  apply (unfold_locales)
  apply (auto simp add: md_defs messagesProtoTr_def messagesProtoTrHonest_def
                        initStateMd_def md_defs
              split: event.split split_if dest: parts.fst_set)

  apply (drule parts.singleton)
  apply auto
  apply (drule parts_Key_Xor)
  apply (drule parts.singleton)
  apply auto
  prefer 2
  apply (drule parts.singleton)
  apply auto
  apply (drule_tac t="trecv" in view_elem_ex)
  apply auto

  apply (drule parts_Key_Xor)
  apply (drule parts.singleton)
  apply auto
  prefer 2
  apply (drule parts.singleton)
  apply auto
  apply (drule_tac t="trec" in view_elem_ex)
  apply auto

  apply (drule parts.singleton)
  apply auto
  apply (drule_tac t="trecv" in view_elem_ex)
  apply auto
  done

text {* Agents only look at their own views and all messages are derivable. *}

interpretation PROTOCOL_EXECUTABLE Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components initStateMd md_proto sys Key
  apply (unfold_locales)
  apply (auto simp add: md_defs initStateMd_def
                        messagesProto_def messagesProtoTrHonest_def)

  apply (rule DM.Hash)
  apply force

  apply (rule DM.Xor)
  apply (drule view_elem_ex)
  apply (erule exE)
  apply (drule Recv_imp_knows_A)
  apply force
  apply force

  apply (rule DM.Crypt)
  apply (rule DM.MPair)
  apply (drule view_elem_ex)
  apply (erule exE)
  apply (drule Recv_imp_knows_A)
  apply force

  apply (rule DM.Xor)
  apply (drule view_elem_ex)
  apply (erule exE)
  apply (drule Recv_imp_knows_A)
  apply force
  apply force
  apply force

  apply (rule DM.MPair)
  apply force
  apply force

  apply (auto simp add: nonce_view_fresh [simplified md_proto_def]
                        nonce_view_used [simplified md_proto_def]
                        recv_a_view_a_r send_a_view_a_r)


  apply (rule_tac x=NP in exI)

  apply (rule_tac x=NV in exI)
  apply (rule_tac x=P in exI)
  apply (rule_tac x=trec1 in exI)
  apply (rule_tac x=trec2 in exI)
  apply (rule_tac x=tsend in exI)
  apply auto
  apply (rule_tac x=NP in exI)
  apply (rule_tac x=NV in exI)
  apply (rule_tac x=P in exI)
  apply (rule_tac x=trec1 in exI)
  apply (rule_tac x=trec2 in exI)
  apply (rule_tac x=tsend in exI)
  apply auto
done

text {* Agent behaviour does not change with constant clock errors. *}

interpretation  PROTOCOL_DELTAONLY Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components initStateMd Key md_proto
  apply unfold_locales
  apply (auto simp add: md_defs in_timetrans)
  apply (rule_tac x=NV in exI)
  apply force
  apply (rule_tac x=NP in exI)
  apply auto
  apply (rule_tac x=NP in exI)
  apply (rule_tac x=NV in exI)
  apply (rule_tac x=P in exI)
  apply (rule_tac x="trec1 - coffset A" in exI)
  apply (rule_tac x="trec2 - coffset A" in exI)
  apply (rule_tac x="tsend - coffset A" in exI)
  apply auto
  apply (simp add: sign_simps)

  apply (rule_tac x=NV in exI)
  apply auto
  apply (rule_tac x="trec + coffset A" in exI, force)
  apply (rule_tac x=NP in exI) 
  apply auto
  apply (rule_tac x="tsend1 + coffset A" in exI, force)
  apply (rule_tac x="trec + coffset A" in exI, force)
  apply (rule_tac x=NP in exI) apply (rule_tac x=NV in exI)
  apply auto
  apply (rule_tac x="trecv + coffset A" in exI)
  apply force
  
  apply (rule_tac x="tsend + coffset A" in exI, force)
  
  apply (rule_tac x=NP in exI) apply (rule_tac x=NV in exI)
  apply (rule_tac x=P in exI)
  apply (rule_tac x="trec1 + coffset A" in exI)
  apply (rule_tac x="trec2 + coffset A" in exI)
  apply (rule_tac x="tsend + coffset A" in exI)
  apply auto
  apply (simp add: sign_simps)
  done


interpretation PROTOCOL_DELTA_EXEC Crypt Nonce MPair Hash Number
                                   parts subterms DM LowHamXor Xor components
                                   initStateMd Key md_proto sys
by unfold_locales

subsection {* Direct Definition for Brands-Chaum protocol *}

inductive_set
  mdb :: "(msg trace) set"
 where
   Nil [intro] : "[] \<in> mdb"
 | Fake: 
   "\<lbrakk> tr \<in> mdb; t >=  maxtime tr;
      X \<in> DM (Intruder I) (knowsI (Intruder I) tr) \<rbrakk>
   \<Longrightarrow> (t, Send (Tx (Intruder I) j) X [])  # tr \<in> mdb"

 | Con : 
   "\<lbrakk> tr \<in> mdb; trecv >= maxtime tr;
      \<forall>X\<in>components {M}.
        \<exists>tsend A i M' L.
           \<exists>Y\<in>components {M'}.
              (tsend, Send (Tx A i) M' L) \<in> set tr \<and>
              cdistM (Tx A i) (Rx B j) = Some tab \<and> tsend + tab \<le> trecv \<and> Xor X Y \<in> LowHamXor \<rbrakk>
    \<Longrightarrow> (trecv, Recv (Rx B j) M) # tr \<in> mdb"

 (* Protocol rules for Brands-Chaum *)
 | MD1:  
   "\<lbrakk> tr \<in> mdb; t >=  maxtime tr;
      \<not> (used tr (Nonce (Honest P) NP)) \<rbrakk>
   \<Longrightarrow> (t, Send (Tr (Honest P)) (Hash (Nonce (Honest P) NP)) [Number 1, Nonce (Honest P) NP]) # tr \<in> mdb"

 | MD2:  
   "\<lbrakk> tr \<in> mdb; t >=  maxtime tr;
      (trec, Recv (Rec (Honest V)) COM) \<in> set tr;
      \<not> (used tr (Nonce (Honest V) NV)) \<rbrakk>
   \<Longrightarrow> (t, Send (Tr (Honest V)) (Nonce (Honest V) NV) [Number 2, COM, Nonce (Honest V) NV]) # tr \<in> mdb"

 | MD3:
   "\<lbrakk> tr \<in> mdb; tsend >=  maxtime tr;
      (trec, Recv (Rec (Honest P)) NV) \<in> set tr;
      (tsend2, Send (Tr (Honest P)) COM [Number 1, Nonce (Honest P) NP]) \<in> set tr \<rbrakk>
  \<Longrightarrow> (tsend, Send (Tr (Honest P))
                   (Xor NV (Nonce (Honest P) NP))
                   [Number 3, Nonce (Honest P) NP, NV])
      # tr \<in> mdb"

 | MD4:  
   "\<lbrakk> tr \<in> mdb; tsend >=  maxtime tr;
      (trecv, Recv (Rec (Honest P)) NV) \<in> set tr;
      (t, Send (Tr (Honest P))
               (Xor NV (Nonce (Honest P) NP))
               [Number 3, Nonce (Honest P) NP, NV])
     \<in> set tr \<rbrakk>
   \<Longrightarrow> (tsend,
       Send (Tr (Honest P))
            (Crypt (priSK (Honest P))
                   \<lbrace> NV, Xor NV (Nonce (Honest P) NP)\<rbrace>) [])
       # tr \<in> mdb"

 | MD5:
   "\<lbrakk> tr \<in> mdb; tdone \<ge> maxtime tr;
      (trec2, Recv (Rec (Honest V))
                  ( Crypt (priSK P)
                   \<lbrace> Nonce (Honest V) NV, Xor (Nonce (Honest V) NV) NP\<rbrace>))
      \<in> set tr;
      (trec1, Recv (Rec (Honest V)) (Xor (Nonce (Honest V) NV) NP))
      \<in> set tr;
      (tsend, Send (Tr (Honest V)) CHAL [Number 2, Hash NP, Nonce (Honest V) NV ]) \<in> set tr \<rbrakk>
  \<Longrightarrow> (tdone, Claim (Honest V) \<lbrace>Agent P, Real ((trec1 - tsend) * vc/2)\<rbrace>) # tr
     \<in> mdb"

text {* obtain a simpler induction rule for protocol since it is executable and deltaonly *}

lemmas proto_induct =
  sys.induct [simplified derivable_removable remove_occursAt timetrans_removable]

subsection {* Equality for direct and parameterized Definition *}

text {* We now show that both inductive definitions define the
        same set of traces. *}

lemma abstr_equal: "mdb = sys"
proof auto
  fix tr
  assume r: "tr \<in> sys"
  show "tr \<in> mdb" using r
  proof (induct tr rule: proto_induct)
    case 1 with prems show ?case by auto
  next
    case 2 with prems show ?case by (auto intro: mdb.Nil)
  next
    case 4 with prems show ?case apply -
      apply (rule mdb.Con)
      by (auto)
  next
    case 3 with prems show ?case by (auto intro: mdb.Fake)
  next
    case 5
    thus ?case
      apply (auto simp add: md_defs)
      apply (auto intro!: mdb.MD1 mdb.MD2 mdb.MD3 [simplified] mdb.MD4 mdb.MD5 simp add: usedI_def)
      apply (auto simp add: mem_def usedI_def)
      done
  qed
next
  fix tr
  assume r: "tr \<in> mdb"
  show "tr \<in> sys" using r
  proof(induct tr rule: mdb.induct)
    case Nil
    with prems show ?case by auto
  next
    case (Fake tr ts X I j)
    with prems show ?case by (auto intro: sys.Fake)
  next
    case (Con tr)
    with prems show ?case apply -
      apply (rule sys.Con)
      by (auto)
  next
    case (MD1 tr ts A NA)
    with prems have "(ts,createEv A (SendEv 0 [Number 1, Nonce (Honest A) NA]) (Hash (Nonce (Honest A) NA))) # tr \<in> sys"
      apply -
      apply (rule_tac step=md1 in sys_Proto_exec)
      apply force
      apply force
      apply force
      apply (force simp add: md_proto_def)
      apply (auto simp add: md_defs)
      apply (rule_tac x=NA in exI)
      apply auto
      apply (auto simp add: usedI_def initStateMd_def)
      apply (force simp: mem_def)
      apply (drule subterms.singleton)
      apply auto
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD2 tr tsend trecv V COM NV)
    with prems have
      "(tsend, 
        createEv V
                 (SendEv 0 [Number 2, COM, Nonce (Honest V) NV])
                 (Nonce (Honest V) NV))
       # tr \<in> sys"
      apply - apply (rule_tac step=md2 in sys_Proto)
      apply (auto simp add: md_defs usedI_def)
      apply (auto simp add: mem_def)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD3 tr tsend trecv P NV tsend2 COM NP)
    with prems have
      "(tsend,
        createEv P (SendEv 0 [Number 3, Nonce (Honest P) NP, NV])
                   (Xor NV (Nonce (Honest P) (NP)))) # tr \<in> sys"
      apply - apply (rule_tac step=md3 in sys_Proto)
      apply (auto simp add: md_defs)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD5 tr tdone trec2 V P NV NP trec1 tsend CHA)
    with prems have
      "(tdone, createEv V ClaimEv \<lbrace>Agent P, Real ((trec1 - tsend) * vc/2)\<rbrace>) # tr \<in> sys"
      apply - apply (rule_tac step=md5 in sys_Proto)
      apply (auto simp add: md_defs)
      apply (intro exI conjI)
      apply auto
      done    
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD4 tr tsend trecv P NV t NP)
    with prems have
      "(tsend, createEv P (SendEv 0 [])
                          (Crypt (priSK (Honest P))
                             \<lbrace>NV, Xor NV (Nonce (Honest P) NP) \<rbrace>)) # tr \<in> sys"
      apply - apply (rule_tac step=md4 in sys_Proto)
      apply (auto simp add: md_defs)
      done
    thus ?case
      apply (auto simp add: createEv.psimps)
      done
  qed
qed


lemmas [simp,intro] = abstr_equal [THEN sym]


lemma Xor_idem[simp]:  "Xor a a = Zero"
  apply (auto simp add: Xor_def Zero_def)
  done

lemma components_xor_n_n_a:
  "components {Xor (Nonce A NA) (Nonce B NB)}
   = {Xor (Nonce A NA) (Nonce B NB)}"
  apply (rule components_non_pair)
  apply (subgoal_tac "NONCE A NA \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "NONCE B NB \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp add: Xor_def MPair_def Nonce_def Agent_def simp del: norm.simps)
  apply (subgoal_tac "MPAIR (Rep_msg X) (Rep_msg Y) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "normed (norm
         (Rep_msg (Abs_msg (NONCE A NA)) \<oplus>
           norm
          (Rep_msg (Abs_msg (NONCE B NB)))))") prefer 2
  apply (force simp add: msg_def)
  
  apply (auto simp add: Abs_msg_inverse  split: split_if_asm)
  done

lemma attack_tr:
   assumes cdPV: "cdistM (Tr (Honest P))   (Rec (Honest V))   = Some dPV" and
           cdVP: "cdistM (Tr (Honest V))   (Rec (Honest P))   = Some dVP" and
           cdIV: "cdistM (Tr (Intruder I)) (Rec (Honest V))   = Some dIV" and
           cdVI: "cdistM (Tr (Honest V))   (Rec (Intruder I)) = Some dVI" and
           cdPI: "cdistM (Tr (Honest P))   (Rec (Intruder I)) = Some dPI" and 
           dist: "dPV + dVP < cdistl (Intruder I) (Honest V) * 2"
  shows " \<exists> tr t d.  (tr \<in> mdb) \<and>
                     ((t, Claim (Honest V) \<lbrace>Agent (Intruder I), Real d\<rbrace>) \<in> set tr) \<and>
                     (d < pdist (Intruder I) (Honest V))"
proof -
  let ?NP  = "Nonce (Honest P) 0"
  let ?NV  = "Nonce (Honest V) 1"
  let ?COM = "(Hash (?NP))"

  
  have dPV_pos: "dPV \<ge> 0" using prems apply - by (auto dest: cdistnoneg_some)
  have dVP_pos: "dVP \<ge> 0" using prems apply - by (auto dest: cdistnoneg_some)
  have dIV_pos: "dIV \<ge> 0" using prems apply - by (auto dest: cdistnoneg_some)
  have dVI_pos: "dVI \<ge> 0" using prems apply - by (auto dest: cdistnoneg_some)
  have dPI_pos: "dPI \<ge> 0" using prems apply - by (auto dest: cdistnoneg_some)
  have zero_lh: "Zero \<in> LowHamXor" by (rule LowHamXor.Zero)

  let ?tr1 = "(0, (Send (Tr (Honest P)) ?COM [Number 1, ?NP])) # []"
  have v1: "?tr1 \<in> mdb" apply -
    apply (rule mdb.MD1)
    by auto

  let ?tr2 = "(dPV, Recv  (Rec (Honest V)) ?COM)#?tr1"
  have v2: "?tr2 \<in> mdb" using v1 prems dPV_pos zero_lh apply -
    apply (rule mdb.Con)
    apply (rule v1)
    by auto

  let ?tr3 = "(dPV, Send (Tr (Honest V)) ?NV [Number 2, ?COM, ?NV])#?tr2"
  have v3: "?tr3 \<in> mdb" using v2 prems dPV_pos apply -
    apply (rule mdb.MD2)
    apply (auto simp add: Xor_Zero)
    apply (subgoal_tac  "?NV \<in> {?NP, ?COM}")
    prefer 2
    apply (simp add: mem_def)
    apply force
    done

  let ?tr4 = "(dPV+dVP, Recv (Rec (Honest P)) ?NV)#?tr3"
  have v4: "?tr4 \<in> mdb" using v3 prems dPV_pos zero_lh dVP_pos apply -
    apply (rule mdb.Con)
    apply (auto)
    apply (rule_tac x="dPV" in exI)
    apply force
    done

  let ?tr5 = "(dPV+dVP, Send (Tr (Honest P))
                              (Xor ?NV ?NP)
                              [Number 3, ?NP, ?NV])#?tr4"
  have v5: "?tr5 \<in> mdb" using v4 prems dPV_pos zero_lh dVP_pos apply -
    apply (rule mdb.MD3)
    apply (auto simp add: Xor_Zero)
    done

  let ?tr6 = "(dPV+dVP+dPV, Recv (Rec (Honest V))
                                 (Xor ?NV ?NP)) #?tr5"
  have v6: "?tr6 \<in> mdb" using v5 prems dPV_pos zero_lh dVP_pos apply -
    apply (rule mdb.Con)
    apply (rule v5)
    apply (clarsimp)
    apply (auto simp add: components_xor_n_n_a)
    apply (rule_tac x="dPV + dVP" in exI)
    apply (rule_tac x="Honest P" in exI)
    apply (rule_tac x="0" in exI)
    apply (rule_tac x=" (Xor (Nonce (Honest V) (Suc 0)) (Nonce (Honest P) 0))" in exI)
    apply auto
    apply (auto simp add: components_xor_n_n_a)
    done
   
   let ?tr7 = "(dPV+dVP+dPV+dVI, Recv (Rec (Intruder I)) ?NV)#?tr6"
   have v7: "?tr7 \<in> mdb" using v6 prems dPV_pos zero_lh dVP_pos dVI_pos dIV_pos apply -
     apply (rule mdb.Con)
     apply (rule v6)
     apply (auto simp add: components_nonce)
     apply (rule_tac x="dPV" in exI)
     apply (rule_tac x="Honest V" in exI)
     apply (rule_tac x="0" in exI)
     apply (rule_tac x=" ?NV" in exI)
     apply (auto simp add: components_nonce)
     done
   
   let ?RESP = "Xor ?NV ?NP"
   
   let ?tr8 = "(dPV+dVP+dPV+dVI+dPI, Recv (Rec (Intruder I)) ?RESP)#?tr7"
   have v8: "?tr8 \<in> mdb" using v7 prems dPV_pos zero_lh dVP_pos dVI_pos dIV_pos dPI_pos apply -
     apply (rule_tac mdb.Con)
     apply (rule v7) prefer 2
     apply (auto simp add: components_xor_n_n_a)
     apply (rule_tac x="dPV + dVP" in exI)
     apply (rule_tac x="Honest P" in exI)
     apply (rule_tac x="0" in exI)
     apply (rule_tac x=" ?RESP" in exI)
     apply (auto simp add: components_xor_n_n_a)
     done

   let ?tr9 = "(dPV+dVP+dPV+dVI+dPI, Send (Tr (Intruder I))
                                      (Crypt (priSK (Intruder I))
                                        \<lbrace>?NV, ?RESP\<rbrace>) [])#?tr8"
   have v9: "?tr9 \<in> mdb" using v8 prems dPV_pos zero_lh dVP_pos dVI_pos dIV_pos dPI_pos apply -
    apply (rule mdb.Fake)
    apply (rule v8)
    apply (force)
    apply (rule DM.Crypt) defer
    apply (auto simp add: knowsI_def initStateMd_def)
    done
(*    apply (rule DM.MPair)
    apply (rule DM.Inj)
    apply (auto simp add: Xor_Zero)
    
    apply (rule DM.MPair) defer
    apply force
    apply (subgoal_tac "Xor ?NV (Xor ?NV ?NP)
                        \<in> DM (Intruder I)
      (insert (Key (priSK (Intruder I)))
            (insert (Xor (Nonce (Honest V) (Suc 0)) (Nonce (Honest P) 0))
              (insert (Nonce (Honest V) (Suc 0)) (Key ` range pubSK))))")
    apply (subgoal_tac "Xor ?NV (Xor ?NV ?NP) = ?NP")
    apply force defer
    apply (rule_tac DM.Xor)
    apply force
    apply force
    apply auto
    done
*)
  let ?tr10 = "(dPV+dVP+dPV+dVI+dPI+dIV, Recv (Rec (Honest V))
                                  (Crypt (priSK (Intruder I))
                                          \<lbrace>?NV, ?RESP\<rbrace>))#?tr9"
   have v10: "?tr10 \<in> mdb" using v9 prems dPV_pos zero_lh dVP_pos dIV_pos dVI_pos dIV_pos dPI_pos apply -
    apply (rule_tac mdb.Con)
    apply (auto simp add: components_crypt)
    apply (rule_tac x="dPV+dVP+dPV+dVI+dPI" in exI)
    apply auto
    apply (rule_tac x="Intruder I" in exI)
    apply (rule_tac x="0" in exI)
    apply (rule_tac x="(Crypt (priSK (Intruder I))
                              \<lbrace>?NV, ?RESP\<rbrace>)" in exI)
    apply (auto simp add: components_crypt)
    done

  let ?tr11 = "(dPV+dVP+dPV+dVI+dPI+dIV, Claim (Honest V)
                                      \<lbrace>Agent (Intruder I), Real ((dPV+dVP+dPV - dPV)*vc/2) \<rbrace>)#?tr10"
  have v11 : "?tr11 \<in> mdb" using v10 prems dPV_pos zero_lh dVP_pos dIV_pos  dVI_pos dIV_pos dPI_pos apply -
    apply (rule mdb.MD5[where CHAL="?NV" and P="Intruder I" and NP="?NP" and NV=1 and V="V"])
    apply (rule v10)
    apply force
    apply (simp (no_asm) only: set.simps)
    apply (rule Set.insertCI)
    apply (rule HOL.refl)
    apply (simp (no_asm) only: set.simps)
    apply (rule Set.insertI2)
    apply (rule Set.insertI2)
    apply (rule Set.insertI2)
    apply (rule Set.insertI2)
    apply (rule Set.insertCI) defer
    
    apply (simp (no_asm) only: set.simps)
    apply (rule Set.insertI2)
    apply (rule Set.insertI2)
    apply (rule Set.insertI2)
    apply (rule Set.insertI2)
    apply (rule Set.insertI2)
    apply (rule Set.insertI2)
    apply (rule Set.insertI2)
    apply (rule Set.insertCI)
    apply (rule HOL.refl)
    apply (simp add: Xor_comm Xor_comm2 Xor_assoc)
    done
  thus ?thesis using prems
    apply -
    apply (rule_tac x="?tr11" in exI)
    apply (rule_tac x="dPV+dVP+dPV+dVI+dPI+dIV" in exI)
    apply (rule_tac x="((dPV+dVP+dPV - dPV)*vc/2)" in exI)
    apply (rule conjI)
    apply (rule v11)
    apply (rule conjI)
    apply force
    apply (auto simp add: cdistl_def)
    apply (insert vc_pos)
    apply (case_tac "vc=0")
    apply auto
    apply (rule_tac c="1/vc" in mult_right_less_imp_less)
    apply auto
    done
qed

end
