(******************************************************************************

 ProtoVeriPhy

 Module: distance/BrandsChaum_implicit.thy

 Copyright (C) 2011-12 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
 License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Security Analysis of a fixed version of the Brands-Chaum protocol
          that uses implicit binding to prevent Distance Hijacking attacks.
          We prove that the resulting protocol is secure in our model
          Note that we abstract away from the individual bits exchanged in the rapid
          bit exchange phase, by performing the message exchange in 2 steps instead
          2*k steps. *}

theory BrandsChaum_implicit imports SystemCoffset SystemOrigination MessageTheoryXor3 begin

locale INITSTATE_SIG_NN = INITSTATE_PKSIG + INITSTATE_NONONCE

definition
  initStateMd :: "agent \<Rightarrow> msg set" where
  "initStateMd A == Key`({priSK A} \<union> (pubSK`UNIV))"

interpretation INITSTATE_SIG_NN Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components
               initStateMd Key
  apply (unfold_locales, auto simp add: initStateMd_def dest: injective_symKey)
  apply (drule subterms.singleton)
  apply (auto dest: injective_symKey)
  apply (drule subterms.singleton)
  apply (auto dest: injective_symKey)
  apply (drule subterms.singleton)
  apply (auto dest: injective_symKey simp add: MACM_def)
done

(* Informal multiset rewriting description of the protocol:

  The commit also includes the identity of the prover.
  md1(P): Fr(NP_P) \<longrightarrow> S_1(NP_P), Out(Hash(NP_P,P))

  md2(V): In(Com), Fr(NV_V) \<longrightarrow> S_2(Com,NV_V), Out (NV_V)

  md3(P): In(NV), S_1(NP_P) \<longrightarrow> Out( NV \<oplus> NP_P), S_3(NP_P,NV)

  md4(P): S_3(NP_P,NV) \<longrightarrow> Out( SIG_{SK(P)}( NV,NP_P,V ))

  md5(V): S_2(Hash(NP,P) ,NV_V),
          In(NV_V \<oplus> NP)
          In(SIG_{SK(P)}( NV_V, NP, P) ))
          \<longrightarrow>
          Claim(..)

*)

definition
  md1 :: "msg step"
 where
  "md1 tr P t =
     (UN NP. {ev. ev = ( Hash  \<lbrace> Nonce (Honest P) NP, Agent (Honest P)\<rbrace>
                       , SendEv 0 [Number 1, Nonce (Honest P) NP]) \<and>
                  Nonce (Honest P) NP \<notin> usedI tr})"


definition
  md2 :: "msg step"
 where
  "md2 tr V t =
     (UN NV COM trec.
        {ev. ev = (Nonce (Honest V) NV,  SendEv 0 [Number 2,COM, Nonce (Honest V) NV]) \<and>
             Nonce (Honest V) NV \<notin> usedI tr \<and>
             (trec, Recv (Rec (Honest V)) COM) \<in> set tr})"

definition
  md3 :: "msg step"
 where
  "md3 tr P t =
     (UN NP NV trec tsend1 COM.
        {ev. ev = ( Xor NV (Nonce (Honest P) NP)
                  , SendEv 1 [Number 3, Nonce (Honest P) NP, NV]) \<and>
             (\<forall> t m nv k. (t, Send (Tx (Honest P) k) m [Number 3, Nonce (Honest P) NP, nv]) \<notin> set tr) \<and>
             (tsend1, Send (Tr (Honest P)) COM [Number 1, Nonce (Honest P) NP]) \<in> set tr \<and>
             (trec,   Recv (Rec (Honest P)) NV) \<in> set tr})"

definition
  md4 :: "msg step"
 where
  "md4 tr P t =
     (UN NP NV V tsend trecv.
        {ev. ev = ( Crypt (priSK (Honest P))
                   \<lbrace> NV, \<lbrace>Nonce (Honest P) NP,Agent V\<rbrace>\<rbrace>
                  , SendEv 0 []) \<and>
             (trecv, Recv (Rec (Honest P)) NV) \<in> set tr \<and> (* not strictly neccessary *)
             (tsend, Send (Tu (Honest P))
                          (Xor NV (Nonce (Honest P) NP))
                          [Number 3, Nonce (Honest P) NP, NV])
             \<in> set tr})"

definition
  md5 :: "msg step"
 where
  "md5 tr V t =
     (UN NP NV P trec1 trec2 tsend CHAL.
       {ev. ev = (\<lbrace>Agent P, Real ((trec1 - tsend) * vc/2)\<rbrace>, ClaimEv) \<and>
            P \<noteq> (Honest V) \<and>  (* FIXME: would be nice to remove this *)
            (trec2, Recv (Rec (Honest V))
                  ( Crypt (priSK P)
                   \<lbrace> Nonce (Honest V) NV, \<lbrace> NP, Agent (Honest V)\<rbrace>\<rbrace>)) \<in> set tr \<and>
            (trec1, Recv (Ru (Honest V)) (Xor (Nonce (Honest V) NV)  NP)) \<in> set tr \<and>
            (tsend, Send (Tr (Honest V)) CHAL [Number 2, Hash \<lbrace> NP, Agent P\<rbrace> , Nonce (Honest V) NV ]) \<in> set tr})"

definition
  md_proto :: "msg proto" where
  "md_proto = {md1,md2,md3,md4,md5}"

lemmas md_defs = md_proto_def md1_def md2_def md3_def md4_def md5_def


locale PROTOCOL_MD = PROTOCOL_PKSIG_NOKEYS+PROTOCOL_NONONCE+INITSTATE_SIG_NN

interpretation PROTOCOL_MD Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components initStateMd Key md_proto
  apply (unfold_locales)
  apply (auto simp add: md_defs messagesProtoTr_def messagesProtoTrHonest_def
                        initStateMd_def
              split: event.split split_if dest: parts.fst_set)
  apply (drule parts.singleton)
  apply auto
  apply (drule parts_Key_Xor)
  apply (drule parts.singleton)
  apply auto
  apply (drule_tac t="trec" in view_elem_ex)
  apply auto
  
  apply (drule parts.singleton)
  apply auto
  apply (drule_tac t="trecv" in view_elem_ex)
  apply auto
  done

text {* Agents only look at their own views and all messages are derivable. *}

interpretation PROTOCOL_EXECUTABLE Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components initStateMd md_proto sys Key
  apply (unfold_locales)
  apply (auto simp add: md_defs initStateMd_def
                        messagesProto_def messagesProtoTrHonest_def MACM_def)

  apply (rule DM.Hash)
  apply (rule DM.MPair)
  apply force
  apply force

  apply (rule DM.Xor)
  apply (drule view_elem_ex)
  apply (erule exE)
  apply (drule Recv_imp_knows_A)
  apply force
  apply force

  apply (rule DM.Crypt)
  apply (rule DM.MPair)
  apply (drule view_elem_ex)
  apply (erule exE)
  apply (drule Recv_imp_knows_A)
  apply force
  
  apply (rule DM.MPair)
  apply force
  apply force
  apply force

  apply (rule DM.MPair) 
  apply force
  apply force
  

  apply (auto simp add: nonce_view_fresh [simplified md_proto_def]
                        nonce_view_used [simplified md_proto_def]
                        recv_a_view_a_r send_a_view_a_r)


  apply (rule_tac x=NP in exI)
  apply auto defer

  apply (rule_tac x=NP in exI)
  apply (rule_tac x=NV in exI)
  apply auto
  apply (rule_tac x=tsend in exI)
  apply (force simp add: view_def in_timetrans)
  apply (rule_tac x=NP in exI)
  apply (rule_tac x=NV in exI)
  apply (rule_tac x=P in exI)
  apply (rule_tac x=trec1 in exI)
  apply (rule_tac x=trec2 in exI)
  apply (rule_tac x=tsend in exI)
  apply auto
  apply (force simp add: view_def in_timetrans)
  apply (rule_tac x=NP in exI)
  apply auto defer
  apply (rule_tac x=NP in exI)
  apply (rule_tac x=NV in exI)
  apply auto
  apply (rule_tac x=tsend in exI)
  apply (force simp add: view_def in_timetrans)
  apply (rule_tac x=NP in exI)
  apply (rule_tac x=NV in exI)
  apply (rule_tac x=P in exI)
  apply (rule_tac x=trec1 in exI)
  apply (rule_tac x=trec2 in exI)
  apply (rule_tac x=tsend in exI)
  apply (force simp add: view_def in_timetrans)
  apply (force simp add: view_def in_timetrans)
  apply (auto simp add: view_def in_timetrans)
  apply (erule_tac x="a + coffset A" in allE)
  apply (erule_tac x=m in allE)
  apply (erule_tac x=nv in allE)
  apply (erule_tac x=k in allE)
  apply (auto simp add: view_def in_timetrans)
  done

text {* Agent behaviour does not change with constant clock errors. *}

interpretation  PROTOCOL_DELTAONLY Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components initStateMd Key md_proto
  apply unfold_locales
  apply (auto simp add: md_defs in_timetrans)
  apply (rule_tac x=NV in exI)
    apply auto
  
  apply (rule_tac x=NP in exI)
  apply auto defer
  apply (rule_tac x=NP in exI)
  apply (rule_tac x=NV in exI)
  apply (rule_tac x=P in exI)
  apply (rule_tac x="trec1 - coffset A" in exI)
  apply (rule_tac x="trec2 - coffset A" in exI)
  apply (rule_tac x="tsend - coffset A" in exI)
  apply auto
  apply (simp add: sign_simps)

  apply (rule_tac x=NV in exI)
  apply auto
  apply (rule_tac x="trec + coffset A" in exI, force)
  apply (rule_tac x=NP in exI) 
  apply auto
  apply (rule_tac x="tsend1 + coffset A" in exI, force)
  apply (rule_tac x="trec + coffset A" in exI, force)
  apply (rule_tac x=NP in exI) apply (rule_tac x=NV in exI)
  apply auto
  apply (rule_tac x="trecv + coffset A" in exI)
  apply force
  
  apply (rule_tac x="tsend + coffset A" in exI, force)
  
  apply (rule_tac x=NP in exI) apply (rule_tac x=NV in exI)
  apply (rule_tac x=P in exI)
  apply (rule_tac x="trec1 + coffset A" in exI)
  apply (rule_tac x="trec2 + coffset A" in exI)
  apply (rule_tac x="tsend + coffset A" in exI)
  apply auto
  apply (simp add: sign_simps)
  apply (erule_tac x="t + coffset A" in allE)
  apply (erule_tac x=m in allE)
  apply (erule_tac x=nv in allE)
  apply (erule_tac x=k in allE)
  apply auto
  done

interpretation PROTOCOL_DELTA_EXEC Crypt Nonce MPair Hash Number
                                   parts subterms DM LowHamXor Xor components
                                   initStateMd Key md_proto sys
by unfold_locales

subsection {* Direct Definition for Brands-Chaum Variant *}

inductive_set
  mdb :: "(msg trace) set"
 where
   Nil [intro] : "[] \<in> mdb"
 | Fake: 
   "\<lbrakk> tr \<in> mdb; t >=  maxtime tr;
      X \<in> DM (Intruder I) (knowsI (Intruder I) tr) \<rbrakk>
   \<Longrightarrow> (t, Send (Tx (Intruder I) j) X [])  # tr \<in> mdb"

 | Con : 
   "\<lbrakk> tr \<in> mdb; trecv >= maxtime tr;
      \<forall>X\<in>components {M}.
        \<exists>tsend A i M' L.
           \<exists>Y\<in>components {M'}.
              (tsend, Send (Tx A i) M' L) \<in> set tr \<and>
              cdistM (Tx A i) (Rx B j) = Some tab \<and> tsend + tab \<le> trecv \<and> Xor X Y \<in> LowHamXor \<rbrakk>
    \<Longrightarrow> (trecv, Recv (Rx B j) M) # tr \<in> mdb"

 (* Protocol rules for Brands-Chaum *)
 | MD1:  
   "\<lbrakk> tr \<in> mdb; t >=  maxtime tr;
      \<not> (used tr (Nonce (Honest P) NP)) \<rbrakk>
   \<Longrightarrow> (t, Send (Tr (Honest P)) (Hash \<lbrace> Nonce (Honest P) NP, Agent (Honest P)\<rbrace>) [Number 1, Nonce (Honest P) NP]) # tr \<in> mdb"

 | MD2:  
   "\<lbrakk> tr \<in> mdb; t >=  maxtime tr;
      (trec, Recv (Rec (Honest V)) COM) \<in> set tr;
      \<not> (used tr (Nonce (Honest V) NV)) \<rbrakk>
   \<Longrightarrow> (t, Send (Tr (Honest V)) (Nonce (Honest V) NV) [Number 2, COM, Nonce (Honest V) NV]) # tr \<in> mdb"

 | MD3:
   "\<lbrakk> tr \<in> mdb; tsend >=  maxtime tr;
      (trec, Recv (Rec (Honest P)) NV) \<in> set tr;
      (tsend2, Send (Tr (Honest P)) COM [Number 1, Nonce (Honest P) NP]) \<in> set tr;
      (\<forall> t m nv k. (t, Send (Tx (Honest P) k) m [Number 3, Nonce (Honest P) NP, nv]) \<notin> set tr)\<rbrakk>
  \<Longrightarrow> (tsend, Send (Tu (Honest P))
                   (Xor NV (Nonce (Honest P) NP))
                   [Number 3, Nonce (Honest P) NP, NV])
      # tr \<in> mdb"

 | MD4:  
   "\<lbrakk> tr \<in> mdb; tsend >=  maxtime tr;
      (trecv, Recv (Rec (Honest P)) NV) \<in> set tr;
      (t, Send (Tu (Honest P))
               (Xor NV (Nonce (Honest P) NP))
               [Number 3, Nonce (Honest P) NP, NV])
     \<in> set tr \<rbrakk>
   \<Longrightarrow> (tsend,
       Send (Tr (Honest P))
            (Crypt (priSK (Honest P))
                   \<lbrace> NV, \<lbrace> Nonce (Honest P) NP, Agent V\<rbrace>\<rbrace>) [])
       # tr \<in> mdb"

 | MD5:
   "\<lbrakk> tr \<in> mdb; tdone \<ge> maxtime tr;
      (trec2, Recv (Rec (Honest V))
                  ( Crypt (priSK P)
                   \<lbrace> Nonce (Honest V) NV, \<lbrace> NP, Agent (Honest V)\<rbrace>\<rbrace>))
      \<in> set tr;
      (trec1, Recv (Ru (Honest V)) (Xor (Nonce (Honest V) NV) NP))
      \<in> set tr;
      (tsend, Send (Tr (Honest V)) CHAL [Number 2, Hash \<lbrace> NP, Agent P\<rbrace>, Nonce (Honest V) NV ]) \<in> set tr;
      P \<noteq> Honest V \<rbrakk>
  \<Longrightarrow> (tdone, Claim (Honest V) \<lbrace>Agent P, Real ((trec1 - tsend) * vc/2)\<rbrace>) # tr
     \<in> mdb"

text {* obtain a simpler induction rule for protocol since it is executable and deltaonly *}

lemmas proto_induct =
  sys.induct [simplified derivable_removable remove_occursAt timetrans_removable]

subsection {* Equality for direct and parameterized Definition *}

text {* We now show that both inductive definitions define the
        same set of traces. *}

lemma abstr_equal: "mdb = sys"
proof auto
  fix tr
  assume r: "tr \<in> sys"
  show "tr \<in> mdb" using r
  proof (induct tr rule: proto_induct)
    case 1 with prems show ?case by auto
  next
    case 2 with prems show ?case by (auto intro: mdb.Nil)
  next
    case 4 with prems show ?case apply -
      apply (rule  mdb.Con)
      apply auto
      done
  next
    case 3 with prems show ?case
      by (auto intro: mdb.Fake)
  next
    case 5
    thus ?case
      apply (auto simp add: md_defs)
      apply (auto intro!: mdb.MD1 mdb.MD2 mdb.MD3 [simplified] mdb.MD4 mdb.MD5 simp add: usedI_def)
      apply (auto simp add: mem_def usedI_def)
      done
  qed
next
  fix tr
  assume r: "tr \<in> mdb"
  show "tr \<in> sys" using r
  proof(induct tr rule: mdb.induct)
    case Nil
    with prems show ?case by auto
  next
    case (Fake tr ts X I j)
    with prems show ?case by (auto intro: sys.Fake)
  next
    case (Con tr)
    with prems show ?case apply - apply (rule sys.Con) by auto
  next
    case (MD1 tr ts A NA)
    with prems have "(ts,createEv A (SendEv 0 [Number 1, Nonce (Honest A) NA]) (Hash \<lbrace>Nonce (Honest A) NA, Agent (Honest A)\<rbrace>)) # tr \<in> sys"
      apply -
      apply (rule_tac step=md1 in sys_Proto_exec)
      apply force
      apply force
      apply force
      apply (force simp add: md_proto_def)
      apply (auto simp add: md_defs)
      apply (rule_tac x=NA in exI)
      apply auto
      apply (auto simp add: usedI_def initStateMd_def)
      apply (force simp: mem_def)
      apply (drule subterms.singleton)
      apply auto
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD2 tr tsend trecv V COM NV)
    with prems have
      "(tsend, 
        createEv V
                 (SendEv 0 [Number 2, COM, Nonce (Honest V) NV])
                 (Nonce (Honest V) NV))
       # tr \<in> sys"
      apply - apply (rule_tac step=md2 in sys_Proto)
      apply (auto simp add: md_defs usedI_def)
      apply (auto simp add: mem_def)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD3 tr tsend trecv P NV tsend2 COM NP)
    with prems have
      "(tsend,
        createEv P (SendEv 1 [Number 3, Nonce (Honest P) NP, NV])
                   (Xor NV (Nonce (Honest P) (NP)))) # tr \<in> sys"
      apply - apply (rule_tac step=md3 in sys_Proto)
      apply (auto simp add: md_defs)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD5 tr tdone trec2 V P NV NP trec1 tsend CHA)
    with prems have
      "(tdone, createEv V ClaimEv \<lbrace>Agent P, Real ((trec1 - tsend) * vc/2)\<rbrace>) # tr \<in> sys"
      apply - apply (rule_tac step=md5 in sys_Proto)
      apply (auto simp add: md_defs)
      apply (intro exI conjI)
      apply auto
      done    
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD4 tr tsend trecv P NV t NP V)
    with prems have
      "(tsend, createEv P (SendEv 0 [])
                          (Crypt (priSK (Honest P))
                             \<lbrace>NV, \<lbrace>Nonce (Honest P) NP, Agent V\<rbrace>\<rbrace>)) # tr \<in> sys"
      apply - apply (rule_tac step=md4 in sys_Proto)
      apply (auto simp add: md_defs)
      done
    thus ?case by (auto simp add: createEv.psimps)
  qed
qed

lemmas [simp,intro] = abstr_equal [THEN sym]

subsection {* Some invariants capturing the Behavior of honest Agents *}

lemma nonce_fresh_challenge:
  assumes mdb: "tr \<in> mdb" and
          send: "(ta, Send (Tx (Honest A) i) CHAL [Number 2,COM, Nonce (Honest A) NA]) \<in> set tr"
  shows   "Nonce (Honest A) NA
           \<notin> usedI (beforeEvent (ta, Send (Tx (Honest A) i) CHAL [Number 2, COM, Nonce (Honest A) NA]) tr)"
  using prems(1-)
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend) with Fake.hyps prems show ?case by auto
next
  case (Con tr tc C mc D tab) with Con.hyps prems show ?case by auto
next
  case (MD1 tr t P' NP' A')
  thus ?case using MD1.hyps prems by auto
next
  case (MD2 tr t_l trec_l V_l COM_l NV_l A)
  thus ?case using MD2.hyps prems
    apply auto
    apply (auto simp add: usedI_def initStateMd_def)
    apply (force simp add: mem_def)
    apply (drule subterms.singleton)
    apply auto
    done
next
  case (MD4 tr tsend trecv P NV t NP V A)
  with MD4.hyps prems show ?case by auto
next
  case (MD5 tr tdone trec2 V P NV NP trc1 tsend CHAL A)
  with MD5.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
next
  case (MD3 tr tsend_l trec_l P_l NV_l tsend2_l COM_l NP_l A)
  let ?sendev = "(ta, Send (Tx (Honest A) i) CHAL [Number 2, COM, Nonce (Honest A) NA])"
  have "?sendev \<in> set tr" using prems by auto
  thus ?case apply -
    apply (drule prems(4))
    apply auto
    done
qed

lemma nonce_fresh_commit:
  assumes mdb: "tr \<in> mdb" and
          send: "(ta, Send (Tx (Honest A) i) (Hash \<lbrace> NP, Agent P \<rbrace>)
                                             [Number 1, NP]) \<in> set tr"
  shows   "
           (\<exists> NA.
             P = Honest A \<and>
             NP = Nonce (Honest A) NA \<and>
             Nonce (Honest A) NA
             \<notin> usedI (beforeEvent
                        (ta, Send (Tx (Honest A) i) (Hash \<lbrace> Nonce (Honest A) NA,  Agent (Honest A) \<rbrace>)
                                                    [Number 1, Nonce (Honest A) NA]) tr))"
  using mdb send
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend) with Fake.hyps prems show ?case by auto
next
  case (Con tr tc C mc D tab) with Con.hyps prems show ?case by auto
next
  case (MD2 tr t trec V COM NV A')
  thus ?case using MD2.hyps prems by auto
next
  case (MD3 tr tsend trec P' NV tsend1 COM NP A')
  with MD3.hyps prems show ?case by auto
next
  case (MD4 tr tsend trecv P NV t NP V A)
  with MD4.hyps prems show ?case by auto
next
  case (MD5 tr tdone trec2 V P NV NP trc1 tsend CHAL A)
  with MD5.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
next
  case (MD1 tr t P' NP' A')
  let ?eva = "(ta, Send (Tx (Honest A') i) (Hash \<lbrace>NP, Agent P\<rbrace>) [Number 1, NP])"
  let ?newev = "(t, Send (Tr (Honest P')) (Hash \<lbrace>Nonce (Honest P') NP', Agent (Honest P')\<rbrace>)
            [Number 1, Nonce (Honest P') NP'])"
  show ?case proof cases
    assume eq: "?eva = ?newev"
    thus ?case using MD1.hyps prems apply -
      apply (rule_tac x=NP' in exI)
      apply (simp add: usedI_def)
      apply (auto simp add: mem_def)
      done
  next
    assume "?eva \<noteq> ?newev"
    hence "?eva \<in> set tr" using `?eva \<in> set (?newev#tr)` by auto
    thus ?case apply -
      apply (frule MD1.hyps(2))
      apply (elim conjE exE)
      apply auto
      done
  qed
qed

lemma nonce_fresh_commit2:
  assumes mdb: "tr \<in> mdb" and
          send: "(ta, Send (Tx (Honest A) i) (Hash \<lbrace> Nonce (Honest A) NA, Agent (Honest A)\<rbrace>)
                                                    [Number 1, Nonce (Honest A) NA])
                 \<in> set tr"
  shows   "Nonce (Honest A) NA
           \<notin> usedI (beforeEvent
                       (ta, Send (Tx (Honest A) i) (Hash \<lbrace> Nonce (Honest A) NA, Agent (Honest A)\<rbrace>)
                                                    [Number 1, Nonce (Honest A) NA])
                      tr)"
  using mdb send
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend) with Fake.hyps prems show ?case by auto
next
  case (Con tr tc C mc D tab) with Con.hyps prems show ?case by auto
next
  case (MD2 tr t trec V COM NV A')
  thus ?case using MD2.hyps prems by auto
next
  case (MD3 tr tsend trec P' NV tsend1 COM NP A')
  with MD3.hyps prems show ?case by auto
next
  case (MD4 tr tsend trecv P NV t NP V A)
  with MD4.hyps prems show ?case by auto
next
  case (MD5 tr tdone trec2 V P NV NP trc1 tsend CHAL A)
  with MD5.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
next
  case (MD1 tr t P' NP' A')
  let ?eva = "(ta, Send (Tx (Honest A') i) (Hash \<lbrace>Nonce (Honest A') NA, Agent (Honest A')\<rbrace>)
      [Number 1, Nonce (Honest A') NA])"
  let ?newev = "(t, Send (Tr (Honest P')) (Hash \<lbrace>Nonce (Honest P') NP', Agent (Honest P')\<rbrace>)
            [Number 1, Nonce (Honest P') NP'])"
  show ?case proof cases
    assume eq: "?eva = ?newev"
    thus ?case using MD1.hyps prems apply -
      apply (simp add: usedI_def)
      apply (auto simp add: mem_def)
      done
  next
    assume "?eva \<noteq> ?newev"
    hence "?eva \<in> set tr" using `?eva \<in> set (?newev#tr)` by auto
    thus ?case apply -
      apply (frule MD1.hyps(2))
      apply auto
      done
  qed
qed

lemma outside_hash_deducible_implies_received:
 assumes     sys_proto: "tr \<in> mdb"
         and ded:       "m \<in> DM B (knowsI B tr)"
         and neq:       "B \<noteq> A"
         and protected: "out_context (Nonce A NA) (Hash \<lbrace> Nonce A NA, Agent A\<rbrace>) m"
 shows "\<exists>trs X i.
            (trs, Recv (Rx B i) X) \<in> set tr
          \<and> out_context (Nonce A NA) (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>) X"
  using ded sys_proto neq protected
proof (induct  rule: DM.induct)
  case (Agent ag)
  thus ?thesis apply - by (auto dest: out_context_inverse)
next
  case (Number n)
  thus ?thesis apply - by (auto dest: out_context_inverse)
next
  case (Real n)
  thus ?thesis apply - by (auto dest: out_context_inverse)
next
  case (Nonce n)
  thus ?thesis apply - by (auto dest: out_context_inverse)
next
  case (Inj Y)
  thus ?thesis apply -
    apply clarsimp
    apply (drule knowsI_A_imp_Recv_initState)
    apply (auto simp add: initStateMd_def knowsI_def)
    apply (drule out_context_imp_subterms)
    apply auto
    apply (drule out_context_imp_subterms)
    apply auto
    done
next
  case (MPair Y Z)
  thus ?thesis apply -
    apply auto
    apply (drule out_context_inverse)
    apply auto
    done
next
  case (Crypt Y K)
  thus ?thesis apply -
    apply auto
    apply (drule out_context_inverse)
    apply auto
    done
next
  case (Hash Y)
  thus ?thesis apply -
    apply auto
    apply (drule out_context_inverse)
    apply auto
    done 
next
  case (Xor Y Z)
  thus ?thesis apply -
    apply auto
    apply (drule out_context_inverse)
    apply auto
    
    apply (subgoal_tac "out_context (Nonce A NA) (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>) Y
                        \<or> out_context (Nonce A NA) (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>) Z")
    apply force
    apply (drule factors_Xor_Nonce)
    apply auto
    apply (case_tac "Y = Nonce A NA")
    apply auto
    apply (case_tac "Z = Nonce A NA")
    apply auto defer

    apply (drule_tac k=k in out_context.Crypt)
    apply force
    apply (drule factors_Xor_Crypt)
    apply auto
    apply (case_tac "Y = (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>)")
    apply auto
    apply (case_tac "Z = (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>)")
    apply auto

    apply (drule_tac Y=Ya in out_context.PairL)
    apply force
    apply (drule factors_Xor_MPair)
    apply auto
    apply (case_tac "Y = (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>)")
    apply auto
    apply (case_tac "Z = (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>)")
    apply auto

    apply (drule_tac Y=Ya in out_context.PairR)
    apply force
    apply (drule factors_Xor_MPair)
    apply auto
    apply (case_tac "Y = (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>)")
    apply auto
    apply (case_tac "Z = (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>)")
    apply auto

    apply (drule factors_Xor)
    apply auto
    apply (case_tac "Y = (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>)")
    apply auto
    apply (case_tac "Z = (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>)")
    apply auto

    apply (drule out_context_inverse)
    apply auto
    apply (drule_tac out_context.Hash)
    apply force
    apply (drule factors_Xor_Hash)
    apply auto
    apply (case_tac "Y = (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>)")
    apply auto
    apply (case_tac "Z = (Hash \<lbrace>Nonce A NA, Agent A\<rbrace>)")
    apply auto
    done
next
  case (Fst Y Z)
  thus ?thesis by auto
next
  case (Snd Y Z)
  thus ?thesis by auto
next
  case (Decrypt K Y)
  thus ?thesis by auto
qed

lemma prover_step_1:
  "\<lbrakk> tr \<in> mdb;
     (t, Send (Tx (Honest P) k) COM [Number 1, Nonce (Honest P) NP]) \<in> set tr \<rbrakk>
   \<Longrightarrow> COM = Hash \<lbrace>Nonce (Honest P) NP, Agent (Honest P)\<rbrace>"
  apply (induct rule: mdb.induct)
  by auto

lemma prover_step_3_unique:
  assumes mdb:  "tr \<in> mdb"
  and     step: "(t, Send (Tx (Honest P) k) RESP  [Number 3, Nonce (Honest P) NP, NV]) \<in> set tr"
  and     step': "(t', Send (Tx (Honest P) k') RESP'  [Number 3, Nonce (Honest P) NP, NV']) \<in> set tr"
  shows   "NV = NV'"
  using mdb step step'
  apply (induct rule: mdb.induct)
  apply auto
done

lemma prover_step_3_unique_all:
  assumes mdb:  "tr \<in> mdb"
  and     step: "(t, Send (Tx (Honest P) k) RESP  [Number 3, Nonce (Honest P) NP, NV]) \<in> set tr"
  and     step': "(t', Send (Tx (Honest P) k') RESP'  [Number 3, Nonce (Honest P) NP, NV']) \<in> set tr"
  shows   "NV = NV' \<and> t = t' \<and> RESP = RESP' \<and> NV = NV' \<and> k = k'"
  using mdb step step'
  apply (induct rule: mdb.induct)
  apply auto
done

lemma verifier_claim_not_himself:
  assumes mdb:  "tr \<in> mdb"
  and     step: "(t, Claim (Honest V) \<lbrace>Agent P,d\<rbrace>) \<in> set tr"
  shows   "P \<noteq> Honest V"
  using mdb step
  apply (induct rule: mdb.induct)
  apply auto
done


lemma prover_step_3:
  assumes mdb:  "tr \<in> mdb"
  and     step: "(t, Send (Tx (Honest P) k) RESP  [Number 3, Nonce (Honest P) NP, NV]) \<in> set tr"
  shows   "RESP = (Xor NV (Nonce (Honest P) NP)) \<and>
           (\<exists> trecv. (trecv, Recv (Rec (Honest P)) NV) \<in> set
             (beforeEvent (t, Send (Tx (Honest P) k) (Xor NV (Nonce (Honest P) NP)) 
                                                     [Number 3, Nonce (Honest P) NP, NV]) tr))"
  using mdb step
proof (induct arbitrary: t P k RESP NV NP rule: mdb.induct)
  case Nil thus ?case by auto
next
  case (Fake tr t_fake X_fake I j_fake t P k RESP NV NP)
  thus ?case by (auto dest: prems(4))
next
  case (Con tr)
  thus ?case by (auto dest: prems(4))
next
  case (MD1  tr_ t_ P_ NP_ t P k RESP NV NP)
  thus ?case  by (auto dest: prems(4))
next
  case MD2
  thus ?case apply -
    apply clarsimp
    apply (elim disjE)
    apply force
    apply (drule prems(4))
    by auto
next
  case MD4
  thus ?case by (auto dest: prems(4))
next
  case MD5
  thus ?case by (auto dest: prems(4))
next
  case (MD3 tr tsend3 trec3 P3 NV3 tsend2_3 COM3 NP3 t P k RESP NV NP)
  let ?newev = "(tsend3,
        Send (Tu (Honest P3)) (Xor NV3 (Nonce (Honest P3) NP3)) [Number 3, Nonce (Honest P3) NP3, NV3])"
  let ?ev    = "(t, Send (Tx (Honest P) k) RESP [Number 3, Nonce (Honest P) NP, NV])"
  show ?case proof cases
    assume "?newev \<in> set tr"
    hence intr: "(t, Send (Tx (Honest P) k) RESP [Number 3, Nonce (Honest P) NP, NV]) \<in> set tr"
      using prems(8) apply -
      apply auto
      done
    show ?case proof cases
      assume "?newev = ?ev"
      have seteq: "set ((tsend3, Send (Tu (Honest P3)) (Xor NV3 (Nonce (Honest P3) NP3))
                                      [Number 3, Nonce (Honest P3) NP3, NV3]) # tr)
                   = set tr" using prems(10)
        apply force
        done
      thus ?case using prems(3-) apply -
        apply (simp (no_asm_use) add: seteq)
        apply (drule prems(4))
        apply (rule conjI)
        apply simp
        apply (intro impI conjI)
        apply force
        apply (elim exE)
        apply (drule beforeEvent_subset)
        apply force
        apply auto
        done
    next
      assume "?newev \<noteq> ?ev"
      hence before: "RESP = (Xor NV (Nonce (Honest P) NP)) \<Longrightarrow>
      beforeEvent (t, Send (Tx (Honest P) k) (Xor NV (Nonce (Honest P) NP))
                           [Number 3, Nonce (Honest P) NP, NV])
                  ((tsend3, Send (Tu (Honest P3)) (Xor NV3 (Nonce (Honest P3) NP3))
                                 [Number 3, Nonce (Honest P3) NP3, NV3]) # tr) =
      beforeEvent (t, Send (Tx (Honest P) k) (Xor NV (Nonce (Honest P) NP))
                                             [Number 3, Nonce (Honest P) NP, NV]) tr"
        apply auto
        done
      thus ?case using prems(3-) intr apply -
        apply (drule_tac t=t in prems(4))
        apply (rule conjI)
        apply force
        apply (elim conjE exE)
        apply (rule_tac x="trecv" in exI)
        apply (subst before)
        apply force
        apply assumption
        done
    qed
  next
    assume "?newev \<notin> set tr"
    show ?case proof cases
      assume "?newev = ?ev"
      thus ?case using prems(3-) apply -
        apply (rule conjI)
        apply force
        apply (rule_tac x="trec3" in exI)
        apply (subgoal_tac "NV=NV3") prefer 2
        apply force
        apply simp
        done
    next
      assume "?newev \<noteq> ?ev"
      thus ?case using prems(3-) apply -
         apply clarsimp
         apply (elim disjE conjE)
         apply simp
         apply (drule prems(4))
         apply auto
         done
     qed
   qed
qed

lemma out_context_componentsE_raw:
  "\<lbrakk> normed M; out_context (Nonce B NB) (Hash \<lbrace>Nonce B NB, Agent B\<rbrace>) X;
     X \<in> components {Abs_msg M} \<rbrakk>
   \<Longrightarrow> out_context (Nonce B NB) (Hash \<lbrace>Nonce B NB, Agent B\<rbrace>) (Abs_msg M)"
  apply (subgoal_tac "M \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (induct rule: normed.induct)
  apply (auto simp add: components_def Abs_msg_inverse)
  apply (subgoal_tac "Abs_msg (MPAIR a b) = MPair (Abs_msg a) (Abs_msg b)") prefer 2
  apply (simp add: MPair_def)
  apply simp
  apply (rule out_context.PairL) prefer 2
  apply force
  apply (subgoal_tac "\<exists>ma. Abs_msg m = Abs_msg ma \<and> ma \<in> fcomponents a") prefer 2
  apply force
  apply (subgoal_tac "a \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply auto

  apply (subgoal_tac "Abs_msg (MPAIR a b) = MPair (Abs_msg a) (Abs_msg b)") prefer 2
  apply (simp add: MPair_def)
  apply simp
  apply (rule out_context.PairR) prefer 2
  apply force
  apply (subgoal_tac "\<exists>ma. Abs_msg m = Abs_msg ma \<and> ma \<in> fcomponents b") prefer 2
  apply (force)
  apply (subgoal_tac "b \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply auto
  done

lemma out_context_componentsE:
  "\<lbrakk> out_context (Nonce B NB) (Hash \<lbrace>Nonce B NB, Agent B\<rbrace>) X;
     X \<in> components {M} \<rbrakk>
   \<Longrightarrow> out_context (Nonce B NB) (Hash \<lbrace>Nonce B NB, Agent B\<rbrace>) M"
  apply (subgoal_tac "normed (Rep_msg M) \<and> Abs_msg (Rep_msg M) = M")
  apply (elim conjE)
  apply (drule out_context_componentsE_raw)
  apply auto
  apply (simp add: Rep_msg_inverse)
  done

lemma out_context_componentsI_raw:
  "\<lbrakk> normed M; out_context (Nonce B NB) (Hash \<lbrace>Nonce B NB, Agent B\<rbrace>) (Abs_msg M) \<rbrakk>
   \<Longrightarrow> \<exists> X \<in> components {Abs_msg M}. out_context (Nonce B NB) (Hash \<lbrace>Nonce B NB, Agent B\<rbrace>) X"
  apply (subgoal_tac "M \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (induct rule: normed.induct)
  apply (auto simp add: components_def Abs_msg_inverse)
  apply (subgoal_tac "Abs_msg (MPAIR a b) = MPair (Abs_msg a) (Abs_msg b)") prefer 2
  apply (simp add: MPair_def)
  apply simp
  apply (drule out_context_inverse)
  apply auto
  apply (subgoal_tac "a \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply auto
  apply (subgoal_tac "b \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply auto
  done

lemma out_context_componentsI:
  "\<lbrakk> out_context (Nonce B NB) (Hash \<lbrace>Nonce B NB, Agent B\<rbrace>) M \<rbrakk>
   \<Longrightarrow> \<exists> X \<in> components {M}. out_context (Nonce B NB) (Hash \<lbrace>Nonce B NB, Agent B\<rbrace>) X"
  apply (subgoal_tac "normed (Rep_msg M) \<and> Abs_msg (Rep_msg M) = M")
  apply (elim conjE)
  apply (drule out_context_componentsI_raw)
  apply auto
  apply (simp add: Rep_msg_inverse)
  done

lemma nonce_use_outside:
  assumes mdb:       "tr \<in> mdb"
      and nonce:     "(tsend, Send (Tx (Honest B) k)
                              (Hash \<lbrace> Nonce (Honest B) NB, Agent (Honest B) \<rbrace>) [Number 1, Nonce (Honest B) NB])
                      \<in> set tr"
      and oev:       "oev \<in> set tr"
      and msg:       "oev = (t, Send (Tx A i) m L) \<or> oev = (t, Recv (Rx A i) m)"
      and outside:   "out_context (Nonce (Honest B) NB) (Hash \<lbrace>Nonce (Honest B) NB, Agent (Honest B)\<rbrace>) m"
  shows "\<exists> NV Y trep. 
           (((trep, Send (Tu (Honest B)) Y [Number 3, Nonce (Honest B) NB, NV])
              \<in> set (beforeEvent oev tr))
            \<or> (oev = (trep, Send (Tu (Honest B)) Y [Number 3, Nonce (Honest B) NB, NV])
               \<and> (trep, Send (Tu (Honest B)) Y [Number 3, Nonce (Honest B) NB, NV])
                 \<in> set tr))
            \<and> (t \<ge> trep + cdistl (Honest B) A)"
   using mdb nonce oev msg outside
proof (induct arbitrary: B k X NB A i m L t tsend oev)
  case Nil thus ?case by auto
next
  case (Con tr trecv_l M_l B_l j_l tab_l)
  let ?lastev = "(trecv_l, Recv (Rx B_l j_l) M_l)"
  let ?recvev = "(t, Recv (Rx A i) m)"
  let ?sendev = "(t, Send (Tx A i) m L)"

  show ?case proof cases
    assume "?sendev \<in> set tr \<and> ?sendev = oev"
    thus ?case using prems(5-) apply -
      apply clarsimp
      apply (drule prems(7)[where oev="?sendev"])
      apply (auto dest: beforeEvent_subset)
      done
  next
    assume "\<not> (?sendev \<in> set tr \<and> ?sendev = oev)"
    hence r_oev: "?recvev = oev" using prems by auto
    thus ?case proof cases
      assume "?recvev \<in> set tr"
      thus ?case using prems(5-) r_oev apply -
        apply clarsimp
        apply (drule  prems(7)[where oev="?recvev"])
        apply auto
        done
    next
      assume notintr: "?recvev \<notin> set tr"
      hence "?lastev = ?recvev" using prems by auto

      then obtain X where
        comp: "X \<in> components {M_l}" and
        out:  "out_context (Nonce (Honest B) NB) (Hash \<lbrace>Nonce (Honest B) NB, Agent (Honest B)\<rbrace>) X"
        using prems(6-) apply -
        apply clarsimp
        apply (drule_tac M="m" in out_context_componentsI)
        apply (elim bexE)
        apply auto
        done

      thus ?case using prems(6-13) notintr r_oev comp out apply -
        apply clarsimp
        apply (erule_tac x=X in ballE) prefer 2
        apply force
        apply (elim exE conjE bexE)
        apply (drule_tac oev="(tsend, Send (Tx A i) M' L)" in prems(7))
        apply force
        apply force
        apply (drule distort_LowHam)
        apply (elim bexE)
        apply simp
        apply (drule out_context_distort)
        apply simp
        apply (rule_tac X=Y in out_context_componentsE)
        apply simp
        apply simp
        apply (subgoal_tac "tab_l \<ge> cdistl A B_l") prefer 2
        apply (subgoal_tac "cdistM (Tx A i) (Rx B_l j_l) = None \<or> cdistl A B_l \<le> the (cdistM (Tx A i) (Rx B_l j_l))") prefer 2
        apply (rule noflt)
        apply clarsimp
        apply (elim exE)
        apply auto
        apply (rule_tac x=NV in exI)
        apply (rule_tac x=Ya in exI)
        apply (rule_tac x=trep in exI)
        apply (rule conjI)
        apply (force dest: beforeEvent_subset)
        apply (rule_tac y="trep + cdistl (Honest B) A + cdistl A B_l"
               in order_trans)
        apply (auto intro: cdistl_triangle)
        apply (rule_tac x=NV in exI)
        apply (rule_tac x=Ya in exI)
        apply (rule_tac x=trep in exI)
        apply auto
        done
    qed
  qed
next
  case (Fake tr t_l X_l I_l j_l B k NB A i m L t tsend oev)
  let ?lastev = "(t_l, Send (Tx (Intruder I_l) j_l) X_l [])"
  let ?sendev = "(t, Send (Tx A i) m L)"
  let ?recvev = "(t, Recv (Rx A i) m)"
  
  show ?case proof cases
    assume "?recvev = oev"
    thus ?case using prems(6-13) apply -
      apply clarsimp
      apply (drule prems(7)[where oev="?recvev"])
      apply (auto dest: beforeEvent_subset)
      done
  next
    assume "?recvev \<noteq> oev"
    hence "?sendev = oev" using prems by auto
    thus ?case proof cases
      assume "?sendev \<in> set tr"
      thus ?case using prems(6-) apply -
        apply auto
        apply (drule prems(7)[where oev="?sendev"])
        by auto
    next
      assume "?sendev \<notin> set tr"
      hence "?sendev = ?lastev" using prems by auto
      thus ?case using prems(6-) apply -
        apply auto
        apply (frule_tac A="Honest B" in outside_hash_deducible_implies_received)
        apply assumption
        apply force
        apply force
        apply (elim exE conjE)
        apply (drule_tac t=trs and m=X in prems(7)) prefer 2
        apply (rule disjI2) prefer 2
        apply assumption
        apply simp
        apply force
        apply force
        apply (elim exE)
        apply (rule_tac x=NV in exI)
        apply (rule_tac x=Y in exI)
        apply (rule_tac x=trep in exI)
        apply (rule conjI) defer
        apply (subgoal_tac "trs \<le> t_l")
        apply force
        apply (erule maxtime_geq_elem)
        apply force
        apply (auto dest: beforeEvent_subset)
        done
    qed
  qed
next
  case (MD5 tr tdone_l trec2_l V_l P_l NV_l NP_l trec1_l tsend_l CHAL_l
            B k NB A i m L t tsend oev)
  let ?recvev = "(t, Recv (Rx A i) m)"
  let ?sendev = "(t, Send (Tx A i) m L)"

  show ?case proof cases
    assume "?sendev = oev"
    thus ?case using prems(6-) apply -
      apply clarsimp
      apply (drule prems(7)[where oev="?sendev"])
      apply (auto dest: beforeEvent_subset)
      done
  next
    assume "?sendev \<noteq> oev"
    hence "?recvev \<in> set tr \<and> ?recvev = oev" using prems by auto
    thus ?case using prems(6-) apply -
      by (auto dest: prems(7)[where oev="?recvev"])
  qed
next
  case (MD1 tr t_l P_l NP_l B k NB A i m L t tsend oev)
  let ?lastev = " (t_l, Send (Tr (Honest P_l))
                                 (Hash \<lbrace>Nonce (Honest P_l) NP_l, Agent (Honest P_l)\<rbrace>)
                                 [Number 1, Nonce (Honest P_l) NP_l])"
  let ?sendev  = "(t, Send (Tx A i) m L)"
  let ?recvev  = "(t, Recv (Rx A i) m)"
  let ?nonceev = "(tsend,
                    Send (Tx (Honest B) k)
                         (Hash \<lbrace>Nonce (Honest B) NB, Agent (Honest B)\<rbrace>)
                         [Number 1, Nonce (Honest B) NB])"

  show ?case proof cases
    assume "?nonceev \<in> set tr"
    show ?case proof cases
      assume "?sendev \<in> set tr \<and> ?sendev = oev"
      thus ?case using prems(5-) apply -
        apply clarsimp
        apply (drule prems(7)[where oev="?sendev"])
        apply (auto dest: beforeEvent_subset)
        done
    next
      assume "\<not> (?sendev \<in> set tr \<and> ?sendev = oev)"
      show ?case proof cases
        assume "?sendev = oev"
        hence notr: "?sendev \<notin> set tr" using prems(3-) by auto
        show ?case proof cases
          assume seq: "?sendev = ?lastev"
          thus ?case using prems(5-) apply -
            apply auto
              (* Nonce not out of context in commit message *)
            apply (case_tac "Nonce (Honest P_l) NP_l = Nonce (Honest B) NB")
            apply auto
            apply (drule out_context_imp_subterms) back
            apply auto
            done
        next
          assume "?sendev \<noteq> ?lastev"
          thus ?case using prems(3-) notr by auto
        qed
      next
        assume "?sendev \<noteq> oev"
        hence  "?recvev \<in> set tr \<and> ?recvev = oev" using prems by auto
        thus ?case using prems(6-) apply -
          apply clarsimp
          apply (drule prems(7)[where oev="?recvev"])
          apply auto
          done
      qed
    qed
  next
    assume "?nonceev \<notin> set tr"
    hence lev: "?nonceev = ?lastev" using prems by auto
    show ?case proof cases
      assume "?sendev \<in> set tr \<and> ?sendev = oev"
      thus ?case using prems(5-) apply -
        apply auto
        apply (drule_tac t="t" and Y="Nonce (Honest P_l) NP_l"
               in Send_imp_parts_used)
        apply (rule out_context_imp_subterms)
        apply (auto simp add: mem_def)
        done
    next
      assume "\<not> (?sendev \<in> set tr \<and>  ?sendev = oev)"
      show ?case proof cases
        assume "?sendev = oev"
        hence notr: "?sendev \<notin> set tr"  using prems(3-) by auto
        show ?case proof cases
          assume seq: "?sendev = ?lastev"
          thus ?case using prems(5-) apply - by auto
        next
          assume "?sendev \<noteq> ?lastev"
          thus ?case using notr prems(3-) by auto
        qed
      next
        assume noev: "?sendev \<noteq> oev"
        hence  "?recvev \<in> set tr" and "?recvev = oev" using prems by auto
        thus ?case using prems(6-13) noev lev apply -
          apply (drule out_context_imp_subterms)
          apply (drule_tac nonce_components_subterm)
          apply (elim bexE)
          apply (drule_tac send_before_recv[simplified])
          apply assumption
          apply assumption
          apply (elim exE conjE bexE)
          apply (drule distort_LowHam)
          apply auto
          apply (drule nonce_not_LowHam)
          apply assumption
          apply (drule_tac Y="Y" in subterms_component_trans)
          apply simp
          apply (drule_tac t=tsend in Send_imp_parts_used)
          apply assumption
          apply (force simp add: mem_def)
          done
      qed
    qed
  qed
next
  case (MD2 tr t_l trec_l V_l COM_l NV_l B k NB A i m L t tsend oev)
  let ?lastev  = "(t_l, Send (Tr (Honest V_l)) (Nonce (Honest V_l) NV_l)
                             [Number 2, COM_l, Nonce (Honest V_l) NV_l])"
  let ?nonceev = "(tsend, Send (Tx (Honest B) k) (Hash \<lbrace>Nonce (Honest B) NB, Agent (Honest B)\<rbrace>)
                               [Number 1, Nonce (Honest B) NB])"
  let ?recvev  = "(t, Recv (Rx A i) m)"
  let ?sendev  = "(t, Send (Tx A i) m L)"

  show ?case proof cases
    assume "?sendev \<in> set tr \<and> oev = ?sendev"
    thus ?case using prems(6-) apply -
      apply clarsimp
      apply (drule prems(7)[where oev="?sendev"])
      apply assumption
      apply force
      apply force
      apply (auto dest: beforeEvent_subset)
      done
  next
    assume "\<not> (?sendev \<in> set tr \<and> oev = ?sendev)"
    show ?case proof cases
      assume "oev = ?sendev"
      hence notr: "?sendev \<notin> set tr" using prems(3-) by auto
      show ?case proof cases
        assume seq: "?sendev = ?lastev"
        have "Nonce (Honest V_l) NV_l = Nonce (Honest B) NB"
          using prems(6-14) seq notr apply -
          apply (case_tac "Nonce (Honest V_l) NV_l = Nonce (Honest B) NB")
          apply force
          apply (drule out_context_imp_subterms)
          apply (drule_tac Y="Nonce (Honest B) NB" in Send_imp_parts_used)
          apply auto
          done
        hence False using prems(6-14) seq notr apply -
          apply clarsimp
          apply (drule_tac Y="Nonce (Honest B) NB" in Send_imp_parts_used)
          apply force
          apply (auto simp add: mem_def)
          done
        thus ?case by auto
      next
        assume "?sendev \<noteq> ?lastev"
        thus ?case using notr prems(3-) by auto
      qed
    next
      assume "oev \<noteq> ?sendev"
      hence  "?recvev \<in> set tr \<and> oev = ?recvev" using prems by auto
      thus ?case using prems(6-14) apply -
        apply auto
        apply (drule prems(7)[where oev="?recvev"])
        apply auto
        done
      qed
  qed
next
  case (MD4 tr tsend_l trecv_l P_l NV_l t_l NP_l V_l B k NB A i m L t tsend oev)

  let ?lastev   = "(tsend_l, Send (Tr (Honest P_l))
                     (Crypt (priSK (Honest P_l))
                         \<lbrace>NV_l,
                           \<lbrace>Nonce (Honest P_l) NP_l, Agent V_l\<rbrace>\<rbrace>)
                     [])"
  let ?nonceev  = "(tsend,
                    Send (Tx (Honest B) k)
                      (Hash \<lbrace>Nonce (Honest B) NB, Agent (Honest B)\<rbrace>)
                      [Number 1, Nonce (Honest B) NB])"
  let ?sendev   = "(t, Send (Tx A i) m L)"
  let ?recvev   = "(t, Recv (Rx A i) m)"
  
  show ?case proof cases
    assume "?sendev \<in> set tr \<and> oev = ?sendev"
    thus ?case using prems(6-) apply -
      apply clarsimp
      apply (drule beforeEvent_subset prems(7)[where oev="?sendev"])
      apply auto
      done
  next
    assume "\<not> (?sendev \<in> set tr \<and> oev = ?sendev)"
    show ?case proof cases
      assume oev: "oev = ?sendev"
      hence notr: "?sendev \<notin> set tr" using prems by auto
      hence seq: "?sendev = ?lastev" using prems by auto
      show ?case proof cases
        assume neq: "Nonce (Honest P_l) NP_l = Nonce (Honest B) NB"
        thus ?case using prems(6-14) seq notr oev apply -
          apply (rule_tac x=NV_l in exI)
          apply (rule_tac x="(Xor NV_l (Nonce (Honest B) NB))" in exI)
          apply (rule_tac x=t_l in exI)
          apply (rule conjI)
          apply (rule disjI1) defer
          apply (subgoal_tac "cdistl (Honest B) (Honest B) = 0")
          apply (drule_tac t'=t_l in maxtime_geq_elem)
          apply force
          apply force
          apply (simp add: cdistl_def)
          apply (force intro: pdist_equal_zero)
          apply force
          done
      next
        assume nneq: "Nonce (Honest P_l) NP_l \<noteq> Nonce (Honest B) NB"
        thus ?case using prems(6-14) seq notr oev apply -
          apply auto
          apply (drule_tac t="trecv_l" and m="NV_l" in prems(7))
          apply auto defer

          apply (rule_tac x=NV in exI)
          apply (rule_tac x="Y"    in exI)
          apply (rule_tac x=trep in exI)
          apply (auto dest: beforeEvent_subset)
          apply (subgoal_tac "trecv_l \<le> tsend_l")
          apply auto
          apply (drule maxtime_geq_elem)
          apply auto
          apply (drule out_context_inverse, auto)+
          done
      qed
    next
      assume "oev \<noteq> ?sendev"
      hence "?recvev \<in> set tr \<and> oev = ?recvev" using prems by auto
      thus ?case using prems(6-14) apply -
        apply auto
        apply (drule prems(7)[where oev="?recvev"])
        apply auto
        done
    qed
  qed
next
  case (MD3 tr tsend_l trec_l P_l NV_l tsend2_l COM_l NP_l B k NB A i m L t tsend oev)

  let ?lastev   = "(tsend_l, Send (Tu (Honest P_l)) (Xor NV_l (Nonce (Honest P_l) NP_l))
                                  [Number 3, Nonce (Honest P_l) NP_l, NV_l])"
  let ?nonceev  = "(tsend, Send (Tx (Honest B) k) (Hash \<lbrace>Nonce (Honest B) NB, Agent (Honest B)\<rbrace>)
                           [Number 1, Nonce (Honest B) NB])"
  let ?sendev   = "(t, Send (Tx A i) m L)"
  let ?recvev   = "(t, Recv (Rx A i) m)"
  
  show ?case proof cases
    assume "?sendev \<in> set tr \<and> oev = ?sendev"
    thus ?case using prems(6-) apply -
      apply auto
      apply (drule prems(7)[where oev="?sendev"])
      apply force
      apply force
      apply force
      apply auto
      done
  next
    assume "\<not> (?sendev \<in> set tr \<and> oev = ?sendev)"
    show ?case proof cases
      assume oev: "oev = ?sendev"
      hence seq: "?sendev = ?lastev" using prems by auto
      show ?case proof cases
        assume neq: "Nonce (Honest P_l) NP_l = Nonce (Honest B) NB"
        thus ?case using prems(6-16) seq oev apply -
          apply auto
          apply (auto simp add: cdistl_def)
          apply (insert vc_pos)
          apply (simp add:  pdist_equal_zero)
          done
      next
        assume nneq: "Nonce (Honest P_l) NP_l \<noteq> Nonce (Honest B) NB"
        show ?case proof cases
          assume "Nonce (Honest B) NB \<in> factors NV_l"
          thus ?case using prems(6-14) seq oev apply -
            apply simp
            apply (drule_tac t="trec_l" and m="NV_l" in prems(7))
            apply force
            apply force prefer 2
            apply (elim exE conjE)
            apply (rule_tac x=NV in exI)
            apply (rule_tac x="Y"in exI)
            apply (rule_tac x=trep in exI)
            
            apply (rule conjI)
            apply (rule disjI1)
            apply (force dest: beforeEvent_subset)

            apply (subgoal_tac "trec_l \<le> tsend_l")
            apply force
            apply (drule maxtime_geq_elem)
            apply force
            apply force
            apply (case_tac "NV_l = Nonce (Honest B) NB")
            apply auto
            done
        next
          assume "Nonce (Honest B) NB \<notin> factors NV_l"
          thus ?case using prems(6-15) seq nneq oev apply -
            apply auto
            apply (frule factors_Xor_nonce_not_subterm)
            apply auto
            apply (drule out_context_inverse)
            apply auto
            
            apply (drule_tac t=trec_l in prems(7)) prefer 2
            apply (rule disjI2) prefer 2
            apply assumption
            apply force defer
            apply auto
            apply (rule_tac x=NV in exI)
            apply (rule_tac x=Y in exI)
            apply (rule_tac x=trep in exI)
            apply (auto dest: beforeEvent_subset)
            apply (subgoal_tac "trec_l \<le> tsend_l")
            apply auto
            apply (drule maxtime_geq_elem)
            apply auto
            apply (case_tac "NV_l = Nonce (Honest B) NB")
            apply auto
            
            apply (drule out_context_inverse)
            apply auto
            
            apply (drule factors_Xor_Nonce)
            apply auto

            apply (drule factors_Xor_Hash)
            apply auto
            apply (case_tac "NV_l = Hash X")
            apply auto
            
            apply (drule factors_Xor_Crypt)
            apply auto
            apply (case_tac "NV_l = Crypt k X")
            apply auto
            
            apply (drule factors_Xor_MPair)
            apply auto
            apply (case_tac "NV_l = \<lbrace>X,Y\<rbrace>")
            apply auto

            apply (drule factors_Xor_MPair)
            apply auto
            apply (case_tac "NV_l = \<lbrace>X,Y\<rbrace>")
            apply auto

            apply (drule factors_Xor)
            apply auto defer
            
            apply (drule out_context_inverse)
            apply auto
            
            apply (drule out_context_inverse)
            apply auto
            apply (case_tac "NV_l = X")
            apply auto
            done
        qed
      qed
    next
      assume "oev \<noteq> ?sendev"
      hence "?recvev \<in> set tr \<and> oev = ?recvev" using prems by auto
      thus ?case using prems(6-15) apply -
        apply auto
        apply (drule prems(7)[where oev="?recvev"])
        apply (auto dest: beforeEvent_subset)
      done
    qed
  qed
qed

lemma nonce_use_outside_tr:
    assumes mdb:       "tr \<in> mdb"
      and nonce:     "(tsend, Send (Tx (Honest B) k)
                              (Hash \<lbrace> Nonce (Honest B) NB, Agent (Honest B) \<rbrace>) [Number 1, Nonce (Honest B) NB])
                      \<in> set tr"
      and msg:       "(t, Send (Tx A i) m L) \<in> set tr \<or> (t, Recv (Rx A i) m) \<in> set tr"
      and outside:   "out_context (Nonce (Honest B) NB) (Hash \<lbrace>Nonce (Honest B) NB, Agent (Honest B)\<rbrace>) m"
  shows "\<exists> NV Y trep. (trep, Send (Tu (Honest B)) Y [Number 3, Nonce (Honest B) NB, NV])
                      \<in> set tr
                      \<and> (t \<ge> trep + cdistl (Honest B) A)"
  using mdb nonce msg outside apply -
  apply (elim disjE)
  apply (drule_tac oev=" (t, Send (Tx A i) m L)" in nonce_use_outside)
  apply assumption
  apply assumption
  apply force
  apply force
  apply (elim exE)
  apply auto prefer 2
  apply (drule_tac oev=" (t, Recv (Rx A i) m)" in nonce_use_outside)
  apply (auto dest: beforeEvent_subset)
done

lemma sig_msg_originates:
  assumes mdb: "tr \<in> mdb"
  and fsend: "(tf, Send (Tx (Honest P) j) mf Lf) \<in> set tr"
  and mfsubterm: "Crypt (priSK (Honest P)) \<lbrace>Nonce (Honest V) NV, \<lbrace>NP', Agent (Honest V)\<rbrace>\<rbrace>
                  \<in> subterms {mf}"
  and ffresh: "Crypt (priSK (Honest P)) \<lbrace>Nonce (Honest V) NV, \<lbrace>NP', Agent (Honest V)\<rbrace>\<rbrace>
               \<notin> used (beforeEvent (tf, Send (Tx (Honest P) j) mf Lf) tr)"
  shows "\<exists> NP. (NP' = Nonce (Honest P) NP)
               \<and> Lf = []
               \<and> mf = Crypt (priSK (Honest P)) \<lbrace>Nonce (Honest V) NV, \<lbrace>Nonce (Honest P) NP, Agent (Honest V)\<rbrace>\<rbrace>" using prems
proof (induct tr arbitrary: tf F j mf LF rule: mdb.induct)
  case (Fake tr mintr I tsend) with Fake.hyps prems show ?case by auto
next
  case (Con tr tc C mc D tab) with Con.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
next
  case (MD1 tr t_l P_l NP_l)
  show ?case using prems by auto
next
  case (MD2 tr t_l trec_l V_l COM_l NV_l)
  show ?case using prems by auto
next
  case (MD5 tr)
  thus ?case by auto
next
  case (MD4 tr tsend_l trecv_l P_l NV_l t_l NP_l V_l tf j mf)
  thus ?case
    apply auto
    apply (rule_tac x=NP_l in exI)
    apply auto
    apply (frule crypt_components_subterm)
    apply auto
    apply (drule_tac X="M" in send_before_recv[simplified])
    apply assumption
    apply assumption
    apply auto
    apply (drule distort_LowHam)
    apply auto
    apply (drule crypt_not_LowHam)
    apply simp
    apply (drule_tac Y=Y in subterms_component_trans)
    apply assumption
    apply (drule Send_imp_parts_used) back
    apply assumption
    apply (auto split: split_if_asm)
    done
next
  case (MD3 tr tsend_l trec_l P_l NV_l tsend2_l COM_l NP_l)
  thus ?case
    apply auto
    apply (drule subterms_Crypt_Xor)
    apply (drule subterms.singleton)
    apply auto
    apply (frule crypt_components_subterm)
    apply auto
    apply (drule_tac X="M" in send_before_recv[simplified])
    apply assumption
    apply assumption
    apply auto
    apply (drule distort_LowHam)
    apply auto
    apply (drule crypt_not_LowHam)
    apply simp
    apply (drule_tac Y=Y in subterms_component_trans)
    apply assumption
    apply (drule Send_imp_parts_used) back
    apply assumption
    apply (auto split: split_if_asm)
    done
qed

lemma originate_unique:
  assumes "m \<notin> used (beforeEvent (ta, Send TA ma La) tr)"
  and     "m \<notin> used (beforeEvent (tb, Send TB mb Lb) tr)"
  and     "(tb, Send TB mb Lb) \<noteq> (ta, Send TA ma La)"
  and     "(tb, Send TB mb Lb) \<in> set tr"
  and     "(ta, Send TA ma La) \<in> set tr"
  and     " m \<in> subterms {ma}"
  shows   "m \<notin> subterms {mb}" using prems
  apply (induct tr)
  apply simp
  apply (case_tac "a=(ta, Send TA ma La) \<and> a \<notin> set tr")
  apply (elim conjE)
  apply simp
  apply (case_tac "m \<in> subterms {mb}") prefer 2
  apply force
  apply (subgoal_tac "(tb, Send TB mb Lb) \<in> set tr") prefer 2
  apply force
  apply (frule_tac Y=m in Send_imp_parts_used)
  apply force
  apply force
  apply (case_tac "a=(tb, Send TB mb Lb) \<and> a \<notin> set tr")
  apply (elim conjE)
  apply simp
  apply (subgoal_tac "(ta, Send TA ma La) \<in> set tr") prefer 2
  apply force
  apply (frule_tac Y=m in Send_imp_parts_used)
  apply force
  apply force
  apply auto
done

lemma beforeEvent_not_equal:
  "\<lbrakk> a \<notin> set (beforeEvent b tr); a \<noteq> b; b \<in> set tr; a \<in> set tr \<rbrakk> \<Longrightarrow> b \<in> set (beforeEvent a tr)"
  apply (induct tr)
  apply (auto split: split_if_asm)
  done

lemma mdb_commit:
  assumes mdb: "tr \<in> mdb"
  and believe: "(tchal, Send (Tx (Honest V) j) CHAL [Number 2, COM, Nonce (Honest V) NV]) \<in> set tr"
  shows   "CHAL = Nonce (Honest V) NV \<and>
           (\<exists> trecv_com. (trecv_com, Recv (Rec (Honest V)) COM)
                         \<in> set (beforeEvent (tchal, Send (Tx (Honest V) j) (Nonce (Honest V) NV) [Number 2, COM, Nonce (Honest V) NV]) tr) 
                         \<and> (trecv_com \<le> tchal))" using prems
  apply (induct tr)
  apply force
  apply force
  apply force
  apply force defer defer
  apply force
  apply force
  apply clarsimp
  apply (elim disjE) prefer 2
  apply force prefer 2
  apply clarsimp
  apply (elim disjE) prefer 2
  apply force
  apply auto
  apply (drule maxtime_geq_elem)
  apply auto
done


lemma resp_implies_commit_send:
  assumes mdb: "tr \<in> mdb"
  and sign:    "(tresp, Send (Tx (Honest A) j) X [Number 3, Nonce (Honest A) NA, NV]) \<in> set tr"
  shows "(X = Xor NV (Nonce (Honest A) NA)) \<and>
         (\<exists> tcom.
             (tcom, Send (Tr (Honest A)) (Hash \<lbrace> Nonce (Honest A) NA, Agent (Honest A)\<rbrace>) [Number 1, Nonce (Honest A) NA]) \<in> set tr)"
  using prems
  apply (induct tr)
  apply auto
  apply (drule prover_step_1)
  apply auto
done

lemma sig_implies_commit_send:
  assumes mdb: "tr \<in> mdb"
  and sign:    "(tsig, Send (Tx (Honest A) j) (Crypt (priSK (Honest A)) \<lbrace>NV, \<lbrace> Nonce (Honest A) NA, Agent V\<rbrace>\<rbrace>) []) \<in> set tr"
  shows "\<exists> tcom.
             (tcom, Send (Tr (Honest A)) (Hash \<lbrace> Nonce (Honest A) NA, Agent (Honest A)\<rbrace>) [Number 1, Nonce (Honest A) NA]) \<in> set tr"
  using prems
  apply (induct tr)
  apply auto
  apply (drule resp_implies_commit_send)
  apply auto
  done


lemma sig_implies_fastrep_send:
  assumes mdb: "tr \<in> mdb"
  and sign:    "(tsig, Send (Tx (Honest A) j) (Crypt (priSK (Honest A)) \<lbrace>NV, \<lbrace> Nonce (Honest A) NA, Agent V\<rbrace>\<rbrace>) []) \<in> set tr"
  shows "\<exists> trep.
             (trep, Send (Tu (Honest A)) (Xor NV (Nonce (Honest A) NA))  [Number 3, Nonce (Honest A) NA, NV]) \<in> set tr"
  using prems
  apply (induct tr)
  apply auto
done

lemma verifier_NV_notin_factors_NP:
    assumes mdb: "tr \<in> mdb"
    and    believe: "(tchal, Send (Tx (Honest V) i) CHAL [Number 2, Hash \<lbrace>NP, Agent P \<rbrace>, Nonce (Honest V) NV]) \<in> set tr"
  shows "Nonce (Honest V) NV \<notin> factors NP" using prems
  apply (induct tr)
  apply auto
  apply (drule_tac X="Hash \<lbrace>NP, Agent P\<rbrace>" in send_before_recv[simplified])
  apply assumption
  apply force
  apply auto
  apply (drule distort_LowHam)
  apply (elim bexE)
  apply simp
  apply (subgoal_tac "Hash \<lbrace>NP, Agent P\<rbrace> \<in> subterms {Xor Y d}") prefer 2
  apply force
  apply (drule hash_not_LowHam)
  apply assumption
  apply (drule_tac Y=Y in subterms_component_trans)
  apply simp
  apply (drule factors_imp_subterms)
  apply (drule_tac G="{NP}" and H="{Hash \<lbrace> NP, Agent P \<rbrace>}" in subterms.trans)
  apply (simp (no_asm_use))
  apply force
  apply (drule_tac H="{M'}" and G="{Hash \<lbrace> NP, Agent P \<rbrace>}" in subterms.trans)
  apply simp
  apply (drule_tac Y="Nonce (Honest V) NV" in Send_imp_parts_used)
  apply assumption
  apply (auto simp add: mem_def)
  done

subsection {* Security proof for Honest Provers *}

lemma mdb_secure:
  assumes mdb: "tr \<in> mdb"
  and believe: "(tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>) \<in> set tr"
  shows   "d \<ge> pdist (Honest V) (Honest P)" using prems
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with Fake.hyps prems show ?case by (auto)
next
  case (Con tr tc C mc D tab)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
next
  case (MD2 tr)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr"
    by auto
  thus ?case using MD2.hyps prems by auto
next
  case (MD3 tr)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr"
    by auto
  with MD3.hyps prems show ?case by auto
next
  case (MD4 tr)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD4.hyps prems show ?case by auto
next
  case (MD1 tr)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD1.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
next
  -- {* the only nontrivial case since it adds Claim events *}
  case (MD5 tr tdone_l trec2_l V_l P_l NV_l NP_l trec1_l tsend_l CHAL_l)
  
  let ?lastev = "(tdone_l, Claim (Honest V_l) \<lbrace>Agent P_l, Real ((trec1_l - tsend_l) * vc / 2)\<rbrace>)"
  and ?claimev = "(tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)"
  show ?case proof cases
    -- {* the added event is the Claim event from the premise,
          the other case follows trivially from the IH *}
    assume "?lastev = ?claimev"
    hence Veq: "V_l=V" and Peq: "P_l=Honest P" and deq: "d=(trec1_l - tsend_l)*vc/2" by auto

    let ?NV = "Nonce (Honest V) NV_l"
    let ?msigned = "\<lbrace>?NV, \<lbrace>NP_l, Agent (Honest V)\<rbrace>\<rbrace>"
    let ?sigmsg = "Crypt (priSK (Honest P)) ?msigned" and
        ?commsg = "Hash \<lbrace> NP_l, Agent (Honest P)\<rbrace>"

    have NV_fresh:
      "?NV
       \<notin> usedI (beforeEvent (tsend_l, Send (Tr (Honest V)) CHAL_l [Number 2, ?commsg, ?NV]) tr)"
      using prems `tr \<in> mdb` apply -
      by (rule nonce_fresh_challenge, auto)
    -- {* The trivial case where V runs the protocol with himself is excluded *}
    show ?case proof cases
    assume PeqV: "P=V"
    thus ?case using prems
      apply -
      apply (drule verifier_claim_not_himself)
      apply auto
      done
  next
    assume PnotV: "P \<noteq> V"
    have sig_recv: "(trec2_l, Recv (Rec (Honest V)) ?sigmsg) \<in> set tr" using prems Veq Peq
      apply auto
      done

    have "?sigmsg \<in> components {?sigmsg}"
      by auto
    
    have "\<exists>A i tsend L M'.
      \<exists>Y\<in>components {M'}.
         (tsend, Send (Tx A i) M' L) \<in> set tr \<and>
           Xor ?sigmsg Y \<in> LowHamXor \<and> cdistM (Tx A i) (Rec (Honest V)) \<noteq> None
           \<and> tsend \<le> trec2_l - cdist (Tx A i)  (Rec (Honest V))"
      using prems Veq Peq
      apply -
      apply (rule send_before_recv)
      apply simp
      apply simp
      apply simp
      done

    then obtain
      E i tesend Le m' Y
      where p3: "Y \<in> components {m'}" and
            p2: "Xor ?sigmsg Y \<in> LowHamXor" and
            p1: "(tesend, Send (Tx E i) m' Le) \<in> set tr" and
(*                      "trec2_l >= cdist (Tx E i) (Rec (Honest V))" and
                      "cdistM (Tx E i) (Rec (Honest V)) \<noteq> None" and *)
            p4: "tesend \<le> trec2_l - cdist (Tx E i) (Rec (Honest V))"
      by auto
    hence "\<exists>tp mp j Lp.
             (tp, Send (Tx (Honest P) j) mp Lp) \<in> set tr
             \<and> (Crypt (priSK (Honest P)) ?msigned) \<in> subterms {mp}
             \<and> (Crypt (priSK (Honest P)) ?msigned)
               \<notin> used (beforeEvent (tp, Send (Tx (Honest P) j) mp Lp) tr)" using prems apply -
      apply (rule_tac tc=tesend and  msig="?msigned" in crypt_originates)
      apply force prefer 2
      apply assumption
      apply simp
      apply (subgoal_tac "?sigmsg \<in> subterms {Y}")
      apply (drule_tac Y=Y in subterms_component_trans)
      apply simp
      apply simp
      apply (drule distort_LowHam)
      apply auto
      apply (subgoal_tac "?sigmsg \<in> subterms {Xor Y d}") defer
      apply force
      apply (drule crypt_not_LowHam)
      apply assumption
      apply auto
      done
    then obtain tp mp j Lp where ftr: "(tp, Send (Tx (Honest P) j) mp Lp) \<in> set tr"
                             and mpsubterm: "(Crypt (priSK (Honest P)) ?msigned)
                                             \<in> subterms {mp}"
                             and ffresh: "(Crypt (priSK (Honest P)) ?msigned)
                                          \<notin> used (beforeEvent (tp, Send (Tx (Honest P) j) mp Lp) tr)"
      by auto
    hence ex: "\<exists> NPP. (NP_l = Nonce (Honest P) NPP)
               \<and> Lp = []
               \<and> mp = Crypt (priSK (Honest P))
                            \<lbrace>Nonce (Honest V) NV_l, \<lbrace>Nonce (Honest P) NPP, Agent (Honest V)\<rbrace>\<rbrace>" apply -
      apply (rule sig_msg_originates)
      apply (auto intro: prems )
      done
    then obtain NPP where ef:
      "(tp, Send (Tx (Honest P) j) (Crypt (priSK (Honest P))
                                     \<lbrace>Nonce (Honest V) NV_l, \<lbrace>Nonce (Honest P) NPP, Agent (Honest V)\<rbrace>\<rbrace>)
                                   []) \<in> set tr"
      and NPP: "NP_l = Nonce (Honest P) NPP"
      apply (insert ftr ex)
      by auto
    
    let ?fastmsg = "Xor (Nonce (Honest V) NV_l) (Nonce (Honest P) NPP)"
    let ?commsg = "Hash \<lbrace>Nonce (Honest P) NPP, Agent (Honest P)\<rbrace>"

    have fast_recv: "(trec1_l, Recv (Ru (Honest V)) ?fastmsg) \<in> set tr" using prems NPP
      by auto

    have chal_eq_ex : "CHAL_l = Nonce (Honest V) NV_l \<and>
             (\<exists> trecv_com.  (trecv_com, Recv (Rec (Honest V)) ?commsg )
                            \<in> set (beforeEvent (tsend_l, Send (Tr (Honest V)) (Nonce (Honest V) NV_l)
                                                           [Number 2, ?commsg, Nonce (Honest V) NV_l])
                                      tr) 
                            \<and> trecv_com \<le> tsend_l)" using prems NPP
      apply -
      apply (rule mdb_commit)
      apply auto
      done
    
    then obtain trecv_com where
      recv_com: "(trecv_com, Recv (Rec (Honest V)) ?commsg )
                 \<in> set (beforeEvent (tsend_l, Send (Tr (Honest V)) (Nonce (Honest V) NV_l)
                                                   [Number 2, ?commsg, Nonce (Honest V) NV_l]) tr)" and
      trecv_com_before: "trecv_com \<le> tsend_l"
      by auto
    
    have chal_eq : "CHAL_l = Nonce (Honest V) NV_l" using chal_eq_ex by auto
      
    obtain tcom where
      com_ev: "(tcom, Send (Tr (Honest P)) (Hash \<lbrace>Nonce (Honest P) NPP, Agent (Honest P)\<rbrace>) [Number 1, Nonce (Honest P) NPP])
               \<in> set tr"
      using ef `tr \<in> mdb` apply -
      apply (drule sig_implies_commit_send)
      apply auto
      done
    
    hence "\<exists> NV' Y trep.
             (trep, Send (Tu (Honest P)) Y [Number 3, Nonce (Honest P) NPP, NV']) \<in> set tr \<and>
             trep + cdistl (Honest P) (Honest V) \<le> trec1_l" using prems(3-)
      apply -
      apply (drule_tac t="trec1_l" in nonce_use_outside_tr)
      apply auto
      apply (subgoal_tac "Nonce (Honest P) NPP \<notin> factors (Nonce (Honest V) NV_l)")
      apply (drule factors_Xor_nonce_not_subterm)
      apply auto
      apply (erule contrapos_np) back back
      apply (rule_tac m="Nonce (Honest P) NPP" in out_context.Xor)
      apply force
      apply force
      apply force
      apply (erule contrapos_pp)
      apply simp
      apply (drule_tac f="factors" in HOL.arg_cong)
      apply auto
      done

    then obtain NV' Y trep where
      rep :       "(trep, Send (Tu (Honest P)) Y [Number 3, Nonce (Honest P) NPP, NV']) \<in> set tr" and
      trep_delay: "trep + cdistl (Honest P) (Honest V) \<le> trec1_l"
      by auto

    then obtain trep2 where
      "(trep2, Send (Tu (Honest P)) (Xor (Nonce (Honest V) NV_l)
                                    (Nonce (Honest P) NPP)) [Number 3, Nonce (Honest P) NPP, Nonce (Honest V) NV_l]) \<in> set tr"
      using ef prems(3-)
      apply auto
      apply (drule sig_implies_fastrep_send)
      apply auto
      done
    
    hence "NV' = Nonce (Honest V) NV_l" using prems(3-) rep
      apply -
      apply (rule prover_step_3_unique[where t=trep and t'=trep2 and P=P and NP=NPP and RESP=Y and NV=NV' and k=1 and
                                     RESP'="Xor (Nonce (Honest V) NV_l) (Nonce (Honest P) NPP)"
                                     and k'=1 and NV'="Nonce (Honest V) NV_l"])
      apply assumption      
      apply auto
      done
    
    hence "Y = Xor (Nonce (Honest V) NV_l) (Nonce (Honest P) NPP)" using `tr \<in> mdb` rep
      apply -
      apply simp
      apply (drule resp_implies_commit_send)
      apply auto
      done

    hence fast_send: "trep - tsend_l >= cdistl (Honest V) (Honest P)" using `tr \<in> mdb` PnotV
      apply -
      apply (erule_tac NA="NV_l" and i=0 and ma="Nonce (Honest V) NV_l"
             and mb="Xor (Nonce (Honest V) NV_l) (Nonce (Honest P) NPP)"
             in fresh_nonce_earliest_send[simplified])
      apply force defer
      apply force defer
      apply (insert prems(3-) chal_eq)
      apply simp
      apply simp
      apply (drule nonce_fresh_challenge)
      apply assumption
      apply (force simp add: usedI_def)
      apply (rule subterms_Nonce_Nonce)
      apply force
      done

    have "2* cdistl (Honest V) (Honest P) \<le> cdistl (Honest V) (Honest P) + cdistl (Honest P) (Honest V)"
      by (auto simp add: cdistl_symm)
    also have "... \<le> trep - tsend_l + (trec1_l - trep)" using fast_send trep_delay by auto
    also have "trep - tsend_l + (trec1_l - trep) \<le>  trec1_l - tsend_l" by auto
    finally have "cdistl (Honest V) (Honest P) * 2 \<le>  trec1_l - tsend_l" by auto
    thus ?thesis using deq
      apply (simp add: cdistl_def deq)
      apply (subgoal_tac "(pdist (Honest V) (Honest P) * 2 /vc) * vc \<le> (trec1_l - tsend_l) * vc") defer
      apply (rule mult_right_mono)
      apply force
      apply (insert vc_pos, auto split: split_if_asm)
      done
    qed
  next
    assume "?lastev \<noteq> ?claimev"
    show ?case using prems by auto
  qed
qed

subsection {* Security for dishonest Provers *}

lemma prover_NP_notin_factors_NV:
    assumes mdb: "tr \<in> mdb"
    and    believe: "(tresp, Send (Tx (Honest V) i) RESP [Number 3, Nonce (Honest P) NP, NV]) \<in> set tr"
  shows "Nonce (Honest P) NP \<notin> factors NV" using prems
  apply (induct tr)
  apply auto
  apply (frule prover_step_1)
  apply auto
  apply (drule nonce_use_outside_tr)
  apply assumption
  apply (rule disjI2)
  apply auto
  apply (case_tac "Nonce (Honest V) NP = NV")
  apply auto
  done

lemma steps_nonce_different:
  assumes
    mdb:  "tr \<in> mdb" and
    ev1: "(t1, Send (Tx (Honest A) i) (Nonce (Honest A) NA) [Number 2, COM, Nonce (Honest A) NA]) \<in> set tr" and
    ev2: "(t2, Send (Tx (Honest B) j) (Hash \<lbrace>Nonce (Honest B) NB, Agent (Honest B)\<rbrace>) [Number 1, Nonce (Honest B) NB]) \<in> set tr"
  shows "Nonce (Honest A) NA \<noteq> Nonce (Honest B) NB" using prems
  apply (induct tr)
  apply auto 
  apply (frule Send_imp_parts_used)
  apply auto
  apply (force simp add: mem_def)
  apply (frule_tac Y="(Nonce (Honest B) NB)"in Send_imp_parts_used)
  apply auto
  apply (force simp add: mem_def)
  done

(* MOVE *)
lemma not_before_itself:
  "e \<in> set (beforeEvent e tr) \<Longrightarrow> False"
  apply (induct tr)
  apply (auto split: split_if_asm)
  done

(* MOVE *)
lemma in_before_imp_eq:
  "a \<in> set (beforeEvent b tr) \<Longrightarrow>  beforeEvent a tr = beforeEvent a (beforeEvent b tr)"
  apply (induct tr)
  apply (auto dest: beforeEvent_subset)
  done


lemma cyclic:
  "\<lbrakk> rcom \<in> set tr; schal \<in> set tr; sresp \<in> set tr;
     rcom  \<in> set (beforeEvent schal tr);
     schal \<in> set (beforeEvent sresp tr);
     sresp \<in> set (beforeEvent rcom tr) \<rbrakk>
   \<Longrightarrow> False"
  apply (frule in_before_imp_eq)
  apply auto
  apply (frule in_before_imp_eq) back
  apply auto
  apply (drule beforeEvent_subset) back back
  apply (drule beforeEvent_subset) back back
  apply (drule not_before_itself)
  by auto

text {* We assume that the verifier cannot receive the signal sent on
        Tx V 0 on Rx V 1. This is required because there is a attack where
        a dishonest prover commits to 0 or dmsg otherwise. *}

definition
  rbe_receiver :: "agent \<Rightarrow> nat  \<Rightarrow> bool" where
  "rbe_receiver B j == (cdistM (Tx B 0) (Rx B j) = None)"

lemma honest_send:
  "\<lbrakk> tr \<in> mdb; (t, Send (Tx (Honest A) i) X L) \<in> set tr \<rbrakk>
   \<Longrightarrow>
     (\<exists> NA . i = 0
        \<and> X = Hash \<lbrace>Nonce (Honest A) NA, Agent (Honest A)\<rbrace>
        \<and> L = [Number 1, Nonce (Honest A) NA])
   \<or> (\<exists> NA COM. i = 0
        \<and> X = Nonce (Honest A) NA
        \<and> L = [Number 2, COM, Nonce (Honest A) NA])
   \<or> (\<exists> NV NA . i = 1
        \<and> X = Xor NV (Nonce (Honest A) NA)
        \<and> L = [Number 3, Nonce (Honest A) NA, NV])
   \<or> (\<exists> NV NA V. i = 0
        \<and> X = Crypt (priSK (Honest A)) \<lbrace>NV, \<lbrace>Nonce (Honest A) NA, Agent V\<rbrace>\<rbrace>
        \<and> L = [])"
  apply (induct tr rule: mdb.induct)
  apply auto
done
  
lemma mdb_secure_dishonest:
  assumes mdb: "tr \<in> mdb"
  and    not_recv: "rbe_receiver (Honest V) 1"
  and    believe: "(tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>) \<in> set tr"
  shows   "\<exists> P'. d \<ge> pdist (Honest V) (Intruder P')" using prems
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with Fake.hyps prems show ?case by (auto)
next
  case (Con tr tc C mc D tab)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
next
  case (MD1 tr t A NA)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD1.hyps prems show ?case by (auto)
next
  case (MD2 tr tsend trec B NA NB)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD2.hyps prems show ?case by (auto)
next
  case (MD3 tr tsend trec B NA tsend1 NB A)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD3.hyps prems show ?case by (auto)
next
  case (MD4 tr)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD4.hyps prems show ?case by (auto)
next
  -- {* the only nontrivial case since it adds Claim events *}
 case (MD5 tr tdone_l trec2_l V_l P_l NV_l NP_l trec1_l tsend_l CHAL_l)
 let ?lastev    = "(tdone_l, Claim (Honest V_l) \<lbrace>Agent P_l, Real ((trec1_l - tsend_l) * vc / 2)\<rbrace>)"
   and ?claimev = "(tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)"

  show ?case proof cases
    -- {* the added event is the Claim event from the premise,
          the other case follows trivially from the IH *}
    assume "?lastev = ?claimev"
    hence Veq: "V=V_l" and Beq: "P_l=Intruder P" and deq: "d=(trec1_l - tsend_l)*vc/2" by auto

    let ?commsg = "Hash \<lbrace>NP_l, Agent P_l\<rbrace>"
    let ?NV = "Nonce (Honest V) NV_l"

    have NV_fresh:
      "?NV \<notin> usedI (beforeEvent (tsend_l, Send (Tr (Honest V)) CHAL_l [Number 2, ?commsg, ?NV]) tr)"
      using prems(3-) Veq Beq deq
      apply -
      apply (drule nonce_fresh_challenge)
      apply auto
      done


    have chal_eq_ex : "CHAL_l = Nonce (Honest V) NV_l \<and>
             (\<exists> trecv_com.  (trecv_com, Recv (Rec (Honest V)) ?commsg )
                            \<in> set (beforeEvent (tsend_l, Send (Tr (Honest V)) (Nonce (Honest V) NV_l)
                                                           [Number 2, ?commsg, Nonce (Honest V) NV_l])
                                      tr) 
                            \<and> trecv_com \<le> tsend_l)" using prems
      apply -
      apply (rule mdb_commit)
      apply auto
      done
    
    then obtain trecv_com where
      recv_com: "(trecv_com, Recv (Rec (Honest V)) ?commsg )
                 \<in> set (beforeEvent (tsend_l, Send (Tr (Honest V)) (Nonce (Honest V) NV_l)
                                                   [Number 2, ?commsg, Nonce (Honest V) NV_l]) tr)" and
      trecv_com_before: "trecv_com \<le> tsend_l"
      by auto
    
    have chal_eq : "CHAL_l = Nonce (Honest V) NV_l" using chal_eq_ex by auto

    let ?RESP = "Xor (Nonce (Honest V_l) NV_l) NP_l"

    have NV_not: "?NV \<notin> factors NP_l" using prems(3-) NV_fresh apply -
      apply (rule verifier_NV_notin_factors_NP)
      apply auto
      done

    hence NV_RESP: "?NV \<in> factors ?RESP" using Veq
      apply -
      apply (drule factors_Xor_nonce_not_subterm)
      apply (simp add: Xor_comm)
      apply (elim disjE)
      apply auto
      done


    have "?RESP \<in> components {?RESP}"
      apply (subgoal_tac "\<forall> X Y. ?RESP \<noteq> \<lbrace> X , Y \<rbrace>")
      apply (drule components_non_pair)
      apply simp
      apply (subgoal_tac "?NV \<in> factors (?RESP)")
      apply auto
      apply (rule NV_RESP)
      done
    
    hence "\<exists>A i tsend L M'.
      \<exists>Y\<in>components {M'}.
         (tsend, Send (Tx A i) M' L) \<in> set tr \<and>
           Xor ?RESP Y \<in> LowHamXor \<and> cdistM (Tx A i) (Ru (Honest V)) \<noteq> None
           \<and> tsend \<le> trec1_l - cdist (Tx A i)  (Ru (Honest V))"
      using prems
      apply -
      apply (rule send_before_recv)
      apply simp
      apply simp
      apply simp
      done

    then obtain
      E i tesend Le RESP' Y
      where p3: "Y \<in> components {RESP'}" and
            p2: "Xor ?RESP Y \<in> LowHamXor" and
            p1: "(tesend, Send (Tx E i) RESP' Le) \<in> set tr" and
            p4: "tesend \<le> trec1_l - cdist (Tx E i) (Ru (Honest V))" and
            p5: "cdistM (Tx E i) (Ru (Honest V)) \<noteq> None"
      by auto

    have fast_not_send_himself: "i \<noteq> 0 \<or> E \<noteq> Honest V"
      using p5 not_recv apply -
      apply (auto simp add: rbe_receiver_def)
      done

    have rfactors: "?NV \<in> factors Y"using Veq p2 NV_RESP
      apply -
      apply (drule distort_LowHam)
      apply auto
      apply (drule factors_Xor)
      apply auto
      apply (frule factors_LowHam)
      apply auto
      done

    show ?case proof cases
      assume "\<exists> I. E = Intruder I"
      then obtain I where Eeq: "E = Intruder I" by auto
      
      hence fast_send: "tesend - tsend_l >= cdistl (Honest V) (Intruder I)" using `tr \<in> mdb`
        apply -
        apply (erule_tac NA="NV_l" and i=0 and ma="Nonce (Honest V) NV_l"
          and mb="RESP'"
          in fresh_nonce_earliest_send[simplified])
        apply force defer
        apply (insert prems(3-) chal_eq) defer defer
        apply simp
        apply simp
        apply (insert NV_fresh)
        apply (force simp add: usedI_def)
        apply force
        apply (insert p3)
        apply (rule_tac Y=Y in subterms_component_trans)
        apply (rule factors_imp_subterms)
        apply (rule rfactors)
        apply simp
        done

      from p4 p5 Eeq have r: "trec1_l - tesend >= cdistl (Intruder I) (Honest V)" apply -
        apply auto
        apply (auto simp add: cdist_def)
        apply (frule noflt_some2)
        apply auto
        done
      have "2* cdistl (Honest V) (Intruder I) \<le> cdistl (Honest V) (Intruder I) + cdistl (Intruder I) (Honest V)"
        by (auto simp add: cdistl_symm)
      also have "... \<le> tesend - tsend_l + (trec1_l - tesend)" using fast_send r apply -
        apply (rule ordered_ab_semigroup_add_class.add_mono)
        by auto
      also have "tesend - tsend_l + (trec1_l - tesend) \<le>  trec1_l - tsend_l" by auto
      finally have "cdistl (Honest V) (Intruder I) * 2 \<le>  trec1_l - tsend_l" by auto
      thus ?thesis using deq
        apply (simp add: cdistl_def deq)
        apply (subgoal_tac "(pdist (Honest V) (Intruder I) * 2 /vc) * vc \<le> (trec1_l - tsend_l) * vc") defer
        apply (rule mult_right_mono)
        apply force
        apply (insert vc_pos, auto split: split_if_asm)
      done
  next
    assume "\<not> (\<exists> I. E = Intruder I)"
    then obtain A where Eeq: "E = Honest A" apply -
      apply (case_tac E, auto)
      done

    show ?case proof cases
      assume asm: "\<exists>NA. i = 0 \<and>
                     RESP' = Hash \<lbrace>Nonce (Honest A) NA, Agent (Honest A)\<rbrace> \<and>
                     Le = [Number 1, Nonce (Honest A) NA]"
      hence "Y = RESP'" using p3 by auto
      thus ?case using asm rfactors by auto
   next
     assume n1: "\<not> (\<exists>NA. i = 0 \<and>
                     RESP' = Hash \<lbrace>Nonce (Honest A) NA, Agent (Honest A)\<rbrace> \<and>
                     Le = [Number 1, Nonce (Honest A) NA])"
   show ?case proof cases
     assume "\<exists>NA COM.
                i = 0 \<and>
                RESP'  = Nonce (Honest A) NA \<and>
                Le    = [Number 2, COM, Nonce (Honest A) NA]"
     then obtain NA COM where LeEq: "Le = [Number 2, COM, Nonce (Honest A) NA]"
       and Req: "RESP' = Nonce (Honest A) NA" and izero: "i=0"
       by auto

     hence Yeq: "Y = RESP'" using p3 by auto
      
     hence "Nonce (Honest A) NA = ?NV" using rfactors Req by auto
        
     thus ?case using fast_not_send_himself Eeq izero
       by auto
  next
    assume n2: "\<not> (\<exists>NA COM.
                   i = 0 \<and>
                   RESP' = Nonce (Honest A) NA \<and>
                   Le    = [Number 2, COM, Nonce (Honest A) NA])"
        
  show ?case proof cases
    assume
    "(\<exists>NV NA.
       i = 1 \<and>
       RESP' = Xor NV (Nonce (Honest A) NA) \<and>
       Le = [Number 3, Nonce (Honest A) NA, NV])"

    then obtain NV NA where LeEq: "Le = [Number 3, Nonce (Honest A) NA, NV]"
      and Req: "RESP' = Xor NV (Nonce (Honest A) NA)" by auto
      
    let ?NA = "Nonce (Honest A) NA"

    have NAneqNV: "?NV \<noteq> ?NA" using prems(3-) Req apply -
      apply (frule resp_implies_commit_send)
      apply simp
      apply (frule mdb_commit)
      apply simp
      apply (elim conjE exE)
      apply (rule steps_nonce_different)
      apply assumption
      apply auto
      done

    have facNV: "?NA \<notin> factors NV" using prems(3-) apply -
      apply (rule prover_NP_notin_factors_NV)
      apply auto
      done
    
    
    hence Yeq: "Y = RESP'" using Req p3 
      apply auto
      apply (subgoal_tac "\<forall> X Y. (Xor NV (Nonce (Honest A) NA)) \<noteq> \<lbrace> X , Y \<rbrace>")
      apply (drule components_non_pair)
      apply simp
      apply (subgoal_tac "?NA \<in> factors (Xor NV (Nonce (Honest A) NA))")
      apply auto
      apply (drule factors_Xor_nonce_not_subterm)
      apply auto
      done

    have facRESP': "?NA \<in> factors RESP'" using Req facNV rfactors apply -
      apply simp
      apply (drule factors_Xor_nonce_not_subterm)
      by auto

    hence facRESP: "?NA \<in> factors ?RESP" using Req p2 Yeq apply -
      apply auto
      apply (drule distort_LowHam)
      apply auto
      apply (subgoal_tac "Nonce (Honest A) NA \<in> factors (Xor (Xor NV d) (Nonce (Honest A) NA))")
      apply (simp add: Xor_rewrite)
      apply (subgoal_tac "Nonce (Honest A) NA \<notin> factors (Xor NV d)")
      apply (frule factors_Xor_nonce_not_subterm)
      apply auto
      apply (drule factors_Xor) back
      apply auto
      apply (insert facNV, force)
      apply (drule factors_LowHam)
      apply auto
      done

    hence "?NA \<in> factors NP_l" using  NAneqNV Veq apply -
      apply (drule factors_Xor)
      apply auto
      done

    hence out: "out_context ?NA (Hash \<lbrace>?NA, Agent (Honest A)\<rbrace>) (Hash \<lbrace>NP_l, Agent P_l\<rbrace>)" using prems
      apply auto
      apply (rule out_context.Hash)
      apply auto
      apply (rule out_context.PairL)
      apply auto
      apply (case_tac "NP_l = ?NA")
      apply auto
      done

    let ?rcom  = "(trecv_com, Recv (Rec (Honest V)) (Hash \<lbrace>NP_l, Agent P_l\<rbrace>))"
    let ?schal = "(tsend_l, Send (Tr (Honest V_l)) CHAL_l [Number 2, Hash \<lbrace>NP_l, Agent P_l\<rbrace>, Nonce (Honest V_l) NV_l])"
    let ?sresp = "(tesend, Send (Tx E i) RESP' Le)"
      
    have a: "?rcom  \<in> set (beforeEvent ?schal tr)" using prems chal_eq
      apply auto
      done

    have b: "?schal \<in> set (beforeEvent ?sresp tr)"using prems(3-) Yeq apply -
      apply (subgoal_tac "?sresp \<notin> set (beforeEvent ?schal tr)")
      apply (drule beforeEvent_not_equal)
      apply auto
      apply (drule nonce_fresh_challenge)
      apply assumption
      apply (auto simp add: usedI_def)
      apply (insert rfactors)
      apply (drule factors_imp_subterms)
      apply (subgoal_tac "Nonce (Honest V) NV_l \<in>
        used
            (beforeEvent
              (tsend_l, Send (Tr (Honest V)) CHAL_l [Number 2, Hash \<lbrace>NP_l, Agent (Intruder P)\<rbrace>, Nonce (Honest V) NV_l]) tr)")
      apply force
      apply (rule_tac L="[Number 3, Nonce (Honest A) NA, NV]" and
                      A="(Tx (Honest A) 1)" and t=tesend and X="RESP'" in Send_imp_used_parts)
      apply (insert Req Yeq, auto)
      done

    have c: "?sresp \<in> set (beforeEvent ?rcom  tr)" using prems(3-) out Beq apply -
      apply (frule_tac tresp=tesend in resp_implies_commit_send)
      apply force
      apply (elim exE conjE)
      apply (frule_tac oev="?rcom" and tsend=tcom in nonce_use_outside)
      apply (force dest: beforeEvent_subset)
      apply (force dest: beforeEvent_subset)
      apply (force dest: beforeEvent_subset)
      apply (simp)
      apply auto
      apply (drule_tac t=trep and t'=tesend in prover_step_3_unique_all)
      apply (auto dest: beforeEvent_subset)
      done

    then show ?case using prems(3-) a b c apply -
      apply (subgoal_tac "False")
      apply force
      apply (rule_tac rcom="?rcom" and schal="?schal" and sresp="?sresp"and tr=tr
             in cyclic)
      apply (auto dest: beforeEvent_subset)
      done
  next
    assume n3: "\<not> (\<exists>NV NA.
                    i = 1 \<and>
                    RESP' = Xor NV (Nonce (Honest A) NA) \<and>
                    Le = [Number 3, Nonce (Honest A) NA, NV])"
    hence asm: "(\<exists>NV NA V.
                 i = 0 \<and>
                 RESP' = Crypt (priSK (Honest A)) \<lbrace>NV, \<lbrace>Nonce (Honest A) NA, Agent V\<rbrace>\<rbrace> \<and>
                 Le = [])"using `tr \<in> mdb` n1 n2 p1 Eeq apply -
      apply simp
      apply (drule honest_send)
      apply auto
      done
    hence Yeq: "Y = RESP'" using p3 
      apply auto
      done

    thus ?case using asm rfactors by auto
  qed qed qed qed
next
  assume "?lastev \<noteq> ?claimev"
  show ?case using prems by auto
qed
next
  case Nil thus ?case by auto
qed

end
