(******************************************************************************

 ProtoVeriPhy

 Module: distance/MeadowsDB.thy

 Copyright (C) 2008-12 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
 License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Security Analysis of the two distance bounding protocols presented by Meadow et al. *}

theory MeadowsDBHash imports SystemCoffset SystemOrigination MessageTheoryXor2 begin

locale INITSTATE_SYMKEYS_NN = INITSTATE_SYMKEYS + INITSTATE_NONONCE

text {* These protocols have been presented in "Distance Bounding Protocols: Authentication Logic Analysis and Collusion Attacks " by Catherine Meadows, Radha Poovendran, Dusko Pavlovic, Liwu Chang, Paul Syverson *}

text {* Here we define the locale describing the necessary initial state
for the protocol, i.e., agents have pairwise shared keys and their
knowledge initially does not contain nonces of other agents. *}

definition
  initStateMd :: "agent \<Rightarrow> msg set" where
  "initStateMd A == Key`((symKey A)`UNIV)"

interpretation INITSTATE_SYMKEYS_NN Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor
               initStateMd Key
  apply (unfold_locales, auto simp add: initStateMd_def dest: injective_symKey)
  apply (drule subterms.singleton)
  apply (auto dest: injective_symKey)
  apply (drule subterms.singleton)
  apply (auto dest: injective_symKey)
  apply (drule subterms.singleton)
  apply (auto dest: injective_symKey simp add: MACM_def)
  apply (drule subterms.singleton)
  apply (auto dest: injective_symKey)
done

definition
  md1 :: "msg step"
 where
  "md1 tr V t =
     (UN NV. {ev. ev = (Nonce (Honest V) NV, SendEv 0 []) \<and>
                  Nonce (Honest V) NV \<notin> usedI tr})"

definition
  md2 :: "msg step"
 where
  "md2 tr P t =
     (UN NP NV trec.
        {ev. ev = ( Hash  \<lbrace> NV,  \<lbrace> Nonce (Honest P) NP , Agent (Honest P) \<rbrace>\<rbrace>
                  , SendEv 0 [NV,Nonce (Honest P) NP]) \<and>
             Nonce (Honest P) NP \<notin> usedI tr \<and>
             (trec, Recv (Rec (Honest P)) NV) \<in> set tr})"

definition
  md3 :: "msg step"
 where
  "md3 tr P t =
     (UN NP NV V tsend trec.
        {ev. ev = ( Hash[Key (symKey (Honest P) V)]
                   \<lbrace> NV, \<lbrace>Nonce (Honest P) NP,Agent (Honest P)\<rbrace>\<rbrace>
                  , SendEv 0 []) \<and>
             (trec, Recv (Rec (Honest P)) NV) \<in> set tr \<and>
             (tsend,
              Send (Tr (Honest P))
                (Hash \<lbrace> NV,  \<lbrace> Nonce (Honest P) NP , Agent (Honest P) \<rbrace>\<rbrace>)
                [NV,Nonce (Honest P) NP])
             \<in> set tr})"

definition
  md4 :: "msg step"
 where
  "md4 tr V t =
     (UN NP NV P trec1 trec2 tsend.
       {ev. ev = (\<lbrace>Agent P, Real ((trec1 - tsend) * vc/2)\<rbrace>, ClaimEv) \<and>
            (trec2, Recv (Rec (Honest V))
                  ( Hash[Key (symKey P (Honest V))]
                   \<lbrace> Nonce (Honest V) NV, \<lbrace> NP, Agent P\<rbrace>\<rbrace>)) \<in> set tr \<and>
            (trec1, Recv (Rec (Honest V)) (Hash  \<lbrace> Nonce (Honest V) NV,  \<lbrace> NP , Agent P \<rbrace>\<rbrace>)) \<in> set tr \<and>
            (tsend, Send (Tr (Honest V)) (Nonce (Honest V) NV) []) \<in> set tr})"

definition
  md_proto :: "msg proto" where
  "md_proto = {md1,md2,md3,md4}"

lemmas md_defs = md_proto_def md1_def md2_def md3_def md4_def


locale PROTOCOL_MD = PROTOCOL_SYMKEYS_NOKEYS+PROTOCOL_NONONCE+INITSTATE_SYMKEYS_NN

interpretation PROTOCOL_MD Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor initStateMd Key md_proto
  apply (unfold_locales)
  apply (auto simp add: md_defs messagesProtoTr_def messagesProtoTrHonest_def initStateMd_def MACM_def
              split: event.split split_if dest: parts.fst_set)
  apply (drule parts.singleton)
  apply auto
  apply (drule view_elem_ex)
  apply auto
  done

text {* Agents only look at their own views and all messages are derivable. *}

interpretation PROTOCOL_EXECUTABLE Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor initStateMd md_proto sys Key
  apply (unfold_locales)
  apply (auto simp add: md_defs initStateMd_def
                        messagesProto_def messagesProtoTrHonest_def MACM_def)

  apply (rule DM.Hash)
  apply (rule DM.MPair)
  apply (drule view_elem_ex)
  apply (erule exE)
  apply (drule Recv_imp_knows_A)
  apply force
  apply (rule DM.MPair) 
  apply force
  apply force

  apply (rule DM.MPair)
  apply (rule DM.Hash)
  apply (rule DM.MPair)
  apply force
  apply (rule DM.MPair)
  
  apply (drule view_elem_ex)
  apply (erule exE)
  apply (drule Recv_imp_knows_A)
  apply force

  apply (rule DM.MPair)
  apply force
  apply force
  apply (rule DM.MPair)
  apply (drule view_elem_ex)
  apply (erule exE)
  apply (drule Recv_imp_knows_A)
  apply force
  apply (rule DM.MPair)
  apply force
  apply force
  apply (rule DM.MPair)
  apply force
  apply force
  
  apply (rule_tac x=NV in exI)
  apply (auto simp add: nonce_view_fresh [simplified md_proto_def]
                        nonce_view_used [simplified md_proto_def])
  apply (auto simp add: recv_a_view_a_r send_a_view_a_r)
  apply (rule_tac x=NP in exI)

  apply (rule_tac x=NV in exI)
  apply (rule_tac x=P in exI)
  apply (rule_tac x=trec1 in exI)
  apply (rule_tac x=trec2 in exI)
  apply (rule_tac x=tsend in exI)
  apply auto
  apply (rule_tac x=NP in exI)
  apply (rule_tac x=NV in exI)
  apply (rule_tac x=P in exI)
  apply (rule_tac x=trec1 in exI)
  apply (rule_tac x=trec2 in exI)
  apply (rule_tac x=tsend in exI)
  apply auto
done

text {* Agent behaviour does not change with constant clock errors. *}

interpretation  PROTOCOL_DELTAONLY Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor initStateMd Key md_proto
  apply unfold_locales
  apply (auto simp add: md_defs in_timetrans)
  apply (rule_tac x=NP in exI)
  apply (rule conjI)
  apply force
  apply force
  apply (rule_tac x=NP in exI) apply (rule_tac x=NV in exI)
  apply force
  apply (intro exI conjI, auto)
  apply (rule_tac f="MPair (Agent P)" in HOL.arg_cong) (* XXX: add intro rule for this *)
  apply (rule_tac f="Real" in HOL.arg_cong)
  apply force
  apply (intro exI)
  apply auto
  apply (rule_tac x="trec + coffset A" in exI, force)
  apply (intro exI)
  apply auto
  apply (rule_tac x="trec + coffset A" in exI, force)
  apply (rule_tac x="tsend + coffset A" in exI, force)
  apply (rule exI, rule exI, rule exI)
  apply (rule_tac x="trec1 + coffset A" in exI,
         rule_tac x="trec2 + coffset A" in exI,
         rule_tac x="tsend + coffset A" in exI)
  apply auto
  apply (rule_tac f="MPair (Agent P)" in HOL.arg_cong) (* XXX: add intro rule for this *)
  apply (rule_tac f="Real" in HOL.arg_cong)
  apply auto
done

interpretation PROTOCOL_DELTA_EXEC Crypt Nonce MPair Hash Number
                                   parts subterms DM LowHamXor Xor
                                   initStateMd Key md_proto sys
by unfold_locales

subsection {* Direct Definition for Meadows Distance Bounding Protocol *}

inductive_set
  mdb :: "(msg trace) set"
 where
   Nil [intro] : "[] \<in> mdb"
 | Fake: 
   "\<lbrakk> tr \<in> mdb; t >=  maxtime tr;
      X \<in> DM (Intruder I) (knowsI (Intruder I) tr) \<rbrakk>
   \<Longrightarrow> (t, Send (Tx (Intruder I) j) X [])  # tr \<in> mdb"

 | Con : 
   "\<lbrakk> tr \<in> mdb; trecv >= maxtime tr;
      (tsend, Send (Tx A i) m L) \<in> set tr;
      cdistM (Tx A i) (Rx B j) = Some tab;
      trecv \<ge> tsend + tab;
      d \<in> LowHamXor \<rbrakk>
    \<Longrightarrow> (trecv, Recv (Rx B j) (Xor m d)) # tr \<in> mdb"

 (* Protocol rules for Meadows DB *)
 | MD1:  
   "\<lbrakk> tr \<in> mdb; t >=  maxtime tr;
      \<not> (used tr (Nonce (Honest V) NV)) \<rbrakk>
   \<Longrightarrow> (t, Send (Tr (Honest V)) (Nonce (Honest V) NV) []) # tr \<in> mdb"

 | MD2:
   "\<lbrakk> tr \<in> mdb; tsend >=  maxtime tr;
      (trec, Recv (Rec (Honest P)) NV) \<in> set tr;
      \<not> (used tr (Nonce (Honest P) NP)) \<rbrakk>
   \<Longrightarrow> (tsend, Send (Tr (Honest P))
                   (Hash \<lbrace> NV,  \<lbrace> Nonce (Honest P) NP, Agent (Honest P)\<rbrace>\<rbrace>  )
                   [NV, Nonce (Honest P) NP])
      # tr \<in> mdb"

 | MD3:  
   "\<lbrakk> tr \<in> mdb; tsend >=  maxtime tr;
      (trec, Recv (Rec (Honest P)) NV) \<in> set tr;
     (tsend1, Send (Tr (Honest P))
                   (Hash \<lbrace> NV, \<lbrace> Nonce (Honest P) NP, Agent (Honest P)\<rbrace>\<rbrace> )
                   [NV, Nonce (Honest P) NP])
     \<in> set tr \<rbrakk>
   \<Longrightarrow> (tsend,
       Send (Tr (Honest P))
            Hash[Key (symKey (Honest P) V)]
                   \<lbrace> NV, \<lbrace> Nonce (Honest P) NP, Agent (Honest P)\<rbrace>\<rbrace> [])
       # tr \<in> mdb"

 | MD4:
   "\<lbrakk> tr \<in> mdb; tdone \<ge> maxtime tr;
      (trec2, Recv (Rec (Honest V))
                  ( Hash[Key (symKey P (Honest V))]
                   \<lbrace> Nonce (Honest V) NV, \<lbrace> NP, Agent P\<rbrace>\<rbrace>))
      \<in> set tr;
      (trec1, Recv (Rec (Honest V)) (Hash \<lbrace> Nonce (Honest V) NV, \<lbrace> NP,  Agent P\<rbrace> \<rbrace>))
      \<in> set tr;
      (tsend, Send (Tr (Honest V)) (Nonce (Honest V) NV) []) \<in> set tr \<rbrakk>
  \<Longrightarrow> (tdone, Claim (Honest V) \<lbrace>Agent P, Real ((trec1 - tsend) * vc/2)\<rbrace>) # tr
     \<in> mdb"

text {* obtain a simpler induction rule for protocol since it is executable and deltaonly *}

lemmas proto_induct =
  sys.induct [simplified derivable_removable remove_occursAt timetrans_removable]

subsection {* Equality for direct and parameterized Definition *}

text {* We now show that both inductive definitions define the
        same set of traces. *}

lemma abstr_equal: "mdb = sys"
proof auto
  fix tr
  assume r: "tr \<in> sys"
  show "tr \<in> mdb" using r
  proof (induct tr rule: proto_induct)
    case 1 with prems show ?case by auto
  next
    case 2 with prems show ?case by (auto intro: mdb.Nil)
  next
    case 4 with prems show ?case by (auto intro: mdb.Con)
  next
    case 3 with prems show ?case by (auto intro: mdb.Fake)
  next
    case 5
    thus ?case
      apply (auto simp add: md_defs)
      apply (auto intro!: mdb.MD1 mdb.MD2 mdb.MD3 [simplified] mdb.MD4 simp add: usedI_def)
      apply (auto simp add: mem_def)
      done
  qed
next
  fix tr
  assume r: "tr \<in> mdb"
  show "tr \<in> sys" using r
  proof(induct tr rule: mdb.induct)
    case Nil
    with prems show ?case by auto
  next
    case (Fake tr ts X I j)
    with prems show ?case by (auto intro: sys.Fake)
  next
    case (Con tr tsend C i X D j dCD trecv)
    with prems show ?case by (auto intro: sys.Con)
  next
    case (MD1 tr ts C NA)
    with prems have "(ts,createEv C (SendEv 0 []) (Nonce (Honest C) NA)) # tr \<in> sys"
      apply -
      apply (rule_tac step=md1 in sys_Proto_exec)
      apply force
      apply force
      apply force
      apply (force simp add: md_proto_def)
      apply (simp add: md1_def)
      apply (simp add: usedI_def)
      apply (auto simp add: mem_def)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD2 tr tsend trecv P NV NP)
    with prems have
      "(tsend, 
        createEv P
                 (SendEv 0 [NV, Nonce (Honest P) NP])
                 (Hash \<lbrace> NV, \<lbrace> Nonce (Honest P) NP, Agent (Honest P)\<rbrace>\<rbrace>  ))
       # tr \<in> sys"
      apply - apply (rule_tac step=md2 in sys_Proto)
      apply (auto simp add: md_defs usedI_def)
      apply (auto simp add: mem_def)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD3 tr tsend trecv P NV tsend1 NP V)
    with prems have
      "(tsend,
        createEv P (SendEv 0 [])
                   (Hash[Key (symKey (Honest P) V)]
                        \<lbrace>NV, \<lbrace>Nonce (Honest P) NP, Agent (Honest P)\<rbrace>\<rbrace>)) # tr \<in> sys"
      apply - apply (rule_tac step=md3 in sys_Proto)
      apply (auto simp add: md_defs)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD4 tr tdone trec2 V P NV NP trec1 tsend)
    with prems have
      "(tdone, createEv V ClaimEv \<lbrace>Agent P, Real ((trec1 - tsend) * vc/2)\<rbrace>) # tr \<in> sys"
      apply - apply (rule_tac step=md4 in sys_Proto)
      apply (auto simp add: md_defs)
      apply (intro exI conjI)
      apply auto
      done
    thus ?case by (auto simp add: createEv.psimps)
  qed
qed

lemmas [simp,intro] = abstr_equal [THEN sym]

subsection {* Some invariants capturing the Behavior of honest Agents *}

lemma nonce_fresh_challenge:
  assumes mdb: "tr \<in> mdb" and
          send: "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) []) \<in> set tr"
  shows   "Nonce (Honest A) NA
           \<notin> usedI (beforeEvent (ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) []) tr)"
  using prems(1-)
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend) with Fake.hyps prems show ?case by auto
next
  case (Con tr tc C mc D tab) with Con.hyps prems show ?case by auto
next
  case (MD1 tr t V' NV)
  show ?case proof cases
    assume "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) []) =
            (t, Send (Tr (Honest V')) (Nonce (Honest V') NV) [])"
    thus ?case using MD1.hyps prems
      apply (auto simp add: usedI_def)
      by (simp add: mem_def)
  next
    assume "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) []) \<noteq>
            (t, Send (Tr (Honest V')) (Nonce (Honest V') NV)  [])"
    thus ?case using MD1.hyps prems by auto
  qed
next
  case (MD2 tr tsend trec P' NV NP)
  thus ?case using MD2.hyps prems by auto
next
  case (MD3 tr tsend trec P' NV tsend1 NP V')
  with MD3.hyps prems show ?case by auto
next
  case (MD4 tr t trec2 V' P' NV NP trec1 tsend)
  with MD4.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
qed

lemma nonce_fresh_response:
  assumes mdb: "tr \<in> mdb" and
          send: "(ta, Send (Tx (Honest A) i) (Hash \<lbrace> NV, \<lbrace> NP, Agent P \<rbrace> \<rbrace>)
                                             [NV, NP]) \<in> set tr"
  shows   "
           (\<exists> NA.
             P = Honest A \<and>
             NP = Nonce (Honest A) NA \<and>
             Nonce (Honest A) NA
             \<notin> usedI (beforeEvent
                        (ta, Send (Tx (Honest A) i) (Hash \<lbrace> NV, \<lbrace> Nonce (Honest A) NA,  Agent (Honest A) \<rbrace>\<rbrace>)
                                                    [NV, Nonce (Honest A) NA]) tr))"
  using mdb send
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend) with Fake.hyps prems show ?case by auto
next
  case (Con tr tc C mc D tab) with Con.hyps prems show ?case by auto
next
  case (MD3 tr tsend trec P' NV tsend1 NP V')
  with MD3.hyps prems show ?case by auto
next
  case (MD4 tr t trec2 V' P' NV NP trec1 tsend)
  with MD4.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
next
  case (MD1 tr t V NV) thus ?case by auto
next
  case (MD2 tr tsend trec C ND NC)
  let ?eva = "(ta, Send (Tx (Honest A) i) (Hash \<lbrace>NV, \<lbrace>NP,  Agent P \<rbrace>\<rbrace>) [NV, NP])"
  let ?newev = "(tsend, Send (Tr (Honest C)) (Hash \<lbrace> ND, \<lbrace> Nonce (Honest C) NC, Agent (Honest C)\<rbrace>\<rbrace>)
                                             [ND, Nonce (Honest C) NC])"
  show ?case proof cases
    assume eq: "?eva = ?newev"
    thus ?case using MD2.hyps prems apply -
      apply (rule_tac x=NC in exI)
      apply (subgoal_tac "NP = Nonce (Honest A) NC") prefer 2
      apply force
      apply (subgoal_tac "P = Honest A")
      prefer 2
      apply force
      apply simp

      apply (simp add: usedI_def)
      apply (auto simp add: mem_def)
      done
  next
    assume "?eva \<noteq> ?newev"
    hence "?eva \<in> set tr" using `?eva \<in> set (?newev#tr)` by auto
    thus ?case apply -
      apply (frule MD2.hyps(2))
      apply (elim conjE exE)
      apply auto
      done
  qed
qed

lemma nonce_fresh_response2:
  assumes mdb: "tr \<in> mdb" and
          send: "(ta, Send (Tx (Honest A) i) (Hash \<lbrace>NV, \<lbrace> Nonce (Honest A) NA, Agent (Honest A)\<rbrace>\<rbrace>)
                                                    [NV, Nonce (Honest A) NA])
                 \<in> set tr"
  shows   "Nonce (Honest A) NA
           \<notin> usedI (beforeEvent
                     (ta, Send (Tx (Honest A) i) (Hash \<lbrace> NV , \<lbrace> Nonce (Honest A) NA, Agent (Honest A)\<rbrace>\<rbrace>)
                                                 [NV, Nonce (Honest A) NA]) tr)"
  using mdb send
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend) with Fake.hyps prems show ?case by auto
next
  case (Con tr tc C mc D tab) with Con.hyps prems show ?case by auto
next
  case (MD3 tr tsend trec P' NV tsend1 NP V')
  with MD3.hyps prems show ?case by auto
next
  case (MD4 tr t trec2 V' P' NV NP trec1 tsend)
  with MD4.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
next
  case (MD1 tr t V NV) thus ?case by auto
next
  case (MD2 tr tsend trec C ND NC)
  let ?eva = "(ta, Send (Tx (Honest A) i) (Hash \<lbrace>NV, \<lbrace> Nonce (Honest A) NA,  Agent (Honest A)\<rbrace>\<rbrace>)
                                                   [NV, Nonce (Honest A) NA])"
  let ?newev = "(tsend, Send (Tr (Honest C)) (Hash \<lbrace> ND, \<lbrace> Nonce (Honest C) NC, Agent (Honest C) \<rbrace>\<rbrace>)
                                             [ND, Nonce (Honest C) NC])"
  show ?case proof cases
    assume eq: "?eva = ?newev"
    thus ?case using MD2.hyps prems apply -
      apply (auto simp add: usedI_def)
      apply (auto simp add: mem_def)
      done
  next
    assume "?eva \<noteq> ?newev"
    hence "?eva \<in> set tr" using `?eva \<in> set (?newev#tr)` by auto
    thus ?case apply -
      apply (frule MD2.hyps(2))
      apply auto
      done
  qed
qed


text {* If an honest prover sends a MAC, then he has sent the corresponding fastreply before.
        Then we can use nonce_fresh_response to obtain that the nonce in a fast-reply is
        fresh. *}

lemma MAC_send_prover:
  assumes mdb: "tr \<in> mdb"
      and mac: "(tsend,
                 Send (Tx (Honest B) k)
                      (Hash[Key (symKey (Honest B) A)]
                       \<lbrace> NA, \<lbrace>Nonce (Honest B) NB, Agent (Honest B)\<rbrace>\<rbrace>) [])
                \<in> set tr"
  shows "(\<exists> tfast. 
            (tfast, Send (Tr (Honest B))
                         (Hash \<lbrace> NA, \<lbrace> Nonce (Honest B) NB, Agent (Honest B) \<rbrace>\<rbrace>)
                         [NA,Nonce (Honest B) NB]) \<in> set tr)"
  using prems
  apply (induct tr rule: mdb.induct)
  apply (auto simp add: MACM_def)
done

lemma MAC_send_prover2:
  assumes mdb: "tr \<in> mdb"
      and mac: "(tsend,
                 Send (Tx (Honest B) k)
                      (Hash[Key (symKey (Honest B) A)]
                       \<lbrace> NA, \<lbrace>Nonce (Honest B) NB, Agent (Honest B)\<rbrace>\<rbrace>) [])
                \<in> set tr"
  shows "(\<exists> tfast. 
            (tfast, Send (Tr (Honest B))
                         (Hash \<lbrace> NA , \<lbrace>Nonce (Honest B) NB,  Agent (Honest B) \<rbrace>\<rbrace>)
                         [NA,Nonce (Honest B) NB]) \<in> set tr \<and>
            Nonce (Honest B) NB
            \<notin> usedI (beforeEvent
                      (tfast, Send (Tr (Honest B))
                         (Hash \<lbrace> NA , \<lbrace> Nonce (Honest B) NB,  Agent (Honest B) \<rbrace>\<rbrace>)
                         [NA,Nonce (Honest B) NB]) tr))"
  using prems apply -
  apply (frule MAC_send_prover)
  apply force
  apply (elim exE)
  apply (rule_tac x=tfast in exI)
  apply (frule nonce_fresh_response2)
by auto

text {* The macs are always unique because they contain the symmetric key
        of two honest agents and his own nonce contribution *}

lemma mac_msg_originates:
  assumes mdb: "tr \<in> mdb"
  and fsend "(tf, Send (Tx (Honest F) j) mf Lf) \<in> set tr"
  and mfsubterm: "Hash \<lbrace>Key (symKey (Honest P) (Honest V)), \<lbrace>Nonce (Honest V) NV, \<lbrace>NP', Agent (Honest P)\<rbrace>\<rbrace>\<rbrace>
                  \<in> subterms {mf}"
  and ffresh: "Hash \<lbrace>Key (symKey (Honest P) (Honest V)), \<lbrace>Nonce (Honest V) NV, \<lbrace>NP', Agent (Honest P)\<rbrace>\<rbrace>\<rbrace>
               \<notin> used (beforeEvent (tf, Send (Tx (Honest F) j) mf Lf) tr)"
  shows "\<exists> NP. F=P \<and> (NP' = Nonce (Honest P) NP)
               \<and> Lf = []
               \<and> mf = Hash[Key (symKey (Honest P) (Honest V))]
                          \<lbrace>Nonce (Honest V) NV, \<lbrace>Nonce (Honest P) NP, Agent (Honest P)\<rbrace>\<rbrace>" using prems
proof (induct tr rule: mdb.induct)
  case (Fake tr mintr I tsend) with Fake.hyps prems show ?case by auto
next
  case (Con tr tc C mc D tab) with Con.hyps prems show ?case by auto
next
  case (MD4 tr t trec2 V' P' NV NP trec1 tsend)
  with MD4.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
next
  case (MD1 tr t V NV)
  show ?case using prems by auto
next
  case (MD2 tr tsend trec P' NV' NP)
  let ?msg = "Hash \<lbrace> NV', \<lbrace> Nonce (Honest P') NP,  Agent (Honest P') \<rbrace>\<rbrace>"
  let ?ev = "(tsend, Send (Tr (Honest P')) ?msg [NV',Nonce (Honest P') NP])"
  let ?mac = "Hash \<lbrace>Key (symKey (Honest P) (Honest V)), \<lbrace>Nonce (Honest V) NV, \<lbrace>NP', Agent (Honest P)\<rbrace>\<rbrace>\<rbrace>"
  show ?case proof cases
    assume "(tf, Send (Tx (Honest F) j) mf Lf) = ?ev"
    hence sub: "?mac \<in> subterms {NV'}" using prems by auto
    thus ?thesis
    proof cases
      assume "?ev \<in> set tr"
      thus ?thesis using prems(6-) by auto
    next
      assume "?ev \<notin> set tr"
      thus ?thesis using sub prems(6-) apply -
        apply simp
        apply (drule_tac m="NV'" in send_before_recv[simplified])
        apply assumption
        apply (elim exE conjE)
        apply simp
        apply (drule hash_not_LowHam)
        apply simp
        apply (drule Send_imp_parts_used) 
        apply assumption
        apply simp
        done
    qed
  next
    assume "(tf, Send (Tx (Honest F) j) mf Lf) \<noteq> ?ev"
    thus ?thesis using prems by auto
  qed
next
  case (MD3 tr tsend trec P' NV' tsend1 NP V')
  let ?msg = "Hash[Key (symKey (Honest P') V')] \<lbrace>NV', \<lbrace>Nonce (Honest P') NP, Agent (Honest P')\<rbrace>\<rbrace>"
  let ?ev = "(tsend, Send (Tr (Honest P')) ?msg [])"
  let ?mac = "Hash \<lbrace>Key (symKey (Honest P) (Honest V)), \<lbrace>Nonce (Honest V) NV, \<lbrace>NP', Agent (Honest P)\<rbrace>\<rbrace>\<rbrace>"
  show ?case proof cases
    assume "(tf, Send (Tx (Honest F) j) mf Lf) = ?ev"
    with `?mac \<in> subterms {mf}`
    have or: "?mac \<in> subterms {NV'} \<or> (P'=P \<and> V'=Honest V \<and> NV' = Nonce (Honest V) NV
                                      \<and> NP' = Nonce (Honest P') NP)"
      apply (auto simp add: MACM_def)
      apply (drule injective_symKey)
      apply (case_tac "V=P'")
        apply simp
        apply (case_tac "V'= Honest P'")
      apply auto
      done
    thus ?thesis proof cases
      assume "?mac \<in> subterms {NV'}"
      show ?thesis proof cases
        assume "?ev \<in> set tr"
        thus ?thesis using prems(6-) by auto
      next
        assume "?ev \<notin> set tr"
        thus ?thesis using prems(6-) apply -
          apply (drule_tac m="NV'" in send_before_recv[simplified])
          apply assumption
          apply (elim exE conjE)
          apply simp
          apply (drule hash_not_LowHam)
          apply simp
          apply (drule Send_imp_parts_used)  back
          apply assumption
          apply simp
          done
      qed
    next
      assume "\<not> (?mac \<in> subterms {NV'})"
      hence "P'=P \<and> V'=Honest V \<and> NV' = Nonce (Honest V) NV \<and> NP' = Nonce (Honest P') NP"
        using or by auto
      thus ?thesis using prems(6-) by auto
    qed
  next
    assume "(tf, Send (Tx (Honest F) j) mf Lf) \<noteq> ?ev"
    thus ?thesis using prems by auto
  qed
qed

lemma originate_unique:
  assumes "m \<notin> used (beforeEvent (ta, Send TA ma La) tr)"
  and     "m \<notin> used (beforeEvent (tb, Send TB mb Lb) tr)"
  and     "(tb, Send TB mb Lb) \<noteq> (ta, Send TA ma La)"
  and     "(tb, Send TB mb Lb) \<in> set tr"
  and     "(ta, Send TA ma La) \<in> set tr"
  and     " m \<in> subterms {ma}"
  shows   "m \<notin> subterms {mb}" using prems
  apply (induct tr)
  apply simp
  apply (case_tac "a=(ta, Send TA ma La) \<and> a \<notin> set tr")
  apply (elim conjE)
  apply simp
  apply (case_tac "m \<in> subterms {mb}") prefer 2
  apply force
  apply (subgoal_tac "(tb, Send TB mb Lb) \<in> set tr") prefer 2
  apply force
  apply (frule_tac Y=m in Send_imp_parts_used)
  apply force
  apply force
  apply (case_tac "a=(tb, Send TB mb Lb) \<and> a \<notin> set tr")
  apply (elim conjE)
  apply simp
  apply (subgoal_tac "(ta, Send TA ma La) \<in> set tr") prefer 2
  apply force
  apply (frule_tac Y=m in Send_imp_parts_used)
  apply force
  apply force
  apply auto
done


lemma freshNonce_dishonestAgent_send_recv:
  assumes "tr \<in> mdb"
  and     "(t, Send (Tx (Honest A) i) m L) \<in> set tr \<or> (t, Recv (Rx (Honest A) i) m) \<in> set tr"
  and     "Hash \<lbrace>Nonce (Honest B) NB, \<lbrace> NC, Agent (Intruder I)\<rbrace>\<rbrace> \<in> subterms {m}"
  and     "(tnonce, Send (Tr (Honest B)) (Nonce (Honest B) NB) []) \<in> set tr"
  and     "Nonce (Honest B) NB
           \<notin> usedI (beforeEvent (tnonce, Send (Tr (Honest B)) (Nonce (Honest B) NB) []) tr)"
  shows "\<exists> I'. t - tnonce \<ge> cdistl (Honest B) (Intruder I') + cdistl (Intruder I') (Honest A)"
  using prems
proof (induct tr arbitrary: A B trec t m L i rule: mdb.induct)
  case (Fake tr mintr I tsend)
  hence "(t, Send (Tx (Honest A) i) m L) \<in> set tr \<or> (t, Recv (Rx (Honest A) i) m) \<in> set tr" by auto
  with Fake.hyps prems show ?case by (auto)
next
  case (MD1 tr tc V NV)
  have "Hash \<lbrace>Nonce (Honest B) NB, \<lbrace> NC, Agent (Intruder I)\<rbrace>\<rbrace> \<notin> subterms {Nonce (Honest V) NV}" by auto
  hence "(t, Send (Tx (Honest A) i) m L) \<in> set tr \<or> (t, Recv (Rx (Honest A) i) m) \<in> set tr"
    using MD1.prems by auto
  also have "(tnonce, Send (Tr (Honest B)) (Nonce (Honest B) NB)  []) \<in> set tr"
    using MD1.prems `tr \<in> mdb`
    apply (auto simp add: usedI_def split: split_if_asm)
    apply (subgoal_tac "Nonce (Honest B) NV \<in> subterms {m}") prefer 2
    apply (rule_tac G="{Hash \<lbrace>Nonce (Honest B) NV, \<lbrace>NC, Agent (Intruder I)\<rbrace>\<rbrace>}" in subterms.trans)
    apply force
    apply force
    apply (drule_tac Y="Nonce (Honest B) NV" in Send_imp_parts_used)
    apply assumption
    apply simp
    apply (subgoal_tac "Nonce (Honest B) NV \<in> subterms {m}") prefer 2
    apply (rule_tac G="{Hash \<lbrace>Nonce (Honest B) NV, \<lbrace>NC, Agent (Intruder I)\<rbrace>\<rbrace>}" in subterms.trans)
    apply force
    apply force
    apply (drule send_before_recv[simplified])
    apply assumption
    apply auto
    apply (drule nonce_not_LowHam)
    apply simp
    apply (drule Send_imp_parts_used)
    apply simp
    apply simp
    done
  ultimately show ?case using MD1.hyps prems  apply -
    apply (rule MD1.hyps(2))
    apply force
    apply force
    apply force
    apply force
    done
next
  case (MD4 tr t A NA)
  hence "(t, Send (Tx (Honest A) i) m L) \<in> set tr \<or> (t, Recv (Rx (Honest A) i) m) \<in> set tr"
    using prems by auto
  with MD4.hyps prems show ?case apply -
    apply (rule MD4.hyps(2))
    apply force
    apply force
    apply force
    apply force
    done
next
  case (Con tr trecv tsend C ia mc Lc D id tab dmsg)
  let ?evrecv = "(trecv, Recv (Rx D id) (Xor mc dmsg))"
  let ?evsend = "(tsend, Send (Tx C ia) mc Lc)"
  show ?case proof cases
    assume "(t, Recv (Rx (Honest A) i) m) \<in> set (?evrecv#tr)"
    show ?case proof cases
      assume "(t, Recv (Rx (Honest A) i) m) = ?evrecv"
      hence mceq: "m = Xor mc dmsg" and Deq: "D=Honest A" and trecveq: "t=trecv" by auto
      show ?thesis proof cases
        assume "\<exists> I'. C = Intruder I'"
        then obtain I' where I: "C = Intruder I'" by auto
        have cdist2: "cdistl (Honest B) (Intruder I') \<le> tsend - tnonce" using prems apply -
          apply (rule_tac A="Honest B" and NA=NB and i=0 and ma="(Nonce (Honest B) NB)" and tr=tr
                 and mb="mc"in fresh_nonce_earliest_send)
          apply force
          apply force
          apply (simp add: usedI_def)
          apply force defer
          apply force
          apply force
          apply (rule_tac G="{Hash \<lbrace>Nonce (Honest B) NB, \<lbrace>NC, Agent (Intruder I)\<rbrace>\<rbrace>}" in subterms.trans)
          apply force
          apply clarsimp
          apply (rule hash_not_LowHam)
          apply simp
          apply simp
          done
        have cdist3: "cdistl (Intruder I') (Honest A) \<le> t - tsend" using trecveq I Con.hyps Deq
          apply auto
          apply (frule noflt_some2)
          by auto
        hence "t - tnonce \<ge> cdistl (Honest B) (Intruder I') + cdistl (Intruder I') (Honest A)" using cdist2
          by auto
        thus ?thesis by auto
      next
        assume "\<not> (\<exists> I'. C = Intruder I')"
        then obtain F where F: "C = Honest F" apply (case_tac C) by auto
        hence "\<exists>I'. cdistl (Honest B) (Intruder I') + cdistl (Intruder I') (Honest F) \<le> tsend - tnonce"
          using Con.prems Con.hyps mceq Deq trecveq apply -
          apply (rule_tac m=mc in Con.hyps(2))
          apply force defer
          apply force
          apply (force simp add: usedI_def)
          apply clarsimp
          apply (rule hash_not_LowHam)
          apply simp
          apply simp
          done
        then obtain I where cdist1:
          "cdistl (Honest B) (Intruder I) + cdistl (Intruder I) (Honest F) \<le> tsend - tnonce"
          by auto
        have cdist2: "cdistl (Honest F) (Honest A) \<le> t - tsend" using trecveq Con.hyps Deq F
          apply auto
          apply (frule noflt_some2)
          by auto
        have cdist3: "cdistl (Intruder I) (Honest F) + cdistl (Honest F) (Honest A) \<ge>
                      cdistl (Intruder I) (Honest A)"
          by (rule cdistl_triangle)

        have "t - tnonce \<ge> (cdistl (Honest B) (Intruder I)  + cdistl (Intruder I) (Honest F))
                               + cdistl (Honest F) (Honest A)" using cdist1 cdist2 apply -
          by auto
        then also have "... \<ge> cdistl (Honest B) (Intruder I) + cdistl (Intruder I) (Honest A)" using cdist3
          by auto
        ultimately have "t - tnonce \<ge> cdistl (Honest B) (Intruder I) + cdistl (Intruder I) (Honest A)"
          by auto
        thus ?thesis by auto
      qed
    next
      assume "(t, Recv (Rx (Honest A) i) m) \<noteq> ?evrecv"
      hence "(t, Recv (Rx (Honest A) i) m) \<in> set tr" using prems by auto
      thus ?thesis using Con.prems apply -
        apply (rule Con.hyps(2))
        by auto
    qed
  next
    assume "(t, Recv (Rx (Honest A) i) m) \<notin> set (?evrecv#tr)"
    hence "(t, Send (Tx (Honest A) i) m L) \<in> set tr" using prems
      apply -
      apply auto
      done
    thus ?thesis using Con.prems apply -
      apply (rule Con.hyps(2))
      by auto
  qed
next
  case (MD3 tr tsend trec D NE tsend1 NF C)

  let ?macm = "Hash[Key (symKey (Honest D) C)] \<lbrace>NE, \<lbrace>Nonce (Honest D) NF, Agent (Honest D)\<rbrace>\<rbrace>"
  let ?ev = "(tsend, Send (Tr (Honest D)) ?macm [])"
  show ?case proof cases
    assume "(t, Send (Tx (Honest A) i) m L) \<in> set (?ev#tr)"
    show ?case proof cases
      assume eveq: "(t, Send (Tx (Honest A) i) m L) = ?ev"
      have hashNE: "Hash \<lbrace>Nonce (Honest B) NB, \<lbrace> NC, Agent (Intruder I)\<rbrace>\<rbrace> \<in> subterms {NE}"
        using eveq `Hash \<lbrace>Nonce (Honest B) NB, \<lbrace>NC, Agent (Intruder I)\<rbrace>\<rbrace> \<in> subterms {m}`
        apply (auto simp add: MACM_def)
        done
      
      have "\<exists>F k tsendnc L m' dmsg.
               dmsg \<in> LowHamXor \<and>
               NE = Xor m' dmsg \<and>
               (tsendnc, Send (Tx F k) m' L) \<in> set tr \<and>
               cdist (Tx F k) (Rec (Honest D))  \<le> trec \<and>
               cdistM (Tx F k) (Rec (Honest D)) \<noteq> None \<and>
               tsendnc \<le> trec - cdist (Tx F k) (Rec (Honest D))"
          using prems apply -
          apply (rule send_before_recv)
          apply auto
          done
      then obtain F k tsendnc Lf m' dmsg 
             where nctr: "(tsendnc, Send (Tx F k) m' Lf) \<in> set tr"
             and   some: "cdistM (Tx F k) (Rec (Honest D)) \<noteq> None"
             and   cdist: "tsendnc \<le> trec - cdist (Tx F k) (Rec (Honest D))"
             and   low:   "dmsg \<in> LowHamXor"
             and   NEeq:  "NE = Xor m' dmsg"
          by auto
      (* Send for receive: either an honest agent: use IH + triangle
         or a dishonest user: must have received nonce and sent the message *)
      show ?case proof cases
        assume "\<exists> G. F = Honest G"
        then obtain G where F: "F = Honest G" by auto
        hence "\<exists>I'. cdistl (Honest B) (Intruder I') + cdistl (Intruder I') (Honest G) \<le> tsendnc - tnonce"
          using MD3.prems MD3.hyps nctr hashNE NEeq low apply -
          apply (rule_tac m=m' in MD3.hyps(2))
          apply force defer
          apply force
          apply (force simp add: usedI_def)
          apply (rule hash_not_LowHam)
          apply simp
          apply simp
          done
        then obtain I where cdist1:
           "cdistl (Honest B) (Intruder I) + cdistl (Intruder I) (Honest G) \<le> tsendnc - tnonce"
          by auto
        have cdist2: "cdistl (Honest G) (Honest A) \<le> trec - tsendnc" using some cdist eveq F
          apply simp
          apply (auto simp add: cdist_def)
          apply (frule noflt_some2)
          apply auto
          done
        have "trec \<le> t" using eveq MD3.hyps MD3.prems apply -
          apply (rule maxtime_geq_elem)
          by auto
        hence  cdist3: "cdistl (Honest G) (Honest A) \<le> t - tsendnc" using cdist2 eveq by auto
        have cdist4: "cdistl (Intruder I) (Honest G) + cdistl (Honest G) (Honest A) \<ge>
                      cdistl (Intruder I) (Honest A)"
          by (rule cdistl_triangle)

        have "t - tnonce \<ge> (cdistl (Honest B) (Intruder I)  + cdistl (Intruder I) (Honest G))
              + cdistl (Honest G) (Honest A)" using cdist1 cdist3 apply -
          by auto
        then also have "... \<ge> cdistl (Honest B) (Intruder I) + cdistl (Intruder I) (Honest A)" using cdist4
          by auto
        ultimately have "t - tnonce \<ge> cdistl (Honest B) (Intruder I) + cdistl (Intruder I) (Honest A)"
          by auto
        thus ?thesis by auto
      next
        assume "\<not> (\<exists> G. F = Honest G)"
        hence "\<exists> G. F = Intruder G" apply (case_tac F, auto) done
        then obtain G where F: "F = Intruder G" by auto
        with hashNE prems have cdist2: "cdistl (Honest B) (Intruder G) \<le> tsendnc - tnonce"
          apply (rule_tac A="Honest B" and NA=NB and i=0 and ma="(Nonce (Honest B) NB)" and tr=tr
                 and mb="m'"in fresh_nonce_earliest_send)
          apply simp
          apply simp
          apply (simp add: usedI_def)
          apply simp defer
          apply simp
          apply simp
          apply (rule nonce_not_LowHam)
          apply simp
          apply simp
          apply clarsimp
          apply (rule_tac G="{Hash \<lbrace>Nonce (Honest B) NB, \<lbrace>NC, Agent (Intruder I)\<rbrace>\<rbrace>}" in  subterms.trans)
          apply simp
          apply simp
          done
        have D: "D=A" using prems by auto
        have t: "trec \<le> t" using prems apply -
          apply (rule maxtime_geq_elem)
          by auto
        have cdist3: "cdist (Tx F k) (Rec (Honest D)) \<le> trec - tsendnc" using cdist by auto
        have "cdistl (Honest B) (Intruder G) + cdistl (Intruder G) (Honest A)
              \<le> cdistl (Honest B) (Intruder G) + cdist (Tx (Intruder G) k) (Rec (Honest A))"
          using some F prems
          apply simp
          apply (auto dest: noflt_some2 simp add: cdist_def)
          done
        also have "... \<le> (tsendnc - tnonce) + (trec - tsendnc)"  using cdist3 cdist2 F D apply -
          apply (rule add_mono)
          by auto
        also have "... \<le> t - tnonce" using t by auto
        finally show ?thesis using F D prems
          apply (rule_tac x=G in exI)
          by auto
      qed
    next
      assume "(t, Send (Tx (Honest A) i) m L) \<noteq> ?ev"
      hence "(t, Send (Tx (Honest A) i) m L) \<in> set tr" using prems by auto
      with MD3.hyps prems show ?thesis apply -
        apply (rule MD3.hyps(2))
        apply force
        apply force
        apply force
        apply force
        done
    qed
  next
    assume "(t, Send (Tx (Honest A) i) m L) \<notin> set (?ev#tr)"
    hence "(t, Recv (Rx (Honest A) i) m) \<in> set tr" using prems by auto
    with MD3.hyps prems show ?case apply -
      apply (rule MD3.hyps(2))
      apply force
      apply force
      apply force
      apply force
      done
  qed
next
  case (MD2 tr tsend trec P NV NP)
  let ?fr = "Hash \<lbrace> NV, \<lbrace> Nonce (Honest P) NP,  Agent (Honest P)\<rbrace>\<rbrace>"
  let ?ev = "(tsend, Send (Tr (Honest P)) (Hash \<lbrace> NV, \<lbrace> Nonce (Honest P) NP, Agent (Honest P) \<rbrace>\<rbrace>)
                                          [NV, Nonce (Honest P) NP])"
  show ?case proof cases
    assume "(t, Send (Tx (Honest A) i) m L) \<in> set (?ev#tr)"
    show ?case proof cases
      assume eveq: "(t, Send (Tx (Honest A) i) m L) = ?ev"
      hence hashNV: "Hash \<lbrace>Nonce (Honest B) NB, \<lbrace>NC, Agent (Intruder I)\<rbrace>\<rbrace> \<in> subterms {NV}"
        using `Hash \<lbrace>Nonce (Honest B) NB, \<lbrace>NC, Agent (Intruder I)\<rbrace>\<rbrace> \<in> subterms {m}` eveq
          by auto
      have "\<exists>F k tsendnc Lf m' dmsg.     
                             dmsg \<in> LowHamXor \<and>
                             NV = Xor m' dmsg \<and>
                             (tsendnc, Send (Tx F k) m' Lf) \<in> set tr \<and>
                             cdist (Tx F k) (Rec (Honest P))  \<le> trec \<and>
                             cdistM (Tx F k) (Rec (Honest P)) \<noteq> None \<and>
                 tsendnc \<le> trec - cdist (Tx F k) (Rec (Honest P))"
        using prems apply -
          apply (rule send_before_recv)
          apply auto
        done
      then obtain F k tsendnc Lf m' dmsg where nctr: "(tsendnc, Send (Tx F k) m' Lf) \<in> set tr"
                 and   some:  "cdistM (Tx F k) (Rec (Honest P)) \<noteq> None"
                 and   cdist: "tsendnc \<le> trec - cdist (Tx F k) (Rec (Honest P))"
                 and   NVeq:  "NV = Xor m' dmsg"
                 and   low:   "dmsg \<in> LowHamXor"
        by auto
      
      have nonceNV: "Nonce (Honest B) NB \<in> subterms {NV}" using MD2.prems eveq  NVeq low apply -
          apply simp
          apply (rule_tac G="{Hash \<lbrace>Nonce (Honest B) NB, \<lbrace>NC, Agent (Intruder I)\<rbrace>\<rbrace>}" in subterms.trans)
          apply simp
          apply force
          done


      (* Send for receive: either an honest agent: use IH + triangle
                           or a dishonest user: must have received nonce and sent the message *)
      show ?case proof cases
        assume "\<exists> G. F = Honest G"
        then obtain G where F: "F= Honest G" by auto
        hence "\<exists>I'. cdistl (Honest B) (Intruder I') + cdistl (Intruder I') (Honest G) \<le> tsendnc - tnonce"
          using MD2.prems MD2.hyps nctr nonceNV hashNV low NVeq apply -
          apply (rule_tac m=m' in MD2.hyps(2))
          apply force defer
          apply force
          apply (simp add: usedI_def)
          apply (rule hash_not_LowHam)
          apply simp
          apply simp
          done
        then obtain I where cdist1:
          "cdistl (Honest B) (Intruder I) + cdistl (Intruder I) (Honest G) \<le> tsendnc - tnonce"
           by auto
        have cdist2: "cdistl (Honest G) (Honest A) \<le> trec - tsendnc" using some cdist eveq F
          apply simp
          apply (auto simp add: cdist_def)
          apply (frule noflt_some2)
          apply auto
          done
        have "trec \<le> t" using eveq MD2.hyps MD2.prems apply -
          apply (rule maxtime_geq_elem)
          by auto
        hence  cdist3: "cdistl (Honest G) (Honest A) \<le> t - tsendnc" using cdist2 eveq by auto
        have cdist4: "cdistl (Intruder I) (Honest G) + cdistl (Honest G) (Honest A) \<ge>
                      cdistl (Intruder I) (Honest A)"
          by (rule cdistl_triangle)

        have "t - tnonce \<ge> (cdistl (Honest B) (Intruder I)  + cdistl (Intruder I) (Honest G))
                             + cdistl (Honest G) (Honest A)" using cdist1 cdist3 apply -
          by auto
        then also have "... \<ge> cdistl (Honest B) (Intruder I) + cdistl (Intruder I) (Honest A)" using cdist4
          by auto
        ultimately have "t - tnonce \<ge> cdistl (Honest B) (Intruder I) + cdistl (Intruder I) (Honest A)"
          by auto
        thus ?thesis by auto
      next
        assume "\<not> (\<exists> G. F = Honest G)"
        hence "\<exists> G. F = Intruder G" apply (case_tac F, auto) done
        then obtain G where F: "F = Intruder G" by auto
        with nonceNV MD2.prems MD2.hyps eveq nctr NVeq low
        have cdist2: "cdistl (Honest B) (Intruder G) \<le> tsendnc - tnonce"
          apply (rule_tac A="Honest B" and NA=NB and i=0 and ma="(Nonce (Honest B) NB)" and tr=tr
                 and mb="m'" in fresh_nonce_earliest_send)
          apply simp
          apply simp
          apply (simp add: usedI_def)
          apply simp defer
          apply simp
          apply simp
          apply clarsimp
          apply (rule nonce_not_LowHam)
          apply simp
          apply simp
          done
        have P: "P=A" using prems by auto
        have t: "trec \<le> t" using prems apply -
          apply (rule maxtime_geq_elem)
          by auto
        have cdist3: "cdist (Tx F k) (Rec (Honest P)) \<le> trec - tsendnc" using cdist by auto
        have "cdistl (Honest B) (Intruder G) + cdistl (Intruder G) (Honest A)
              \<le> cdistl (Honest B) (Intruder G) + cdist (Tx (Intruder G) k) (Rec (Honest A))"
          using some F prems
          apply simp
          apply (auto dest: noflt_some2 simp add: cdist_def)
          done
        also have "... \<le> (tsendnc - tnonce) + (trec - tsendnc)"  using cdist3 cdist2 F P apply -
          apply (rule add_mono)
          by auto
        also have "... \<le> t - tnonce" using t by auto
        finally show ?thesis using F P prems
          apply (rule_tac x=G in exI)
          by auto
      qed
    next
      assume "(t, Send (Tx (Honest A) i) m L) \<noteq> ?ev"
      thus ?thesis using MD2.prems MD2.hyps apply -
        apply (rule MD2.hyps(2))
        apply force
        apply force
        apply force
        apply (force simp add: usedI_def)
        done
    qed
  next
    assume "(t, Send (Tx (Honest A) i) m L) \<notin> set (?ev#tr)"
    thus ?thesis using MD2.prems MD2.hyps apply -
      apply (rule MD2.hyps(2))
      apply force
      apply force
      apply force
      apply (force simp add: usedI_def)
      done
  qed
next
  case Nil
  show ?case using prems by auto
qed

subsection {* Security proof for Honest Provers *}

lemma mdb_secure:
  assumes mdb: "tr \<in> mdb"
  and believe: "(tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>) \<in> set tr"
  shows   "d \<ge> pdist (Honest V) (Honest P)" using prems
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with Fake.hyps prems show ?case by (auto)
next
  case (Con tr tc C mc D tab)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
next
  case (MD1 tr t V' NV)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD1.hyps prems show ?case by auto
next
  case (MD2 tr tsend trec P' NV NP)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD2.hyps prems show ?case by (auto)
next
  case (MD3 tr tsend trec P' NV tsend1 NP V')
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD3.hyps prems show ?case by auto
next
  -- {* the only nontrivial case since it adds Claim events *}
  case (MD4 tr t trec2 V' P' NV NP trec1 tsend)
  let ?x = "(t, Claim (Honest V') \<lbrace>Agent P', Real ((trec1 - tsend)*vc/2)\<rbrace>)"
  and ?ev = "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>))"
  show ?case proof cases
    -- {* the added event is the Claim event from the premise,
          the other case follows trivially from the IH *}
    assume "?x = ?ev"
    hence Veq: "V'=V" and Peq: "P'=Honest P" and deq: "d=(trec1 - tsend)*vc/2" by auto
    let ?mmac = "\<lbrace>Nonce (Honest V) NV, \<lbrace>NP, Agent (Honest P)\<rbrace>\<rbrace>"
    let ?macmsg = "Hash[Key (symKey (Honest P) (Honest V))]
                       \<lbrace>Nonce (Honest V) NV, \<lbrace>NP, Agent (Honest P)\<rbrace>\<rbrace>" and
        ?fastmsg = "Hash \<lbrace> Nonce (Honest V) NV, \<lbrace> NP, Agent (Honest P)\<rbrace>\<rbrace>"
    have NC_fresh:
      "Nonce (Honest V) NV
       \<notin> usedI (beforeEvent (tsend, Send (Tr (Honest V)) (Nonce (Honest V) NV) []) tr)"
      using prems `tr \<in> mdb` apply -
      by (rule nonce_fresh_challenge, auto)
    -- {* We first handle the trivial case where V runs the protocol with himself *}
    show ?case proof cases
      assume PeqV: "P=V"
      with `tr \<in> mdb` prems Veq Peq deq
        obtain F tfsend i L m' dmsg
        where lh: "dmsg \<in> LowHamXor" and
              fastEq: "?fastmsg = Xor m' dmsg" and
              fastintrE: "(tfsend, Send (Tx F i) m' L) \<in> set tr" and
                         "trec1 >= cdist (Tx F i) (Rec (Honest V))" and
              q: "tfsend \<le> trec1 - cdist (Tx F i) (Rec (Honest V))" and
              cdist_some: "cdistM (Tx F i) (Rec (Honest V)) \<noteq> None"
          apply - apply (drule_tac m="?fastmsg" in send_before_recv [simplified])
          by auto
      hence "tfsend \<le> trec1" apply auto
        apply (rule_tac y="trec1 - y" in order_trans)
        apply (auto simp add: cdist_def)
        by (rule cdistnoneg_some)
      show ?thesis proof (rule ccontr)
        assume "\<not> pdist (Honest V) (Honest P) \<le> d"
        hence "trec1 < tsend" using PeqV apply -
          apply (auto simp add: minusv_def CauchySchwarz.norm_def
                 pdist_def vlen_def ith_def deq)
          apply (erule contrapos_np)
          apply (subgoal_tac "trec1 \<ge> tsend")
          apply (rule mult_nonneg_nonneg)
          apply (auto intro: vc_pos order_less_imp_le)
          done
      hence "tfsend < tsend" using `tfsend \<le> trec1` by auto
      hence "(tfsend, Send (Tx F i) m' L)
                 \<in> set (beforeEvent (tsend, Send (Tr (Honest V)) (Nonce (Honest V) NV) []) tr)"
        using fastintrE prems apply -
        apply (rule beforeEvent_earlier)
        apply auto
        done
      thus False using NC_fresh lh fastEq apply -
        apply (rotate_tac 2)
        apply (erule contrapos_np)
        apply (auto simp add: usedI_def)
        apply (erule Send_imp_parts_used)
        apply (erule nonce_not_LowHam)
        apply (subgoal_tac " Nonce (Honest V) NV \<in> subterms {Hash \<lbrace>Nonce (Honest V) NV, \<lbrace>NP, Agent (Honest P)\<rbrace>\<rbrace>}")
        apply force
        apply (thin_tac "?a = ?b")
        by auto
    qed
  next
    assume PnotV: "P \<noteq> V"
    have mac_recv:  "(trec2,Recv (Rec (Honest V)) ?macmsg) \<in> set tr" using prems by auto
    have fast_recv: "(trec1, Recv (Rec (Honest V)) ?fastmsg) \<in> set tr" using prems by auto
    with mac_recv `tr \<in> mdb`
      obtain E tesend i Le m' dmsg
      where lh: "dmsg \<in> LowHamXor" and
            macEq: "?macmsg = Xor m' dmsg" and
            macintrE: "(tesend, Send (Tx E i) m' Le) \<in> set tr" and
                      "trec2 >= cdist (Tx E i) (Rec (Honest V))" and
                      "cdistM (Tx E i) (Rec (Honest V)) \<noteq> None" and
            q: "tesend \<le> trec2 - cdist (Tx E i) (Rec (Honest V))"
      apply - apply (drule_tac m="?macmsg" in send_before_recv [simplified])
      by auto
    hence "\<exists>tf mf j F Lf.
             (tf, Send (Tx (Honest F) j) mf Lf) \<in> set tr
             \<and> (F = P \<or> F = V)
             \<and> (Hash (MPair (Key (symKey (Honest P) (Honest V))) ?mmac) \<in> subterms {mf})
             \<and> (Hash (MPair (Key (symKey (Honest P) (Honest V))) ?mmac)
               \<notin> used (beforeEvent (tf, Send (Tx (Honest F) j) mf Lf) tr))" using prems lh macEq apply -
      apply (rule_tac tc=tesend and mc="m'" in mac_originates)
      apply auto
      apply (erule hash_not_LowHam)
      apply (auto simp add: MACM_def)
      apply (subgoal_tac " Hash \<lbrace>Key (symKey (Honest P) (Honest V)), \<lbrace>Nonce (Honest V) NV, \<lbrace>NP, Agent (Honest P)\<rbrace>\<rbrace>\<rbrace>
                           \<in> subterms {\<lbrace>Hash \<lbrace>Key (symKey (Honest P) (Honest V)), \<lbrace>Nonce (Honest V) NV,
                                       \<lbrace>NP, Agent (Honest P)\<rbrace>\<rbrace>\<rbrace>,
                                       \<lbrace>Nonce (Honest V) NV, \<lbrace>NP, Agent (Honest P)\<rbrace>\<rbrace>\<rbrace>}")
      apply force
      apply (thin_tac "?y = Xor m' dmsg")
      apply auto
      done
    then obtain tf mf j F Lf where ftr: "(tf, Send (Tx (Honest F) j) mf Lf) \<in> set tr"
                             and FPV: "(F = P \<or> F = V)"
                             and mfsubterm: "Hash (MPair (Key (symKey (Honest P) (Honest V))) ?mmac)
                                             \<in> subterms {mf}"
                             and ffresh: "Hash (MPair (Key (symKey (Honest P) (Honest V))) ?mmac)
                                          \<notin> used (beforeEvent (tf, Send (Tx (Honest F) j) mf Lf) tr)"
      by auto
    hence ex: "\<exists> NPP. F=P \<and> (NP = Nonce (Honest P) NPP)
               \<and> Lf = []
               \<and> mf = Hash[Key (symKey (Honest P) (Honest V))]
                          \<lbrace>Nonce (Honest V) NV, \<lbrace>Nonce (Honest P) NPP, Agent (Honest P)\<rbrace>\<rbrace>" apply -
      apply (rule mac_msg_originates)
      by (auto intro: prems)
    then obtain NPP where ef:
      "(tf, Send (Tx (Honest P) j) Hash[Key (symKey (Honest P) (Honest V))]
                                   \<lbrace>Nonce (Honest V) NV, \<lbrace>Nonce (Honest P) NPP, Agent (Honest P)\<rbrace>\<rbrace>
                                   []) \<in> set tr"
      and NPP: "NP = Nonce (Honest P) NPP"
      apply (insert ftr ex)
      by auto
    then obtain tfast where
      fastsend: "(tfast, Send (Tr (Honest P))
                               (Hash \<lbrace> Nonce (Honest V) NV, \<lbrace> Nonce (Honest P) NPP,  Agent (Honest P)\<rbrace>\<rbrace>)
                               [Nonce (Honest V) NV, Nonce (Honest P) NPP]) \<in> set tr" and
      fresh: "Nonce (Honest P) NPP
              \<notin> usedI
              (beforeEvent
                (tfast, Send (Tr (Honest P))
                               (Hash \<lbrace>Nonce (Honest V) NV,  \<lbrace> Nonce (Honest P) NPP, Agent (Honest P)\<rbrace>\<rbrace>)
                               [Nonce (Honest V) NV, Nonce (Honest P) NPP]) tr)"
      using `tr \<in> mdb` apply -
        apply (drule MAC_send_prover2)
        apply assumption
        apply (auto intro: prems)
      done
    hence rec1_fast: "trec1 - tfast  >= cdistl (Honest P) (Honest V)" using `tr \<in> mdb` PnotV fastsend
      apply -
      apply (erule_tac NA="NPP" and
             ma="Hash \<lbrace>Nonce (Honest V) NV, \<lbrace> Nonce (Honest P) NPP, Agent (Honest P)\<rbrace>\<rbrace>" and
             mb="?fastmsg" and i=0 and tr=tr and
             A="Honest P" and B="Honest V" and
             j=0 and tr=tr
             in fresh_nonce_earliest_recv [simplified]) prefer 6
      apply force prefer 4
      apply force prefer 4
      apply (rule fast_recv)
      apply (auto simp add: NPP)
      apply (force simp add: usedI_def)
      done
    have fast_send: "tfast - tsend >= cdistl (Honest V) (Honest P)" using `tr \<in> mdb` PnotV
      apply -
      apply (erule_tac NA="NV" and i=0 and ma="Nonce (Honest V) NV"
             and mb="Hash \<lbrace>Nonce (Honest V) NV, \<lbrace> Nonce (Honest P) NPP, Agent (Honest P)\<rbrace>\<rbrace>"
             in fresh_nonce_earliest_send[simplified])
      apply force
      apply (insert NC_fresh, force simp add: usedI_def)
      apply force
      apply (auto intro: prems simp add: Veq)
      apply (auto simp add: Veq[THEN sym] intro: prems)
      done
    have "2* cdistl (Honest V) (Honest P) \<le> cdistl (Honest V) (Honest P) + cdistl (Honest P) (Honest V)"
      by (auto simp add: cdistl_symm)
    also have "... \<le> tfast - tsend + (trec1 - tfast)" using fast_send rec1_fast by auto
    also have "tfast - tsend + (trec1 - tfast) \<le>  trec1 - tsend" by auto
    finally have "cdistl (Honest V) (Honest P) * 2 \<le>  trec1 - tsend" by auto
    thus ?thesis using deq
      apply (simp add: cdistl_def deq)
      apply (subgoal_tac "(pdist (Honest V) (Honest P) * 2 /vc) * vc \<le> (trec1 - tsend) * vc") defer
      apply (rule mult_right_mono)
      apply force
      apply (insert vc_pos, auto split: split_if_asm)
      done
    qed
  next
    assume "?x \<noteq> ?ev"
    show ?case using prems by auto
  qed
next
  case Nil thus ?case by auto
qed

subsection {* Security for dishonest Provers *}

lemma mdb_secure_dishonest:
  assumes mdb: "tr \<in> mdb"
  and    believe: "(tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>) \<in> set tr"
  shows   "\<exists> P'. d \<ge> pdist (Honest V) (Intruder P')" using prems
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with Fake.hyps prems show ?case by (auto)
next
  case (Con tr tc C mc D tab)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
next
  case (MD1 tr t A NA)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD1.hyps prems show ?case by (auto)
next
  case (MD2 tr tsend trec B NA NB)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD2.hyps prems show ?case by (auto)
next
  case (MD3 tr tsend trec B NA tsend1 NB A)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD3.hyps prems show ?case by (auto)
next
  -- {* the only nontrivial case since it adds Claim events *}
  case (MD4 tr t trec2 A B NA NB trec1 tsend)
  let ?x = "(t, Claim (Honest A) \<lbrace>Agent B, Real ((trec1 - tsend)*vc/2)\<rbrace>)"
  and ?ev = "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>))"
  show ?case proof cases
    -- {* the added event is the Claim event from the premise,
          the other case follows trivially from the IH *}
    assume "?x = ?ev"
    hence Aeq: "A=V" and Beq: "B=Intruder P" and deq: "d=(trec1 - tsend)*vc/2" by auto
    let ?macmsg = "Hash[Key (symKey B (Honest A))]
                       \<lbrace>Nonce (Honest V) NA, \<lbrace>NB, Agent (Intruder P)\<rbrace>\<rbrace>" and
        ?fastmsg = "Hash \<lbrace> Nonce (Honest V) NA, \<lbrace> NB, Agent (Intruder P)\<rbrace>\<rbrace>"
    have NC_fresh:
      "Nonce (Honest V) NA
       \<notin> usedI (beforeEvent (tsend, Send (Tr (Honest V)) (Nonce (Honest V) NA) []) tr)"
      using prems `tr \<in> mdb` Aeq Beq deq
      apply -
      by (rule nonce_fresh_challenge, auto)
    have "\<exists> I'. trec1 - tsend \<ge> cdistl (Honest V) (Intruder I') + cdistl (Intruder I') (Honest V)"
      using prems Aeq NC_fresh

      apply -
      apply (rule_tac i=0 and m="?fastmsg" and tr=tr in freshNonce_dishonestAgent_send_recv)
      by auto
    then obtain I where "trec1 - tsend \<ge> cdistl (Honest V) (Intruder I) + cdistl (Intruder I) (Honest V)"
      by auto
    thus ?thesis using deq Aeq vc_pos
      apply (auto simp add: cdistl_def)
      apply (rule_tac x=I in exI)
      apply (auto simp add: pdist_symm)
      apply (subgoal_tac "(2 * pdist (Honest V) (Intruder I) / vc) * vc\<le> (trec1 - tsend) * vc") prefer 2
      apply (rule mult_right_mono)
      apply (auto split: split_if_asm)
      done
  next
    assume "?x \<noteq> ?ev"
    show ?case using prems by auto
  qed
next
  case Nil thus ?case by auto
qed

end
