(******************************************************************************

 ProtoVeriPhy

 Module: distance/BrandsChaum_explicit.thy

 Copyright (C) 2011-12 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
 License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Security Analysis of a fixed version of the Brands-Chaum protocol
          that uses explicit binding with a hash function to prevent
          Distance Hijacking Attacks.
          We prove that the resulting protocol is secure in our model
          Note that we abstract away from the individual bits exchanged in the rapid
          bit exchange phase, by performing the message exchange in 2 steps instead
          2*k steps. *}

theory BrandsChaum_explicit imports SystemCoffset SystemOrigination MessageTheoryXor3 begin

locale INITSTATE_SIG_NN = INITSTATE_PKSIG + INITSTATE_NONONCE

definition
  initStateMd :: "agent \<Rightarrow> msg set" where
  "initStateMd A == Key`({priSK A} \<union> (pubSK`UNIV))"

interpretation INITSTATE_SIG_NN Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components
               initStateMd Key
  apply (unfold_locales, auto simp add: initStateMd_def dest: injective_symKey)
  apply (drule subterms.singleton)
  apply (auto)
  apply (drule subterms.singleton)
  apply (auto)
  apply (drule subterms.singleton)
  apply (auto)
done

(* Informal multiset rewriting description of the protocol:

  md1(V): Fr(NV_V) \<longrightarrow> S_2(NV_V), Out (NV_V)

  // The commit is not required since we assume the hash function to be
  // non-invertible.
  md2(P): In(NV), S_1(NP_P) \<longrightarrow> Out( NV \<oplus> Hash(NP_P,P) ), S_3(NV, NP_P)

  md3(P): S_3(NV,NP_P) \<longrightarrow> Out( SIG_{SK(P)}( NV,NP_P,V ))

  md4(V): S_2(NV_V),
          In(NV_V \<oplus> Hash(NP,P))
          In(MAC_{K_VP}( NV_V, NP, P))
          \<longrightarrow>
          Claim(..)

*)

definition
  md1 :: "msg step"
 where
  "md1 tr V t =
     (UN NV. {ev. ev = (Nonce (Honest V) NV, SendEv 0 []) \<and>
                  Nonce (Honest V) NV \<notin> usedI tr})"

definition
  md2 :: "msg step"
 where
  "md2 tr P t =
     (UN NP NV trec.
        {ev. ev = (Xor NV (Hash\<lbrace> Nonce (Honest P) NP , Agent (Honest P) \<rbrace>)
                  , SendEv 0 [NV,Nonce (Honest P) NP]) \<and>
             Nonce (Honest P) NP \<notin> usedI tr \<and>
             (trec, Recv (Rec (Honest P)) NV) \<in> set tr})"

definition
  md3 :: "msg step"
 where
  "md3 tr P t =
     (UN NP NV V tsend trec.
        {ev. ev = ( Crypt (priSK (Honest P))
                   \<lbrace> NV, \<lbrace>Nonce (Honest P) NP,Agent V\<rbrace>\<rbrace>
                  , SendEv 0 []) \<and>
             (trec, Recv (Rec (Honest P)) NV) \<in> set tr \<and>
             (tsend,
              Send (Tr (Honest P))
                (Xor NV (Hash  \<lbrace> Nonce (Honest P) NP , Agent (Honest P) \<rbrace>))
                [NV,Nonce (Honest P) NP])
             \<in> set tr})"

definition
  md4 :: "msg step"
 where
  "md4 tr V t =
     (UN NP NV P trec1 trec2 tsend.
       {ev. ev = (\<lbrace>Agent P, Real ((trec1 - tsend) * vc/2)\<rbrace>, ClaimEv) \<and>
            (trec2, Recv (Rec (Honest V))
                  ( Crypt (priSK P)
                   \<lbrace> Nonce (Honest V) NV, \<lbrace> NP, Agent (Honest V)\<rbrace>\<rbrace>)) \<in> set tr \<and>
            (trec1, Recv (Rec (Honest V)) (Xor (Nonce (Honest V) NV)  (Hash \<lbrace> NP , Agent P \<rbrace>))) \<in> set tr \<and>
            (tsend, Send (Tr (Honest V)) (Nonce (Honest V) NV) []) \<in> set tr})"

definition
  md_proto :: "msg proto" where
  "md_proto = {md1,md2,md3,md4}"

lemmas md_defs = md_proto_def md1_def md2_def md3_def md4_def


locale PROTOCOL_MD = PROTOCOL_PKSIG_NOKEYS+PROTOCOL_NONONCE+INITSTATE_SIG_NN

interpretation PROTOCOL_MD Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components initStateMd Key md_proto
  apply (unfold_locales)
  apply (auto simp add: md_defs messagesProtoTr_def messagesProtoTrHonest_def initStateMd_def
              split: event.split split_if dest: parts.fst_set)
  apply (drule parts.singleton)
  apply auto
  apply (drule Key_parts_Xor)
  apply (drule parts.singleton)
  apply auto
  apply (drule view_elem_ex)
  apply auto
  apply (drule parts.singleton)
  apply auto
  apply (drule view_elem_ex)
  apply auto
  done

text {* Agents only look at their own views and all messages are derivable. *}

interpretation PROTOCOL_EXECUTABLE Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components initStateMd md_proto sys Key
  apply (unfold_locales)
  apply (auto simp add: md_defs initStateMd_def
                        messagesProto_def messagesProtoTrHonest_def)

  apply (rule DM.Xor)
  apply (drule view_elem_ex)
  apply auto
  apply (drule Recv_imp_knows_A)
  apply auto
  apply (rule DM.Crypt)
  apply (rule DM.MPair)
  apply auto
  apply (drule view_elem_ex)
  apply auto
  apply (drule Recv_imp_knows_A)
  apply simp

  apply (rule_tac x=NV in exI)
  apply (auto simp add: nonce_view_fresh [simplified md_proto_def]
                        nonce_view_used [simplified md_proto_def])
  apply (rule_tac x=NP in exI)
  apply auto
  apply (rule_tac x=trec in exI)
  apply (auto simp add: recv_a_view_a_r send_a_view_a_r)

  apply (rule_tac x=NP in exI)
  apply (rule_tac x=NV in exI)
  apply (rule_tac x=P in exI)
  apply (rule_tac x=trec1 in exI)
  apply (rule_tac x=trec2 in exI)
  apply (rule_tac x=tsend in exI)
  apply auto
  apply (rule_tac x=NP in exI)
  apply (rule_tac x=NV in exI)
  apply (rule_tac x=P in exI)
  apply (rule_tac x=trec1 in exI)
  apply (rule_tac x=trec2 in exI)
  apply (rule_tac x=tsend in exI)
  apply auto
done

text {* Agent behaviour does not change with constant clock errors. *}

interpretation  PROTOCOL_DELTAONLY Crypt Nonce MPair Hash Number parts subterms DM LowHamXor Xor components initStateMd Key md_proto
  apply unfold_locales
  apply (auto simp add: md_defs in_timetrans)
  apply (rule_tac x=NP in exI)
  apply (rule conjI)
  apply force
  apply force
  apply (rule_tac x=NP in exI) apply (rule_tac x=NV in exI)
  apply force
  apply (intro exI conjI, auto)
  apply (rule_tac f="MPair (Agent P)" in HOL.arg_cong) (* XXX: add intro rule for this *)
  apply (rule_tac f="Real" in HOL.arg_cong)
  apply force
  apply (intro exI)
  apply auto
  apply (rule_tac x="trec + coffset A" in exI, force)
  apply (intro exI)
  apply auto
  apply (rule_tac x="trec + coffset A" in exI, force)
  apply (rule_tac x="tsend + coffset A" in exI, force)
  apply (rule exI, rule exI, rule exI)
  apply (rule_tac x="trec1 + coffset A" in exI,
         rule_tac x="trec2 + coffset A" in exI,
         rule_tac x="tsend + coffset A" in exI)
  apply auto
  apply (rule_tac f="MPair (Agent P)" in HOL.arg_cong) (* XXX: add intro rule for this *)
  apply (rule_tac f="Real" in HOL.arg_cong)
  apply auto
done

interpretation PROTOCOL_DELTA_EXEC Crypt Nonce MPair Hash Number
                                   parts subterms DM LowHamXor Xor components
                                   initStateMd Key md_proto sys
by unfold_locales

subsection {* Direct Definition *}

inductive_set
  mdb :: "(msg trace) set"
 where
   Nil [intro] : "[] \<in> mdb"
 | Fake: 
   "\<lbrakk> tr \<in> mdb; t >=  maxtime tr;
      X \<in> DM (Intruder I) (knowsI (Intruder I) tr) \<rbrakk>
   \<Longrightarrow> (t, Send (Tx (Intruder I) j) X [])  # tr \<in> mdb"

 | Con : 
   "\<lbrakk> tr \<in> mdb; trecv >= maxtime tr;
      \<forall>X\<in>components {M}.
        \<exists>tsend A i M' L.
           \<exists>Y\<in>components {M'}.
              (tsend, Send (Tx A i) M' L) \<in> set tr \<and>
              cdistM (Tx A i) (Rx B j) = Some tab \<and> tsend + tab \<le> trecv \<and> Xor X Y \<in> LowHamXor \<rbrakk>
    \<Longrightarrow> (trecv, Recv (Rx B j) M) # tr \<in> mdb"

 (* Protocol rules for Brands-Chaum variant *)
 | MD1:  
   "\<lbrakk> tr \<in> mdb; t >=  maxtime tr;
      \<not> (used tr (Nonce (Honest V) NV)) \<rbrakk>
   \<Longrightarrow> (t, Send (Tr (Honest V)) (Nonce (Honest V) NV) []) # tr \<in> mdb"

 | MD2:
   "\<lbrakk> tr \<in> mdb; tsend >=  maxtime tr;
      (trec, Recv (Rec (Honest P)) NV) \<in> set tr;
      \<not> (used tr (Nonce (Honest P) NP)) \<rbrakk>
   \<Longrightarrow> (tsend, Send (Tr (Honest P))
                   (Xor NV  (Hash \<lbrace> Nonce (Honest P) NP, Agent (Honest P)\<rbrace>))
                   [NV, Nonce (Honest P) NP])
      # tr \<in> mdb"

 | MD3:  
   "\<lbrakk> tr \<in> mdb; tsend >=  maxtime tr;
      (trec, Recv (Rec (Honest P)) NV) \<in> set tr;
     (tsend1, Send (Tr (Honest P))
                   (Xor NV (Hash \<lbrace> Nonce (Honest P) NP, Agent (Honest P)\<rbrace> ))
                   [NV, Nonce (Honest P) NP])
     \<in> set tr \<rbrakk>
   \<Longrightarrow> (tsend,
       Send (Tr (Honest P))
            (Crypt (priSK (Honest P))
                   \<lbrace> NV, \<lbrace> Nonce (Honest P) NP, Agent V\<rbrace>\<rbrace>) [])
       # tr \<in> mdb"

 | MD4:
   "\<lbrakk> tr \<in> mdb; tdone \<ge> maxtime tr;
      (trec2, Recv (Rec (Honest V))
                  ( Crypt (priSK P)
                   \<lbrace> Nonce (Honest V) NV, \<lbrace> NP, Agent (Honest V)\<rbrace>\<rbrace>))
      \<in> set tr;
      (trec1, Recv (Rec (Honest V)) (Xor (Nonce (Honest V) NV) (Hash \<lbrace> NP,  Agent P\<rbrace>)))
      \<in> set tr;
      (tsend, Send (Tr (Honest V)) (Nonce (Honest V) NV) []) \<in> set tr \<rbrakk>
  \<Longrightarrow> (tdone, Claim (Honest V) \<lbrace>Agent P, Real ((trec1 - tsend) * vc/2)\<rbrace>) # tr
     \<in> mdb"

text {* obtain a simpler induction rule for protocol since it is executable and deltaonly *}

lemmas proto_induct =
  sys.induct [simplified derivable_removable remove_occursAt timetrans_removable]

subsection {* Equality for direct and parameterized Definition *}

text {* We now show that both inductive definitions define the
        same set of traces. *}

lemma abstr_equal: "mdb = sys"
proof auto
  fix tr
  assume r: "tr \<in> sys"
  show "tr \<in> mdb" using r
  proof (induct tr rule: proto_induct)
    case 1 with prems show ?case by auto
  next
    case 2 with prems show ?case by (auto intro: mdb.Nil)
  next
    case 4 with prems show ?case apply - apply (rule  mdb.Con) apply auto done
  next
    case 3 with prems show ?case by (auto intro: mdb.Fake)
  next
    case 5
    thus ?case
      apply (auto simp add: md_defs)
      apply (auto intro!: mdb.MD1 mdb.MD2 mdb.MD3 [simplified] mdb.MD4 simp add: usedI_def)
      apply (auto simp add: mem_def)
      done
  qed
next
  fix tr
  assume r: "tr \<in> mdb"
  show "tr \<in> sys" using r
  proof(induct tr rule: mdb.induct)
    case Nil
    with prems show ?case by auto
  next
    case (Fake tr ts X I j)
    with prems show ?case by (auto intro: sys.Fake)
  next
    case (Con tr)
    with prems show ?case apply - apply (rule sys.Con) apply auto done
  next
    case (MD1 tr ts C NA)
    with prems have "(ts,createEv C (SendEv 0 []) (Nonce (Honest C) NA)) # tr \<in> sys"
      apply -
      apply (rule_tac step=md1 in sys_Proto_exec)
      apply force
      apply force
      apply force
      apply (force simp add: md_proto_def)
      apply (simp add: md1_def)
      apply (simp add: usedI_def)
      apply (auto simp add: mem_def)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD2 tr tsend trecv P NV NP)
    with prems have
      "(tsend, 
        createEv P
                 (SendEv 0 [NV, Nonce (Honest P) NP])
                 (Xor NV (Hash \<lbrace> Nonce (Honest P) NP, Agent (Honest P)\<rbrace>)  ))
       # tr \<in> sys"
      apply - apply (rule_tac step=md2 in sys_Proto)
      apply (auto simp add: md_defs usedI_def)
      apply (auto simp add: mem_def)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD3 tr tsend trecv P NV tsend1 NP V)
    with prems have
      "(tsend,
        createEv P (SendEv 0 [])
                   (Crypt (priSK (Honest P))
                        \<lbrace>NV, \<lbrace>Nonce (Honest P) NP, Agent V\<rbrace>\<rbrace>)) # tr \<in> sys"
      apply - apply (rule_tac step=md3 in sys_Proto)
      apply (auto simp add: md_defs)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (MD4 tr tdone trec2 V P NV NP trec1 tsend)
    with prems have
      "(tdone, createEv V ClaimEv \<lbrace>Agent P, Real ((trec1 - tsend) * vc/2)\<rbrace>) # tr \<in> sys"
      apply - apply (rule_tac step=md4 in sys_Proto)
      apply (auto simp add: md_defs)
      apply (intro exI conjI)
      apply auto
      done
    thus ?case by (auto simp add: createEv.psimps)
  qed
qed

lemmas [simp,intro] = abstr_equal [THEN sym]

subsection {* Some invariants capturing the Behavior of honest Agents *}

lemma nonce_fresh_challenge:
  assumes mdb: "tr \<in> mdb" and
          send: "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) []) \<in> set tr"
  shows   "Nonce (Honest A) NA
           \<notin> usedI (beforeEvent (ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) []) tr)"
  using prems(1-)
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend) with Fake.hyps prems show ?case by auto
next
  case (Con tr tc C mc D tab) with Con.hyps prems show ?case by auto
next
  case (MD1 tr t V' NV)
  show ?case proof cases
    assume "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) []) =
            (t, Send (Tr (Honest V')) (Nonce (Honest V') NV) [])"
    thus ?case using MD1.hyps prems
      apply (auto simp add: usedI_def)
      by (simp add: mem_def)
  next
    assume "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) []) \<noteq>
            (t, Send (Tr (Honest V')) (Nonce (Honest V') NV)  [])"
    thus ?case using MD1.hyps prems by auto
  qed
next
  case (MD2 tr tsend trec P' NV NP)
  thus ?case using MD2.hyps prems by auto
next
  case (MD3 tr tsend trec P' NV tsend1 NP V')
  with MD3.hyps prems show ?case by auto
next
  case (MD4 tr t trec2 V' P' NV NP trec1 tsend)
  with MD4.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
qed

lemma nonce_fresh_response:
  assumes mdb: "tr \<in> mdb" and
          send: "(ta, Send (Tx (Honest A) i) (Xor NV (Hash \<lbrace> NP, Agent P \<rbrace>))
                                             [NV, NP]) \<in> set tr"
  shows   "
           (\<exists> NA.
             P = Honest A \<and>
             NP = Nonce (Honest A) NA \<and>
             Nonce (Honest A) NA
             \<notin> usedI (beforeEvent
                        (ta, Send (Tx (Honest A) i) (Xor NV (Hash \<lbrace> Nonce (Honest A) NA,  Agent (Honest A) \<rbrace>))
                                                    [NV, Nonce (Honest A) NA]) tr))"
  using mdb send
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend) with Fake.hyps prems show ?case by auto
next
  case (Con tr tc C mc D tab) with Con.hyps prems show ?case by auto
next
  case (MD3 tr tsend trec P' NV tsend1 NP V')
  with MD3.hyps prems show ?case by auto
next
  case (MD4 tr t trec2 V' P' NV NP trec1 tsend)
  with MD4.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
next
  case (MD1 tr t V NV) thus ?case by auto
next
  case (MD2 tr tsend trec C ND NC)
  let ?eva = "(ta, Send (Tx (Honest A) i) (Xor NV (Hash \<lbrace>NP,  Agent P \<rbrace>)) [NV, NP])"
  let ?newev = "(tsend, Send (Tr (Honest C)) (Xor ND (Hash \<lbrace> Nonce (Honest C) NC, Agent (Honest C)\<rbrace>))
                                             [ND, Nonce (Honest C) NC])"
  show ?case proof cases
    assume eq: "?eva = ?newev"
    thus ?case using MD2.hyps prems eq apply -
      apply (rule_tac x=NC in exI)
      apply (subgoal_tac "NP = Nonce (Honest C) NC") prefer 2
      apply force
      apply (subgoal_tac "P = Honest C")
      prefer 2
      apply simp
      apply (drule Xor_same_arg)
      apply force
      apply (auto simp add: usedI_def)
      apply (force simp: mem_def)
      done
  next
    assume "?eva \<noteq> ?newev"
    hence "?eva \<in> set tr" using `?eva \<in> set (?newev#tr)` by auto
    thus ?case apply -
      apply (frule MD2.hyps(2))
      apply (elim conjE exE)
      apply auto
      done
  qed
qed

lemma nonce_fresh_response2:
  assumes mdb: "tr \<in> mdb" and
          send: "(ta, Send (Tx (Honest A) i) (Xor NV (Hash \<lbrace> Nonce (Honest A) NA, Agent (Honest A)\<rbrace>))
                                                    [NV, Nonce (Honest A) NA])
                 \<in> set tr"
  shows   "Nonce (Honest A) NA
           \<notin> usedI (beforeEvent
                     (ta, Send (Tx (Honest A) i) (Xor NV (Hash \<lbrace> Nonce (Honest A) NA, Agent (Honest A)\<rbrace>))
                                                 [NV, Nonce (Honest A) NA]) tr)"
  using mdb send
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend) with Fake.hyps prems show ?case by auto
next
  case (Con tr tc C mc D tab) with Con.hyps prems show ?case by auto
next
  case (MD3 tr tsend trec P' NV tsend1 NP V')
  with MD3.hyps prems show ?case by auto
next
  case (MD4 tr t trec2 V' P' NV NP trec1 tsend)
  with MD4.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
next
  case (MD1 tr t V NV) thus ?case by auto
next
  case (MD2 tr tsend trec C ND NC)
  let ?eva = "(ta, Send (Tx (Honest A) i) (Xor NV (Hash \<lbrace> Nonce (Honest A) NA,  Agent (Honest A)\<rbrace>))
                                                   [NV, Nonce (Honest A) NA])"
  let ?newev = "(tsend, Send (Tr (Honest C)) (Xor ND (Hash \<lbrace> Nonce (Honest C) NC, Agent (Honest C) \<rbrace>))
                                             [ND, Nonce (Honest C) NC])"
  show ?case proof cases
    assume eq: "?eva = ?newev"
    thus ?case using MD2.hyps prems apply -
      apply (auto simp add: usedI_def)
      apply (auto simp add: mem_def)
      done
  next
    assume "?eva \<noteq> ?newev"
    hence "?eva \<in> set tr" using `?eva \<in> set (?newev#tr)` by auto
    thus ?case apply -
      apply (frule MD2.hyps(2))
      apply auto
      done
  qed
qed


text {* If an honest prover sends a signature, then he has sent the corresponding fastreply before.
        Then we can use nonce fresh response to obtain that the nonce in a fast-reply is
        fresh. *}

lemma sig_send_prover:
  assumes mdb: "tr \<in> mdb"
      and mac: "(tsend,
                 Send (Tx (Honest B) k)
                      (Crypt (priSK (Honest B))
                       \<lbrace> NA, \<lbrace>Nonce (Honest B) NB, Agent A\<rbrace>\<rbrace>) [])
                \<in> set tr"
  shows "(\<exists> tfast. 
            (tfast, Send (Tr (Honest B))
                         (Xor NA (Hash \<lbrace> Nonce (Honest B) NB, Agent (Honest B) \<rbrace>))
                         [NA,Nonce (Honest B) NB]) \<in> set tr)"
  using prems
  apply (induct tr rule: mdb.induct)
  apply (auto)
done

lemma sig_send_prover2:
  assumes mdb: "tr \<in> mdb"
      and mac: "(tsend,
                 Send (Tx (Honest B) k)
                      (Crypt (priSK (Honest B))
                       \<lbrace> NA, \<lbrace>Nonce (Honest B) NB, Agent A\<rbrace>\<rbrace>) [])
                \<in> set tr"
  shows "(\<exists> tfast. 
            (tfast, Send (Tr (Honest B))
                         (Xor NA (Hash \<lbrace>Nonce (Honest B) NB,  Agent (Honest B) \<rbrace>))
                         [NA,Nonce (Honest B) NB]) \<in> set tr \<and>
            Nonce (Honest B) NB
            \<notin> usedI (beforeEvent
                      (tfast, Send (Tr (Honest B))
                         (Xor NA (Hash \<lbrace> Nonce (Honest B) NB,  Agent (Honest B) \<rbrace>))
                         [NA,Nonce (Honest B) NB]) tr))"
  using prems apply -
  apply (frule sig_send_prover)
  apply force
  apply (elim exE)
  apply (rule_tac x=tfast in exI)
  apply (frule nonce_fresh_response2)
by auto

text {* The sigs are always unique because they contain the private
        key of an honest agents and his own nonce contribution *}


lemma sig_msg_originates:
  assumes mdb: "tr \<in> mdb"
  and fsend "(tf, Send (Tx (Honest F) j) mf Lf) \<in> set tr"
  and mfsubterm: "Crypt (priSK (Honest P)) \<lbrace>Nonce (Honest V) NV, \<lbrace>NP', Agent (Honest V)\<rbrace>\<rbrace>
                  \<in> subterms {mf}"
  and ffresh: "Crypt (priSK (Honest P)) \<lbrace>Nonce (Honest V) NV, \<lbrace>NP', Agent (Honest V)\<rbrace>\<rbrace>
               \<notin> used (beforeEvent (tf, Send (Tx (Honest F) j) mf Lf) tr)"
  shows "\<exists> NP. F=P \<and> (NP' = Nonce (Honest P) NP)
               \<and> Lf = []
               \<and> mf = Crypt (priSK (Honest P)) \<lbrace>Nonce (Honest V) NV, \<lbrace>Nonce (Honest P) NP, Agent (Honest V)\<rbrace>\<rbrace>" using prems
proof (induct tr rule: mdb.induct)
  case (Fake tr mintr I tsend) with Fake.hyps prems show ?case by auto
next
  case (Con tr tc C mc D tab) with Con.hyps prems show ?case by auto
next
  case (MD4 tr t trec2 V' P' NV NP trec1 tsend)
  with MD4.hyps prems show ?case by auto
next
  case Nil thus ?case by auto
next
  case (MD1 tr t V NV)
  show ?case using prems by auto
next
  case (MD2 tr tsend trec P' NV' NP)
  let ?msg = "Xor NV' (Hash \<lbrace> Nonce (Honest P') NP,  Agent (Honest P') \<rbrace>)"
  let ?ev = "(tsend, Send (Tr (Honest P')) ?msg [NV',Nonce (Honest P') NP])"
  let ?sig = "Crypt (priSK (Honest P)) \<lbrace>Nonce (Honest V) NV, \<lbrace>NP', Agent (Honest V)\<rbrace>\<rbrace>"
  show ?case proof cases
    assume "(tf, Send (Tx (Honest F) j) mf Lf) = ?ev"
    hence sub: "?sig \<in> subterms {NV'}" using prems
      apply auto
      apply (drule sig_subterms)
      apply (drule subterms.singleton)
      apply auto
      done
    thus ?thesis
    proof cases
      assume "?ev \<in> set tr"
      thus ?thesis using prems(6-) by auto
    next
      assume "?ev \<notin> set tr"
      thus ?thesis using sub prems(6-) apply -
        apply auto
        apply (frule crypt_components_subterm)
        apply auto
        apply (drule_tac X="M" in send_before_recv[simplified])
        apply assumption
        apply assumption
        apply auto
        apply (drule distort_LowHam)
        apply auto
        apply (drule crypt_not_LowHam)
        apply simp
        apply (drule_tac Y=Y in subterms_component_trans)
        apply assumption
        apply (drule Send_imp_parts_used) 
        apply assumption
        apply simp
        done
    qed
  next
    assume "(tf, Send (Tx (Honest F) j) mf Lf) \<noteq> ?ev"
    thus ?thesis using prems by auto
  qed
next
  case (MD3 tr tsend trec P' NV' tsend1 NP V')
  let ?msg = "Crypt (priSK (Honest P')) \<lbrace>NV', \<lbrace>Nonce (Honest P') NP, Agent V'\<rbrace>\<rbrace>"
  let ?ev = "(tsend, Send (Tr (Honest P')) ?msg [])"
  let ?sig = "Crypt (priSK (Honest P)) \<lbrace>Nonce (Honest V) NV, \<lbrace>NP', Agent (Honest V)\<rbrace>\<rbrace>"
  show ?case proof cases
    assume "(tf, Send (Tx (Honest F) j) mf Lf) = ?ev"
    with `?sig \<in> subterms {mf}`
    have or: "?sig \<in> subterms {NV'} \<or> (P'=P \<and> V'=Honest V \<and> NV' = Nonce (Honest V) NV
                                      \<and> NP' = Nonce (Honest P') NP)"
      by auto
    thus ?thesis proof cases
      assume "?sig \<in> subterms {NV'}"
      show ?thesis proof cases
        assume "?ev \<in> set tr"
        thus ?thesis using prems(6-) by auto
      next
        assume "?ev \<notin> set tr"
        thus ?thesis using prems(6-) apply -
          apply (frule_tac S="{NV'}" in crypt_components_subterm)
          apply simp
          apply (elim bexE)
          apply (drule_tac X="M" in send_before_recv[simplified])
          apply assumption
          apply assumption
          apply clarsimp
          apply (drule distort_LowHam)
          apply clarsimp
          apply (drule crypt_not_LowHam)
          apply assumption
          apply (drule_tac Y=Y in subterms_component_trans)
          apply assumption
          apply (drule Send_imp_parts_used)  back
          apply assumption
          apply simp
          done
      qed
    next
      assume "\<not> (?sig \<in> subterms {NV'})"
      hence "P'=P \<and> V'=Honest V \<and> NV' = Nonce (Honest V) NV \<and> NP' = Nonce (Honest P') NP"
        using or by auto
      thus ?thesis using prems(6-) by auto
    qed
  next
    assume "(tf, Send (Tx (Honest F) j) mf Lf) \<noteq> ?ev"
    thus ?thesis using prems by auto
  qed
qed

lemma originate_unique:
  assumes "m \<notin> used (beforeEvent (ta, Send TA ma La) tr)"
  and     "m \<notin> used (beforeEvent (tb, Send TB mb Lb) tr)"
  and     "(tb, Send TB mb Lb) \<noteq> (ta, Send TA ma La)"
  and     "(tb, Send TB mb Lb) \<in> set tr"
  and     "(ta, Send TA ma La) \<in> set tr"
  and     " m \<in> subterms {ma}"
  shows   "m \<notin> subterms {mb}" using prems
  apply (induct tr)
  apply simp
  apply (case_tac "a=(ta, Send TA ma La) \<and> a \<notin> set tr")
  apply (elim conjE)
  apply simp
  apply (case_tac "m \<in> subterms {mb}") prefer 2
  apply force
  apply (subgoal_tac "(tb, Send TB mb Lb) \<in> set tr") prefer 2
  apply force
  apply (frule_tac Y=m in Send_imp_parts_used)
  apply force
  apply force
  apply (case_tac "a=(tb, Send TB mb Lb) \<and> a \<notin> set tr")
  apply (elim conjE)
  apply simp
  apply (subgoal_tac "(ta, Send TA ma La) \<in> set tr") prefer 2
  apply force
  apply (frule_tac Y=m in Send_imp_parts_used)
  apply force
  apply force
  apply auto
done


lemma components_factors:
  "factors m \<noteq> {m} \<Longrightarrow> components {m} = {m}"
  apply (case_tac "Rep_msg m")
  apply (auto simp add: factors_def components_def)
  apply (drule_tac f=Abs_msg in HOL.arg_cong, auto simp add: Rep_msg_inverse)+
  done

lemma ffactors_fcomponents:
  "components {m} \<noteq> {m} \<Longrightarrow> factors m = {m}"
  apply (case_tac "Rep_msg m")
  apply (auto simp add: factors_def components_def)
  apply (drule_tac f=Abs_msg in HOL.arg_cong, auto simp add: Rep_msg_inverse)+
  done


lemma freshNonce_dishonestAgent_send_recv:
  assumes "tr \<in> mdb"
  and     "(t, Send (Tx (Honest A) i) m L) \<in> set tr \<or> (t, Recv (Rx (Honest A) i) m) \<in> set tr"
  and     "X \<in> components {m}"
  and     "Hash \<lbrace> NC, Agent (Intruder I)\<rbrace> \<in> factors X"
  and     "Nonce (Honest B) NB \<in> factors X"
  and     "(tnonce, Send (Tr (Honest B)) (Nonce (Honest B) NB) []) \<in> set tr"
  and     "Nonce (Honest B) NB
           \<notin> usedI (beforeEvent (tnonce, Send (Tr (Honest B)) (Nonce (Honest B) NB) []) tr)"
  shows   "\<exists> I'. t - tnonce \<ge> cdistl (Honest B) (Intruder I') + cdistl (Intruder I') (Honest A)"
  using prems
proof (induct tr arbitrary: A B trec t m L i X rule: mdb.induct)
  case (Fake tr mintr I tsend)
  hence "(t, Send (Tx (Honest A) i) m L) \<in> set tr \<or> (t, Recv (Rx (Honest A) i) m) \<in> set tr" by auto
  with Fake.hyps prems show ?case by (auto)
next
  case (MD1 tr tc V NV)
  have "Hash \<lbrace> NC, Agent (Intruder I)\<rbrace> \<notin> factors (Nonce (Honest V) NV)" by auto
  hence "(t, Send (Tx (Honest A) i) m L) \<in> set tr \<or> (t, Recv (Rx (Honest A) i) m) \<in> set tr"
    using MD1.prems by auto
  also have "(tnonce, Send (Tr (Honest B)) (Nonce (Honest B) NB)  []) \<in> set tr"
    using MD1.prems `tr \<in> mdb`
    apply (auto simp add: usedI_def split: split_if_asm)
    apply (drule_tac Y="Nonce (Honest B) NV" in Send_imp_parts_used)
    apply auto
    apply (drule factors_imp_subterms) back
    apply (drule_tac Y=X in subterms_component_trans)
    apply simp
    apply simp
    apply (drule factors_imp_subterms) back
    apply (drule_tac Y=X in subterms_component_trans)
    apply simp
    apply (frule_tac S="{m}" in nonce_components_subterm)
    apply (elim bexE)
    apply (drule_tac X="m" in send_before_recv[simplified])
    apply assumption
    apply assumption
    apply auto
    apply (drule distort_LowHam)
    apply clarsimp
    apply (drule nonce_not_LowHam)
    apply assumption
    apply (drule_tac Y=Y in subterms_component_trans)
    apply assumption
    apply (drule Send_imp_parts_used)
    apply assumption
    apply simp
    done
  ultimately show ?case using MD1.hyps prems  apply -
    apply (rule MD1.hyps(2))
    apply (simp)+
    done
next
  case (MD4 tr t A NA)
  hence "(t, Send (Tx (Honest A) i) m L) \<in> set tr \<or> (t, Recv (Rx (Honest A) i) m) \<in> set tr"
    using prems by auto
  with MD4.hyps prems show ?case apply -
    apply (rule MD4.hyps(2))
    apply simp+
    done
next
  case Nil
  show ?case using prems by auto
next
  case (MD3 tr tsend trec D NE tsend1 NF C)

  let ?sigm = "Crypt (priSK (Honest D)) \<lbrace>NE, \<lbrace>Nonce (Honest D) NF, Agent C\<rbrace>\<rbrace>"
  let ?ev = "(tsend, Send (Tr (Honest D)) ?sigm [])"
  show ?case proof cases
    assume "(t, Send (Tx (Honest A) i) m L) \<in> set (?ev#tr)"
    show ?case proof cases
      assume eveq: "(t, Send (Tx (Honest A) i) m L) = ?ev"
      thus ?thesis using prems by auto
    next
      assume "(t, Send (Tx (Honest A) i) m L) \<noteq> ?ev"
      hence "(t, Send (Tx (Honest A) i) m L) \<in> set tr" using prems by auto
      with MD3.hyps prems show ?thesis apply -
        apply (rule MD3.hyps(2))
        apply force
        apply force
        apply force
        apply force
        apply force
        apply force
        done
    qed
  next
    assume "(t, Send (Tx (Honest A) i) m L) \<notin> set (?ev#tr)"
    hence "(t, Recv (Rx (Honest A) i) m) \<in> set tr" using prems by auto
    with MD3.hyps prems show ?case apply -
      apply (rule MD3.hyps(2))
      apply force
      apply force
      apply force
      apply force
      apply force
      apply force
      done
  qed
next
  case (MD2 tr tsend trec P NV NP)
  let ?hash = "Hash \<lbrace> Nonce (Honest P) NP,  Agent (Honest P)\<rbrace>"
  let ?fr = "Xor NV ?hash"
  let ?ev = "(tsend, Send (Tr (Honest P)) ?fr [NV, Nonce (Honest P) NP])"
  show ?case proof cases
    assume "(t, Send (Tx (Honest A) i) m L) \<in> set (?ev#tr)"
    show ?case proof cases
      assume eveq: "(t, Send (Tx (Honest A) i) m L) = ?ev"

      have hashFactor: "?hash \<notin> factors NV"
        using `\<not> MESSAGE_DERIVATION.used subterms tr (Nonce (Honest P) NP)` 
              `(trec, Recv (Rec (Honest P)) NV) \<in> set tr` `tr \<in> mdb`
        apply -
        apply auto
        apply (drule factors_imp_subterms)
        apply (subgoal_tac "Nonce (Honest P) NP \<in> subterms {NV}") prefer 2
        apply (rule_tac G="{?hash}" in subterms.trans)
        apply force
        apply force
        apply (drule nonce_components_subterm)
        apply auto
        apply (drule send_before_recv[simplified])
        apply simp
        apply simp
        apply auto
        apply (drule distort_LowHam)
        apply auto
        apply (drule nonce_not_LowHam)
        apply simp
        apply (drule_tac Y=Y in subterms_component_trans)
        apply simp
        apply (drule Send_imp_parts_used)
        apply (auto simp add: mem_def)
        done

      have NVzero: "NV \<noteq> Zero"using `X \<in> components {m}` `Hash \<lbrace>NC, Agent (Intruder I)\<rbrace> \<in> factors X` eveq
        by auto

      have "?hash \<in> factors (Xor NV ?hash)" using hashFactor
        apply -
        apply (drule factors_Xor_hash_not_subterm)
        by auto

      hence XorNotPair: "\<forall> X Y. Xor NV ?hash \<noteq> MPair X Y" by auto

      hence "components {Xor NV ?hash} = {Xor NV ?hash}" using hashFactor apply -
        apply (drule components_non_pair)
        by auto
        
      hence Xeq: "X = Xor NV ?hash" using prems by auto
      
      hence hashNV: "Hash \<lbrace>NC, Agent (Intruder I)\<rbrace> \<in> factors NV"
        using `Hash \<lbrace>NC, Agent (Intruder I)\<rbrace> \<in> factors X` eveq hashFactor apply-
        apply simp
        apply (drule factors_Xor_hash_not_subterm)
        apply auto
        done

      hence nonceNV: "Nonce (Honest B) NB \<in> factors NV"
        using `Nonce (Honest B) NB \<in> factors X` hashFactor eveq Xeq apply-
        apply-
        apply simp
        apply (drule factors_Xor_hash_not_subterm)
        apply auto
        done

      have "components {NV} = {NV}" using nonceNV hashNV apply -
        apply (rule components_factors)
        apply auto
        done
      
      hence "NV \<in> components {NV}" by auto

      thus ?case using prems(8-)
        apply auto
        apply (subgoal_tac "\<exists>I'. cdistl (Honest B) (Intruder I') + cdistl (Intruder I') (Honest P) \<le> trec - tnonce") prefer 2
        apply (rule prems(9))
        apply (rule disjI2)
        apply simp
        apply assumption
        apply (rule hashNV)
        apply (rule nonceNV)
        apply simp
        apply simp
        apply (elim exE)
        apply (rule_tac x=I' in exI)
        apply (subgoal_tac "trec \<le> tsend")
        apply force
        apply (rule maxtime_geq_elem)
        apply auto
        done
    next
      assume "(t, Send (Tx (Honest A) i) m L) \<noteq> ?ev"
      thus ?thesis using MD2.prems MD2.hyps apply -
        apply (rule MD2.hyps(2))
        apply force
        apply force
        apply force
        apply force
        apply force
        apply (force simp add: usedI_def)
        done
    qed
  next
    assume "(t, Send (Tx (Honest A) i) m L) \<notin> set (?ev#tr)"
    thus ?thesis using MD2.prems MD2.hyps apply -
      apply (rule MD2.hyps(2))
      apply force
      apply force
      apply force
      apply force
      apply force
      apply (force simp add: usedI_def)
      done
  qed
next
  print_cases
  case (Con tr trecv_l M_l B_l j_l tab_l)

  let ?evrecv = "(trecv_l, Recv (Rx B_l j_l) M_l)"
  show ?case proof cases
    assume "(t, Recv (Rx (Honest A) i) m) \<in> set (?evrecv#tr)"
    show ?case proof cases
      assume "(t, Recv (Rx (Honest A) i) m) = ?evrecv"
      hence mceq: "m = M_l" and Deq: "B_l=Honest A" and trecveq: "t=trecv_l" by auto
      
      obtain tsend E u M' L' Y where
        p1: "(tsend, Send (Tx E u) M' L') \<in> set tr" and
        p2: "Y\<in>components {M'}" and
        p3: "Xor X Y \<in> LowHamXor" and
        p4: "cdistM (Tx E u) (Rx B_l j_l) = Some tab_l" and
        p5: "tsend + tab_l \<le> trecv_l"
        using prems
        apply -
        apply (erule ballE)
        apply auto
        done
      show ?thesis proof cases
        assume "\<exists> I'. E = Intruder I'"
        then obtain I' where I: "E = Intruder I'" by auto
        have cdist2: "cdistl (Honest B) (Intruder I') \<le> tsend - tnonce" using prems p1 p2 p3 p4 p5 apply -
          apply (rule_tac A="Honest B" and NA=NB and i=0 and ma="(Nonce (Honest B) NB)" and tr=tr
                 and mb="M'"in fresh_nonce_earliest_send)
          apply force
          apply force
          apply (simp add: usedI_def)
          apply force defer
          apply simp
          apply simp
          apply (rule subterms_component_trans) defer
          apply simp
          apply (drule factors_imp_subterms)
          apply (drule distort_LowHam)
          apply (elim bexE)
          apply simp
          apply (drule_tac d=d and m=Y and A="Honest B" and N="NB" in nonce_not_LowHam)
          apply (erule factors_imp_subterms)
          apply simp
          done
        have cdist3: "cdistl (Intruder I') (Honest A) \<le> t - tsend" using p1 p4 p5 trecveq I Con.hyps Deq
          apply auto
          apply (frule noflt_some2)
          by auto
        hence "t - tnonce \<ge> cdistl (Honest B) (Intruder I') + cdistl (Intruder I') (Honest A)" using cdist2
          by auto
        thus ?thesis by auto
      next
        assume "\<not> (\<exists> I'. E = Intruder I')"
        then obtain F where F: "E = Honest F" apply (case_tac E) by auto
        hence "\<exists>I'. cdistl (Honest B) (Intruder I') + cdistl (Intruder I') (Honest F) \<le> tsend - tnonce"
          using Con.prems Con.hyps mceq Deq trecveq p1 p2 p3 p4 p5 apply -
          apply (rule_tac m=M' in Con.hyps(2))
          apply force defer defer defer
          apply force
          apply (force simp add: usedI_def)
          apply assumption

          apply (drule distort_LowHam)
          apply (elim bexE)
          apply simp
          apply (drule factors_Xor)
          apply clarsimp
          apply (drule factors_LowHam)
          apply simp
          apply force

          apply (drule distort_LowHam)
          apply (elim bexE)
          apply simp
          apply (drule factors_Xor) back
          apply clarsimp
          apply (drule factors_LowHam)
          apply simp
          apply force
          done
        then obtain I where cdist1:
          "cdistl (Honest B) (Intruder I) + cdistl (Intruder I) (Honest F) \<le> tsend - tnonce"
          by auto
        have cdist2: "cdistl (Honest F) (Honest A) \<le> t - tsend" using trecveq Con.hyps Deq F p1 p4 p5
          apply auto
          apply (frule noflt_some2)
          by auto
        have cdist3: "cdistl (Intruder I) (Honest F) + cdistl (Honest F) (Honest A) \<ge>
                      cdistl (Intruder I) (Honest A)"
          by (rule cdistl_triangle)

        have "t - tnonce \<ge> (cdistl (Honest B) (Intruder I)  + cdistl (Intruder I) (Honest F))
                               + cdistl (Honest F) (Honest A)" using cdist1 cdist2 apply -
          by auto
        then also have "... \<ge> cdistl (Honest B) (Intruder I) + cdistl (Intruder I) (Honest A)" using cdist3
          by auto
        ultimately have "t - tnonce \<ge> cdistl (Honest B) (Intruder I) + cdistl (Intruder I) (Honest A)"
          by auto
        thus ?thesis by auto
      qed
    next
      assume "(t, Recv (Rx (Honest A) i) m) \<noteq> ?evrecv"
      hence "(t, Recv (Rx (Honest A) i) m) \<in> set tr" using prems by auto
      thus ?thesis using Con.prems apply -
        apply (rule Con.hyps(2))
        by auto
    qed
  next 
    assume "(t, Recv (Rx (Honest A) i) m) \<notin> set (?evrecv#tr)"
    hence "(t, Send (Tx (Honest A) i) m L) \<in> set tr" using prems
      apply -
      apply auto
      done
    thus ?thesis using Con.prems apply -
      apply (rule Con.hyps(2))
      by auto
  qed
qed

subsection {* Security proof for Honest Provers *}


lemma mdb_secure:
  assumes mdb: "tr \<in> mdb"
  and believe: "(tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>) \<in> set tr"
  shows   "d \<ge> pdist (Honest V) (Honest P)" using prems
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with Fake.hyps prems show ?case by (auto)
next
  case (Con tr tc C mc D tab)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
next
  case (MD1 tr t V' NV)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD1.hyps prems show ?case by auto
next
  case (MD2 tr tsend trec P' NV NP)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD2.hyps prems show ?case by (auto)
next
  case (MD3 tr tsend trec P' NV tsend1 NP V')
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD3.hyps prems show ?case by auto
next
  -- {* the only nontrivial case since it adds Claim events *}
  case (MD4 tr t trec2 V' P' NV NP trec1 tsend)
  let ?x = "(t, Claim (Honest V') \<lbrace>Agent P', Real ((trec1 - tsend)*vc/2)\<rbrace>)"
  and ?ev = "((tdone, Claim (Honest V) \<lbrace>Agent (Honest P), Real d\<rbrace>))"
  show ?case proof cases
    -- {* the added event is the Claim event from the premise,
          the other case follows trivially from the IH *}
    assume "?x = ?ev"
    hence Veq: "V'=V" and Peq: "P'=Honest P" and deq: "d=(trec1 - tsend)*vc/2" by auto
    let ?mmac = "\<lbrace>Nonce (Honest V) NV, \<lbrace>NP, Agent (Honest P)\<rbrace>\<rbrace>"
    let ?sigmsg = "Crypt (priSK (Honest P))
                       \<lbrace>Nonce (Honest V) NV, \<lbrace>NP, Agent (Honest V)\<rbrace>\<rbrace>" and
        ?fastmsg = "Xor (Nonce (Honest V) NV) (Hash \<lbrace> NP, Agent (Honest P)\<rbrace>)"
    have NC_fresh:
      "Nonce (Honest V) NV
       \<notin> usedI (beforeEvent (tsend, Send (Tr (Honest V)) (Nonce (Honest V) NV) []) tr)"
      using prems `tr \<in> mdb` apply -
      by (rule nonce_fresh_challenge, auto)
    -- {* We first handle the trivial case where V runs the protocol with himself *}
    show ?case proof cases
      assume PeqV: "P=V"
      
      let ?XOR = "Xor (Nonce (Honest V') NV) (Hash \<lbrace>NP, Agent P'\<rbrace>)"
      
      have "Nonce (Honest V') NV \<in> subterms {?XOR}"
        apply (auto simp add: subterms_xor_nonce_hash)
        done
      then obtain X where "X \<in> components {?XOR}"
                    and "Nonce (Honest V') NV \<in> subterms {X}" using prems apply -
        apply (drule nonce_components_subterm)
        apply auto
        done

      with prems(3-) have
          " \<exists>A i tsend L M'.
              \<exists>Z\<in>components {M'}.
                (tsend, Send (Tx A i) M' L) \<in> set tr \<and>
                Xor X Z \<in> LowHamXor \<and> cdistM (Tx A i) (Rx (Honest V') 0) \<noteq> None
                \<and> tsend \<le> trec1 - cdist (Tx A i)  (Rx (Honest V') 0)"
        apply -
        apply (rule send_before_recv)
        apply auto
        done
      
        then obtain C u tfsend Lc M' Z
	  where     p1: "(tfsend, Send (Tx C u) M' Lc) \<in> set tr"
               and  p2: "Xor X Z \<in> LowHamXor" 
               and p3:  "Z \<in> components {M'}"
               and p4:  "cdistM (Tx C u) (Rx (Honest V') 0) \<noteq> None"
               and p5:  "tfsend \<le> trec1 - cdist (Tx C u)  (Rx (Honest V') 0)"
          by auto

        have p6: "Nonce (Honest V') NV \<in> subterms {M'}"
          using p1 p2 p3 `Nonce (Honest V') NV \<in> subterms {X}` apply -
            apply (drule distort_LowHam)
            apply auto
            apply (drule nonce_not_LowHam)
            apply simp
            apply (rule subterms_component_trans)
            apply auto
            done
      
      hence "tfsend \<le> trec1" using p4 p5 apply auto
        apply (rule_tac y="trec1 - y" in order_trans)
        apply (auto simp add: cdist_def)
        by (rule cdistnoneg_some)
      show ?thesis proof (rule ccontr)
        assume "\<not> pdist (Honest V) (Honest P) \<le> d"
        hence "trec1 < tsend" using PeqV apply -
          apply (auto simp add: minusv_def CauchySchwarz.norm_def
                 pdist_def vlen_def ith_def deq)
          apply (erule contrapos_np)
          apply (subgoal_tac "trec1 \<ge> tsend")
          apply (rule mult_nonneg_nonneg)
          apply (auto intro: vc_pos order_less_imp_le)
          done
      hence "tfsend < tsend" using `tfsend \<le> trec1` by auto
      hence "(tfsend, Send (Tx C u) M' Lc)
                 \<in> set (beforeEvent (tsend, Send (Tr (Honest V)) (Nonce (Honest V) NV) []) tr)"
        using prems apply -
        apply (rule beforeEvent_earlier)
        apply auto
        done
      thus False using NC_fresh p6 Veq apply -
        apply (rotate_tac 2)
        apply (erule contrapos_np)
        apply (auto simp add: usedI_def)
        apply (erule Send_imp_parts_used)
        apply auto
        done
    qed
  next
    assume PnotV: "P \<noteq> V"
    have sig_recv:  "(trec2,Recv (Rec (Honest V)) ?sigmsg) \<in> set tr" using prems by auto
    have fast_recv: "(trec1, Recv (Rec (Honest V)) ?fastmsg) \<in> set tr" using prems by auto
    
    have "?sigmsg \<in> components {?sigmsg}"
      by auto

    with prems(3-) have
          " \<exists>A i tsend L M'.
              \<exists>Z\<in>components {M'}.
                (tsend, Send (Tx A i) M' L) \<in> set tr \<and>
                Xor ?sigmsg Z \<in> LowHamXor \<and> cdistM (Tx A i) (Rx (Honest V') 0) \<noteq> None
                \<and> tsend \<le> trec2 - cdist (Tx A i)  (Rx (Honest V') 0)"
      apply -
      apply (rule send_before_recv)
      apply auto
      done
      
    then obtain C u tfsend Lc M' Z
      where     p1: "(tfsend, Send (Tx C u) M' Lc) \<in> set tr"
           and  p2: "Xor ?sigmsg Z \<in> LowHamXor" 
           and  p3: "Z \<in> components {M'}"
      by auto

    have p4: "?sigmsg \<in> subterms {M'}"
      using p1 p2 p3 apply -
      apply (subgoal_tac "?sigmsg \<in> subterms {Z}")
      apply (rule_tac Y=Z in subterms_component_trans)
      apply simp
      apply simp
      apply (drule distort_LowHam)
      apply (elim bexE)
      apply simp
      apply (subgoal_tac "Xor Z d = ?sigmsg") prefer 2
      apply force
      apply (thin_tac "Crypt ?k ?m = ?u")
      apply simp
      apply (subgoal_tac "?sigmsg \<in> factors (Xor Z d)") prefer 2
      apply simp
      apply (drule factors_Xor)
      apply auto
      apply (erule factors_imp_subterms)
      apply (auto dest: factors_LowHam)
      done

    hence "\<exists>tf mf j Lf.
             (tf, Send (Tx (Honest P) j) mf Lf) \<in> set tr
             \<and> ?sigmsg \<in> subterms {mf}
             \<and> ?sigmsg \<notin> used (beforeEvent (tf, Send (Tx (Honest P) j) mf Lf) tr)" using prems p4 apply -
      apply (rule_tac tc=tfsend in crypt_originates)
      apply force prefer 2
      apply force
      apply force
      done
    then obtain tf mf j Lf where ftr: "(tf, Send (Tx (Honest P) j) mf Lf) \<in> set tr"
                           and mfsubterm: "?sigmsg \<in> subterms {mf}"
                             and ffresh: "?sigmsg
                                          \<notin> used (beforeEvent (tf, Send (Tx (Honest P) j) mf Lf) tr)"
      by auto
    hence ex: "\<exists> NPP. (P = P) \<and> NP = Nonce (Honest P) NPP
               \<and> Lf = []
               \<and> mf = Crypt (priSK (Honest P))
                            \<lbrace>Nonce (Honest V) NV, \<lbrace>Nonce (Honest P) NPP, Agent (Honest V)\<rbrace>\<rbrace>" apply -
      apply (rule sig_msg_originates)
      by (auto intro: prems)
    then obtain NPP where ef:
      "(tf, Send (Tx (Honest P) j) (Crypt (priSK (Honest P))
                                         \<lbrace>Nonce (Honest V) NV, \<lbrace>Nonce (Honest P) NPP, Agent (Honest V)\<rbrace>\<rbrace>)
                                   []) \<in> set tr"
      and NPP: "NP = Nonce (Honest P) NPP"
      apply (insert ftr ex)
      by auto
    then obtain tfast where
      fastsend: "(tfast, Send (Tr (Honest P))
                               (Xor (Nonce (Honest V) NV) (Hash \<lbrace> Nonce (Honest P) NPP,  Agent (Honest P)\<rbrace>))
                               [Nonce (Honest V) NV, Nonce (Honest P) NPP]) \<in> set tr" and
      fresh: "Nonce (Honest P) NPP
              \<notin> usedI
              (beforeEvent
                (tfast, Send (Tr (Honest P))
                               (Xor (Nonce (Honest V) NV) (Hash \<lbrace> Nonce (Honest P) NPP,  Agent (Honest P)\<rbrace>))
                               [Nonce (Honest V) NV, Nonce (Honest P) NPP]) tr)"
      using `tr \<in> mdb` apply -
        apply (drule sig_send_prover2)
        apply assumption
        apply (auto intro: prems)
      done
    hence rec1_fast: "trec1 - tfast  >= cdistl (Honest P) (Honest V)" using `tr \<in> mdb` PnotV fastsend
      apply -
      apply (erule_tac NA="NPP" and
             ma="Xor (Nonce (Honest V) NV) (Hash \<lbrace> Nonce (Honest P) NPP, Agent (Honest P)\<rbrace>)" and
             mb="?fastmsg" and i=0 and tr=tr and
             A="Honest P" and B="Honest V" and
             j=0 and tr=tr
             in fresh_nonce_earliest_recv [simplified]) prefer 6
      apply force prefer 4
      apply force prefer 4
      apply (rule fast_recv)
      apply (auto simp add: NPP)
      apply (force simp add: usedI_def)
      apply (auto simp add: subterms_xor_nonce_hash)
      done
    have fast_send: "tfast - tsend >= cdistl (Honest V) (Honest P)" using `tr \<in> mdb` PnotV
      apply -
      apply (erule_tac NA="NV" and i=0 and ma="Nonce (Honest V) NV"
             and mb="Xor (Nonce (Honest V) NV) (Hash \<lbrace> Nonce (Honest P) NPP, Agent (Honest P)\<rbrace>)"
             in fresh_nonce_earliest_send[simplified])
      apply force
      apply (insert NC_fresh, force simp add: usedI_def)
      apply force
      apply (auto intro: prems simp add: Veq)
      apply (auto simp add: Veq[THEN sym] intro: prems)
      apply (auto simp add: subterms_xor_nonce_hash)
      done
    have "2* cdistl (Honest V) (Honest P) \<le> cdistl (Honest V) (Honest P) + cdistl (Honest P) (Honest V)"
      by (auto simp add: cdistl_symm)
    also have "... \<le> tfast - tsend + (trec1 - tfast)" using fast_send rec1_fast by auto
    also have "tfast - tsend + (trec1 - tfast) \<le>  trec1 - tsend" by auto
    finally have "cdistl (Honest V) (Honest P) * 2 \<le>  trec1 - tsend" by auto
    thus ?thesis using deq
      apply (simp add: cdistl_def deq)
      apply (subgoal_tac "(pdist (Honest V) (Honest P) * 2 /vc) * vc \<le> (trec1 - tsend) * vc") defer
      apply (rule mult_right_mono)
      apply force
      apply (insert vc_pos, auto split: split_if_asm)
      done
    qed
  next
    assume "?x \<noteq> ?ev"
    show ?case using prems by auto
  qed
next
  case Nil thus ?case by auto
qed

subsection {* Security for dishonest Provers *}

lemma mdb_secure_dishonest:
  assumes mdb: "tr \<in> mdb"
  and    believe: "(tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>) \<in> set tr"
  shows   "\<exists> P'. d \<ge> pdist (Honest V) (Intruder P')" using prems
proof (induct tr arbitrary: A B trec t rule: mdb.induct)
  case (Fake tr mintr I tsend)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with Fake.hyps prems show ?case by (auto)
next
  case (Con tr tc C mc D tab)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
next
  case (MD1 tr t A NA)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD1.hyps prems show ?case by (auto)
next
  case (MD2 tr tsend trec B NA NB)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD2.hyps prems show ?case by (auto)
next
  case (MD3 tr tsend trec B NA tsend1 NB A)
  hence "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>)) \<in> set tr" by auto
  with MD3.hyps prems show ?case by (auto)
next
  -- {* the only nontrivial case since it adds Claim events *}
  case (MD4 tr t trec2 A B NA NB trec1 tsend)
  let ?x = "(t, Claim (Honest A) \<lbrace>Agent B, Real ((trec1 - tsend)*vc/2)\<rbrace>)"
  and ?ev = "((tdone, Claim (Honest V) \<lbrace>Agent (Intruder P), Real d\<rbrace>))"
  show ?case proof cases
    -- {* the added event is the Claim event from the premise,
          the other case follows trivially from the IH *}
    assume "?x = ?ev"
    hence Aeq: "A=V" and Beq: "B=Intruder P" and deq: "d=(trec1 - tsend)*vc/2" by auto
    let    ?fastmsg = "Xor (Nonce (Honest V) NA) (Hash \<lbrace> NB, Agent (Intruder P)\<rbrace>)"
    have NC_fresh:
      "Nonce (Honest V) NA
       \<notin> usedI (beforeEvent (tsend, Send (Tr (Honest V)) (Nonce (Honest V) NA) []) tr)"
      using prems `tr \<in> mdb` Aeq Beq deq
      apply -
      by (rule nonce_fresh_challenge, auto)

    have "factors ?fastmsg = {Nonce (Honest V) NA, Hash \<lbrace>NB, Agent (Intruder P)\<rbrace>}"
      apply (subgoal_tac "Nonce (Honest V) NA \<notin> factors (Hash \<lbrace>NB, Agent (Intruder P)\<rbrace>)")
      apply (drule factors_Xor_nonce_not_subterm)
      apply (elim disjE)
      apply force
      apply (force simp add: Xor_rewrite)
      apply force
      done
    hence "\<exists> I'. trec1 - tsend \<ge> cdistl (Honest V) (Intruder I') + cdistl (Intruder I') (Honest V)"
      using prems Aeq NC_fresh
      apply -
      apply simp
      apply (rule_tac i=0 and m="?fastmsg" and tr=tr in freshNonce_dishonestAgent_send_recv)
      apply simp
      apply force defer defer defer
      apply force
      apply force
      apply (subst components_non_pair) prefer 2
      apply force defer
      apply simp
      apply simp
      apply clarsimp
      done
    then obtain I where "trec1 - tsend \<ge> cdistl (Honest V) (Intruder I) + cdistl (Intruder I) (Honest V)"
      by auto
    thus ?thesis using deq Aeq vc_pos
      apply (auto simp add: cdistl_def)
      apply (rule_tac x=I in exI)
      apply (auto simp add: pdist_symm)
      apply (subgoal_tac "(2 * pdist (Honest V) (Intruder I) / vc) * vc\<le> (trec1 - tsend) * vc") prefer 2
      apply (rule mult_right_mono)
      apply (auto split: split_if_asm)
      done
  next
    assume "?x \<noteq> ?ev"
    show ?case using prems by auto
  qed
next
  case Nil thus ?case by auto
qed

end
