(******************************************************************************

  ProtoVeriPhy

  Module: distance/NatEmbed.thy

  Copyright (C) 2009      Patrick Schaller <patrick.schaller@inf.ethz.ch>
                2009-2012 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
  License: BSD (see COPYRIGHT)

******************************************************************************)
theory NatEmbed imports Main Divides Power Factorization begin

text {* We want to find a function f, such that
f(x,y) not equal to f(u,v) if the set with x and y is not equal
to the set with u and v. The reason is to find a key
distribution function, assign to every pair of agents a shared secret
key, such that they differ for every distinct pair of agents.

In contrast to Paulson's construct, where there is only one intruder
and therefore only a injective function from nat to nat is needed, for
our case we need to have symmetric keys for all (even dishonest) pairs
of users. This requires an injective function from Agents x Agents to
Keys, both types (Agents and Keys) are type synonyms for natural
numbers.

Another way of modelling this would be to define an additional
datatype for shared symmetric keys and using the injectivity of the
datatype constructor.  *}
 

definition
  primefactors ::"nat \<Rightarrow> nat \<Rightarrow> nat list"
where
  "primefactors a b = (if a < b 
                         then (replicate (a+1) 2)@(replicate (b+1) 3) 
                         else (replicate (b+1) 2)@(replicate (a+1) 3))"



lemma two_repl_primel:"primel (replicate n 2)"
  by (simp add: primel_def)

lemma three_is_prime: "prime (3::nat)"
  apply (auto simp add: prime_nat_def)
  apply (frule dvd_imp_le)
  apply simp
  apply (case_tac m)
  apply simp
  apply (case_tac nat)
  apply simp
  apply (case_tac nata)
  apply simp
  apply arith+
done


lemma three_repl_primel:"primel (replicate n 3)"
  by (simp add: primel_def)

lemma factor_prime:"primel ((replicate n 2)@(replicate m 3))"
  apply (simp add: primel_append)
  apply (rule conjI)
  apply (rule two_repl_primel)
  apply (rule three_repl_primel)
done

lemma replicate_comp: 
  assumes "replicate n m = a # list" 
  shows "a = m" using prems
by (induct n, auto)

lemma nondec_replicate: 
  assumes "nondec (replicate n m)"
  shows "nondec (m # (replicate n m))" using prems
by (case_tac n, auto)


lemma replicate_nondec:"nondec (replicate n m)"
proof (induct n arbitrary: m)
  case 0 show ?case by simp
next
  case (Suc n m) from this show ?case apply -
    apply (simp only: replicate_Suc)
    apply (rule nondec_replicate)
    apply auto
    done
qed

lemma nondec_replicate_append: 
  assumes A: "n \<le> m"
  shows "nondec( (replicate k n) @ (replicate l m))" using A
proof (induct k arbitrary: l)
  case 0 show ?case by (simp ,rule replicate_nondec)
next
  case (Suc k) then show ?case 
    apply (simp only: replicate_Suc)
    apply (simp only: append_Cons)
    apply (simp only: nondec.simps(2))
    apply (cases "replicate k n @ replicate l m")
    apply simp
    apply auto
    apply (cases k)
    apply simp
    apply (drule replicate_comp) 
    apply arith
    apply auto
    done
qed

lemma rep_two_three_nondec:"nondec ((replicate n 2)@(replicate m 3))"
 by (rule nondec_replicate_append, arith)

lemma primefactors_primrel:"primel (primefactors a b)"
  apply (unfold primefactors_def)
  apply (simp only: split_if)
  apply (rule conjI)
  apply (rule impI)
  apply (rule factor_prime)
  apply (rule impI)
  apply (rule factor_prime)
done

lemma primefactors_nondec:"nondec (primefactors a b)"
  apply (unfold primefactors_def)
  apply (simp only: split_if)
  apply (rule conjI)
  apply (rule impI)
  apply (rule rep_two_three_nondec)
  apply (rule impI)
  apply (rule rep_two_three_nondec)
done

lemma primefactors_not_empty:"primefactors a b \<noteq> []"
by (unfold primefactors_def, cases a, cases b, auto)
 

lemma prod_prim_ge0:"prod (primefactors a b) > Suc 0"
by (rule primel_nempty_g_one, rule primefactors_primrel, rule  primefactors_not_empty)


lemma prod_primefactors_equal:
  assumes A:"prod (primefactors a b) = prod (primefactors c d)"
  shows "(primefactors a b) = (primefactors c d)" using A
  apply -
  apply (insert prod_prim_ge0 [of a b])
  apply (frule_tac n="prod (primefactors a b)" in unique_prime_factorization)
  apply (insert primefactors_nondec [of a b])
  apply (insert primefactors_primrel [of a b])
  apply auto
  apply (insert primefactors_nondec [of c d])
  apply (insert primefactors_primrel [of c d])
  apply auto
done

lemma c:
   assumes "a \<noteq> b" and
           "replicate n1 a @ replicate m1 b =
            replicate n2 a @ replicate m2 b"
   shows  "n1 = n2" using prems
proof (cases "m1 =0 \<or> m2 =0")
case True show ?thesis using prems
  apply auto
  apply (case_tac "m2>0")
  apply auto
  apply (drule_tac f="%x. drop n2 x" in HOL.arg_cong)
  apply auto
  apply (case_tac "m1>0")
  apply auto
  apply (drule_tac f="%x. drop n1 x" in HOL.arg_cong)
  apply auto
  done
next
  case False
  hence m1: "m1 > 0" and m2: "m2 > 0" by auto
  show ?thesis proof cases
    assume "n1 = n2"
    thus ?thesis by auto
  next
    assume neq: "n1 \<noteq> n2"
    let ?l1 = "replicate n1 a @ replicate m1 b"
    let ?l2 = "replicate n2 a @ replicate m2 b"
    show ?thesis proof cases
      assume " n1 < n2"
      have A: "?l1!n1 = b" using m1 apply -
	apply (rule_tac ys="drop 1 (replicate m1 b)" in List.nth_via_drop)
	apply auto
	apply (case_tac m1, auto)
	done
      have B: "?l2!n1 = a" using prems apply -
	apply (rule_tac ys="drop (n1+1) ?l2" in List.nth_via_drop)
	apply auto
	apply (cases "n2-n1")
	apply auto
	done
      hence "a = b" using  A B `?l1=?l2`
	apply (drule_tac f="%x. x !n1" in HOL.arg_cong)
	by simp
      thus ?thesis using `a\<noteq>b` by auto
    next
      assume "\<not> (n1 < n2)"
      hence n2: "n2 < n1" using neq by auto
      have A: "?l2!n2 = b" using m2 apply -
	apply (rule_tac ys="drop 1 (replicate m2 b)" in List.nth_via_drop)
	apply auto
	apply (case_tac m2)
	by auto
      have B: "?l1!n2 = a" using n2 m1 m2 apply -
	apply (rule_tac ys="drop (n2+1) ?l1" in List.nth_via_drop)
	apply auto
	apply (cases "n1-n2")
	apply auto
	done
      hence "a = b" using  A B `?l1=?l2`
	apply (drule_tac f="%x. x !n2" in HOL.arg_cong)
	by simp
      thus ?thesis using `a\<noteq>b` by auto
    qed
  qed
qed

lemma replicate_append_length:
  assumes "replicate n1 a @ replicate m1 b =
           replicate n2 a @ replicate m2 b" and
          "a \<noteq> b"
  shows  "n1 = n2 \<and> m1 =m2" using prems
  apply -
  apply (frule c)
  apply assumption
  apply (rule conjI)
by auto


lemma primefactors_unique: 
  assumes A:"primefactors a b = primefactors c d" 
  shows "{a,b} = {c,d}" using A
  apply (unfold primefactors_def)
  apply (simp del: replicate.simps split: split_if_asm)
  apply (auto dest: replicate_append_length simp del:replicate.simps)
done


lemma prod_primf_is_emb:
  assumes "prod (primefactors a b ) = prod (primefactors c d)"
  shows "{a,b} = {c,d}" using prems
proof -
  assume A:" prod (primefactors a b) = prod (primefactors c d)" 
  have B: "(primefactors a b) = (primefactors c d)" using A by (rule prod_primefactors_equal)
  from B show ?thesis by (rule primefactors_unique)
qed

lemma two_set_equal:
  "\<lbrakk>  {a,b} = {c,d};
     \<lbrakk>  a = c; b = d \<rbrakk>  \<Longrightarrow>  P;
     \<lbrakk>  b = c; a = d \<rbrakk>  \<Longrightarrow>  P
   \<rbrakk>  \<Longrightarrow>  P"
  apply (subgoal_tac "a \<in>  {c,d}") prefer 2
  apply force
  apply (subgoal_tac "b \<in>  {c,d}") prefer 2
  apply force
  apply (case_tac "a=c")
   apply (case_tac "b=d")
    apply force
   apply (case_tac "b=c")
    apply force
    apply force
  apply (case_tac "a=d")
  apply force
  apply force
done

lemma eq_imp_primef_eq:
  assumes A:"{a,b} = {c,d}"
  shows "primefactors a b = primefactors c d" using prems
  apply -
  apply (erule two_set_equal)
  apply (unfold primefactors_def)
  apply (simp split: split_if_asm)
  apply auto
done

lemma eq_imp_prod_eq:
  assumes A:"{a,b} = {c,d}"
  shows "prod (primefactors a b) = prod (primefactors c d)" using prems
by (auto dest: eq_imp_primef_eq)

lemma f_inj_prod_inj:
  assumes   A :"prod (primefactors (f a) (f b))= prod (primefactors (f c) (f d))"
  and B:"inj f"
  shows "{a,b} = {c,d}" using prems
  apply -
  apply (drule prod_primf_is_emb)
  apply (simp add: inj_on_def)
  apply (drule two_set_equal)
  apply auto
done

lemma f_inj_primef_eq:
  assumes A:"{a,b} = {c,d}"
  and     B:"inj f"
  shows " prod (primefactors (f a) (f b)) = prod (primefactors (f c) (f d))" using prems
  apply -
  apply (erule two_set_equal)
  apply (unfold primefactors_def)
  apply (simp split: split_if_asm)
  apply auto
done


end